
//
//  CpuPPC.cpp (CPU PowerPC)
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 25 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#ifdef CC_VMP_MAC
#include <Gestalt.h>						// OS 9
//#include <CoreServices/CoreServices.h>	// OS X
#endif

    //  Book Development Library Modules

#include "CpuAsm.h"



int gHasAltivec = 0;				// Has an AltiVec flag (set by Framework)


//
//  CPU Vendor Lookup Table
//

static char *CpuVendor[] =
{
    "UNKNOWN MFG",
    "MOTOROLA",        // CPUVEN_MOTOROLA     = 1
    "IBM",			   // CPUVEN_IBM          = 2
};


static char *szMacOS[] = { "9", "X", "XI", "XII" };
#define szMacOSMax  (sizeof(szMacOS) / sizeof(char *))
#define MacOSBase	9


void CpuDetect( CpuInfo * const pInfo )
{
	OSErr			err;
	long			processorAttributes;
	long 			response;
	bool			b;
	
	ASSERT_PTR4( pInfo );
	
	memset(pInfo, 0, sizeof(CpuInfo));
		
		// Get System Version  $0701 is 7.0.1
		
	err = Gestalt(gestaltSystemVersion, &response);
	if (noErr == err)
	  {
	  	pInfo->nCpuId = (response & 0xff00) >> 8;
	  }
	  
	gHasAltivec = 0;

		// Get Processor Features
		
	err = Gestalt(gestaltPowerPCProcessorFeatures, &processorAttributes);
	if (noErr == err)
	  {
	    b = (1 << gestaltPowerPCHasSquareRootInstructions) & processorAttributes;
		gHasAltivec = (1 << gestaltPowerPCHasVectorInstructions) & processorAttributes;
	
		if ( gHasAltivec )
		  {
			pInfo->nBits |= CPUBITS_ALTIVEC;
		  }
	  }
}



//
// CPU information string
//

char *CpuInfoStr( char *pBuf, const CpuInfo * const pInfo )
{
	char *p;
	
	ASSERT_PTR4(pBuf);
	ASSERT_PTR4(pInfo);
	
	if (( MacOSBase <= pInfo->nCpuId ) && ( pInfo->nCpuId < MacOSBase + szMacOSMax ))
	  {
	  	p = pBuf + sprintf( pBuf, "PowerPC  Mac OS:%s", szMacOS[ pInfo->nCpuId - MacOSBase ] );
	  }
	else
	  {
		p = pBuf + sprintf( pBuf, "PowerPC  Mac OS:%u", pInfo->nCpuId );
      }
      
    p += sprintf( p, " '%s' ", CpuVendor[pInfo->nMfg ] );
     
    if ( CPUBITS_ALTIVEC & pInfo->nBits )
      {
        p += sprintf( p, " AltiVec" );
      }

    if ( CPUBITS_GEKKO & pInfo->nBits )
      {
        p += sprintf( p, " Gekko" );
      } 

	sprintf( p, "\n" );
	
	return pBuf;
}
