
//
//  Bench.cpp - Array of Structures (versus) Structure of Arrays
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      23 Jul 2001 - JCL - Brought up to date.
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries

#include <iostream.h>
#include <assert.h>

    //  Book Development Library Modules

#include "AoS.h"




//
//	Fill each Structure within the array win a value range of n to n+3
//	for visual inspection and learning purposes!
//
//	The same as sequentially filling memory with an incremental floating point value.
//
//	nCnt	 x		 y		 z		 w
//	 0		0.0		1.0		2.0		3.0
//	 1		4.0		5.0		6.0		7.0
//	 2		8.0		9.0		10.0	11.0
//	 3			etc.

static void SetupAoS( AoS *pAry, uint nCnt )
{
	float v;

	ASSERT_PTR(pAry);
	ASSERT(0!=nCnt);

	v = 0.0f;

	if (nCnt)
	  {
		do {
			pAry->x = v;
			pAry->y = v+1.0f;
			pAry->z = v+2.0f;		
			pAry->w = v+3.0f;
			pAry++;
			v += 4.0f;
		  } while(--nCnt);
	  }
}




//
//	Convert AoS to 4K Block SoA quickly!
//               ***AoS***             >>> to >>>            ***SoA***
//	nCnt	 x		 y		 z		 w			
//	 0		0.0		1.0		2.0		3.0				0.0		4.0		8.0		12.0  xxxx
//	 1		4.0		5.0		6.0		7.0				16.0	20.0	24.0	28.0  xxxx
//	 2		8.0		9.0		10.0	11.0			1.0		5.0		9.0		13.0  yyyy
//	 3		12.0	13.0	14.0	15.0			17.0	21.0	25.0	29.0  yyyy
//	 4		16.0	17.0	18.0	19.0			2.0		6.0		10.0	14.0  zzzz
//	 5		20.0	21.0	22.0	23.0			18.0	22.0	25.0	30.0  zzzz
//	 6		24.0	25.0	26.0	27.0			3.0		7.0		11.0	15.0  wwww
//	 7		28.0	29.0	30.0	31.0			19.0	23.0	26.0	31.0  wwww
//	-------------------------------------           -----------------------------------
//	 8		32.0	33.0	34.0	35.0			32.0	36.0	40.0	44.0  xxxx
//	 9		36.0	37.0	38.0	39.0			48.0	52.0	56.0	60.0  xxxx


static void vmp_SoA8_AoS( SoABlk *pDAry, const AoS *pSAry, uint nCnt )
{
	uint n;
	float *pD, *pS;

	ASSERT_PTR(pSAry);
	ASSERT_PTR(pDAry);
	ASSERT(0!=nCnt);

	pS = (float*)pSAry;
	pD = (float*)pDAry;

		// 8 or more?
	if ( nCnt > 8 )
	  {
		n = nCnt >> 3;
		nCnt &= 0x07;

		do {			// One complete 8x {XYZW} structure each pass!
			*(pD+0)=*(pS+0);  *(pD+8)=*(pS+1);   *(pD+16)=*(pS+2);  *(pD+24)=*(pS+3);
			*(pD+1)=*(pS+4);  *(pD+9)=*(pS+5);   *(pD+17)=*(pS+6);  *(pD+25)=*(pS+7);
			*(pD+2)=*(pS+8);  *(pD+10)=*(pS+9);  *(pD+18)=*(pS+10); *(pD+26)=*(pS+11);
			*(pD+3)=*(pS+12); *(pD+11)=*(pS+13); *(pD+19)=*(pS+14); *(pD+27)=*(pS+15);
			*(pD+4)=*(pS+16); *(pD+12)=*(pS+17); *(pD+20)=*(pS+18); *(pD+28)=*(pS+19);
			*(pD+5)=*(pS+20); *(pD+13)=*(pS+21); *(pD+21)=*(pS+22); *(pD+29)=*(pS+23);
			*(pD+6)=*(pS+24); *(pD+14)=*(pS+25); *(pD+22)=*(pS+26); *(pD+30)=*(pS+27);
			*(pD+7)=*(pS+28); *(pD+15)=*(pS+29); *(pD+23)=*(pS+30); *(pD+31)=*(pS+31);
			pS += SOA_BLK_MAX;	// 32
			pD += SOA_BLK_MAX;
		  } while(--n);
	  }

		//	Get any remainder {XYZW} structures

	if (nCnt)
	  {
	   do {
			*(pD+0)=*(pS+0);  *(pD+8)=*(pS+1);   *(pD+16)=*(pS+2);  *(pD+24)=*(pS+3);
			pS += 4;
			pD++;
		  } while(--nCnt);
	  }
}



//
//	Convert 4K (Blocked) SoA to AoS quickly!
//               ***AoS***             <<< to <<<            ***SoA***
//	nCnt	 x		 y		 z		 w			
//	 0		0.0		1.0		2.0		3.0				0.0		4.0		8.0		12.0  xxxx
//	 1		4.0		5.0		6.0		7.0				16.0	20.0	24.0	28.0  xxxx
//	 2		8.0		9.0		10.0	11.0			1.0		5.0		9.0		13.0  yyyy
//	 3		12.0	13.0	14.0	15.0			17.0	21.0	25.0	29.0  yyyy
//	 4		16.0	17.0	18.0	19.0			2.0		6.0		10.0	14.0  zzzz
//	 5		20.0	21.0	22.0	23.0			18.0	22.0	25.0	30.0  zzzz
//	 6		24.0	25.0	26.0	27.0			3.0		7.0		11.0	15.0  wwww
//	 7		28.0	29.0	30.0	31.0			19.0	23.0	26.0	31.0  wwww
//	-------------------------------------           -----------------------------------
//	 8		32.0	33.0	34.0	35.0			32.0	36.0	40.0	44.0  xxxx
//	 9		36.0	37.0	38.0	39.0			48.0	52.0	56.0	60.0  xxxx


static void vmp_AoS_SoA8( AoS *pDAry, const SoABlk *pSAry, uint nCnt )
{
	uint n;
	float *pD, *pS;

	ASSERT_PTR(pSAry);
	ASSERT_PTR(pDAry);
	ASSERT(0!=nCnt);

	pS = (float*)pSAry;
	pD = (float*)pDAry;

		// 8 or more?
	if ( nCnt > 8 )
	  {
		n = nCnt >> 3;
		nCnt &= 0x07;

		do {
			*(pD+0)=*(pS+0);  *(pD+1)=*(pS+8);   *(pD+2)=*(pS+16);  *(pD+3)=*(pS+24);
			*(pD+4)=*(pS+1);  *(pD+5)=*(pS+9);   *(pD+6)=*(pS+17);  *(pD+7)=*(pS+25);
			*(pD+8)=*(pS+2);  *(pD+9)=*(pS+10);  *(pD+10)=*(pS+18); *(pD+11)=*(pS+26);
			*(pD+12)=*(pS+3); *(pD+13)=*(pS+11); *(pD+14)=*(pS+19); *(pD+15)=*(pS+27);
			*(pD+16)=*(pS+4); *(pD+17)=*(pS+12); *(pD+18)=*(pS+20); *(pD+19)=*(pS+28);
			*(pD+20)=*(pS+5); *(pD+21)=*(pS+13); *(pD+22)=*(pS+21); *(pD+23)=*(pS+29);
			*(pD+24)=*(pS+6); *(pD+25)=*(pS+14); *(pD+26)=*(pS+22); *(pD+27)=*(pS+30);
			*(pD+28)=*(pS+7); *(pD+29)=*(pS+15); *(pD+30)=*(pS+23); *(pD+31)=*(pS+31);
			pS += 32;
			pD += 32;
		  } while(--n);
	  }

		//	Get any remainders
	if (nCnt)
	  {
	   do {
			*(pD+0)=*(pS+0);  *(pD+1)=*(pS+8);   *(pD+2)=*(pS+16);  *(pD+3)=*(pS+24);
			pS++;
			pD += 4;
		  } while(--nCnt);
	  }
}



//	
//	Compare two 4K SoA blocks
//

static bool vmp_CmpAoS( const AoS * const pD, const AoS * const pS, uint nCnt )
{
	ASSERT_PTR(pD);
	ASSERT_PTR(pS);
	ASSERT(0!=nCnt);

	if (nCnt)
	  {
		do {
			if ((pD->x != pS->x) || (pD->y != pS->y) || (pD->z != pS->z) || (pD->w != pS->w))
			  {
				return false;
			  }
		  } while(--nCnt);
	  }

	return true;
}



int main( int argc, const char *argv[] )
{
	SoABlk	soaTmp;
	AoS		aosSrc[ SOA_MAX ], aosDst[ SOA_MAX ];

#pragma unused(argc, argv)

	cout << "Begin AoS demonstration!" << endl;

	SetupAoS( aosSrc, SOA_MAX );		// Fill a 4K (256 structure) block with sequential values 0.0 to 1023.0

		//	Convert the 4K block of AoS to SoA (Quickly)!

	cout << "Converting AoS to SoA" << endl;

	vmp_SoA8_AoS( &soaTmp, aosSrc, SOA_MAX );

		//	Convert the 4K block of AoS back to SoA

	cout << "Converting SoA to AoS" << endl;

	vmp_AoS_SoA8( aosDst, &soaTmp, SOA_MAX );

		//	Now compare them!

	if ( !vmp_CmpAoS( aosSrc, aosDst, SOA_MAX ) )
	{
		cout << "MISMATCH"<< endl;
	}
	else
	{
		cout << "Matched!" << endl;
	}

	cout << "Fin!\n" << flush;

    return 0;
}

