
//
//  PBool.cpp - Packed Boolean Logic
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      19 Dec 2001 - JCL - Brought up to date.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules
    
#include "PBool.h"
#include "PBoolEx.h"



vmp_pnotProc vmp_pnot;
vmp_pnotProc vmp_pnot0;

vmp_pandProc vmp_pand;
vmp_pandProc vmp_pand0;

vmp_pandcProc vmp_pandc;
vmp_pandcProc vmp_pandc0;

vmp_porProc vmp_por;
vmp_porProc vmp_por0;

vmp_pxorProc vmp_pxor;
vmp_pxorProc vmp_pxor0;

vmp_pnorProc vmp_pnor;
vmp_pnorProc vmp_pnor0;



//
//	Logical NOT 1's complement  D=(!A)
//

void vmp_pnotGeneric(  void * const pvD, const void * const pvA )
{
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32*)pvD)+0) = 0xffffffff ^ *(((uint32*)pvA)+0);
	*(((uint32*)pvD)+1) = 0xffffffff ^ *(((uint32*)pvA)+1);
	*(((uint32*)pvD)+2) = 0xffffffff ^ *(((uint32*)pvA)+2);
	*(((uint32*)pvD)+3) = 0xffffffff ^ *(((uint32*)pvA)+3);
}



//
//	Logical AND  D=(A & B)
//

void vmp_pandGeneric(  void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32*)pvD)+0) = *(((uint32*)pvA)+0) & *(((uint32*)pvB)+0);
	*(((uint32*)pvD)+1) = *(((uint32*)pvA)+1) & *(((uint32*)pvB)+1);
	*(((uint32*)pvD)+2) = *(((uint32*)pvA)+2) & *(((uint32*)pvB)+2);
	*(((uint32*)pvD)+3) = *(((uint32*)pvA)+3) & *(((uint32*)pvB)+3);
}



//
//	Logical OR  D=(A | B)
//

void vmp_porGeneric(   void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32*)pvD)+0) = *(((uint32*)pvA)+0) | *(((uint32*)pvB)+0);
	*(((uint32*)pvD)+1) = *(((uint32*)pvA)+1) | *(((uint32*)pvB)+1);
	*(((uint32*)pvD)+2) = *(((uint32*)pvA)+2) | *(((uint32*)pvB)+2);
	*(((uint32*)pvD)+3) = *(((uint32*)pvA)+3) | *(((uint32*)pvB)+3);
}



//
//	Logical XOR (Exclusive OR)  D=(A^B)
//

void vmp_pxorGeneric(  void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32*)pvD)+0) = *(((uint32*)pvA)+0) ^ *(((uint32*)pvB)+0);
	*(((uint32*)pvD)+1) = *(((uint32*)pvA)+1) ^ *(((uint32*)pvB)+1);
	*(((uint32*)pvD)+2) = *(((uint32*)pvA)+2) ^ *(((uint32*)pvB)+2);
	*(((uint32*)pvD)+3) = *(((uint32*)pvA)+3) ^ *(((uint32*)pvB)+3);
}



//
//	Logical And w/1's Complement  D=(A & !B)
//
void vmp_pandcGeneric( void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32*)pvD)+0) = (0xffffffff ^ *(((uint32*)pvB)+0)) & *(((uint32*)pvA)+0);
	*(((uint32*)pvD)+1) = (0xffffffff ^ *(((uint32*)pvB)+1)) & *(((uint32*)pvA)+1);
	*(((uint32*)pvD)+2) = (0xffffffff ^ *(((uint32*)pvB)+2)) & *(((uint32*)pvA)+2);
	*(((uint32*)pvD)+3) = (0xffffffff ^ *(((uint32*)pvB)+3)) & *(((uint32*)pvA)+3);
}



//
//	Logical NOR  1's complement of OR  D=!(A | B)
//

void vmp_pnorGeneric( void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32*)pvD)+0) = 0xffffffff ^ ( *(((uint32*)pvA)+0) | *(((uint32*)pvB)+0) );
	*(((uint32*)pvD)+1) = 0xffffffff ^ ( *(((uint32*)pvA)+1) | *(((uint32*)pvB)+1) );
	*(((uint32*)pvD)+2) = 0xffffffff ^ ( *(((uint32*)pvA)+2) | *(((uint32*)pvB)+2) );
	*(((uint32*)pvD)+3) = 0xffffffff ^ ( *(((uint32*)pvA)+3) | *(((uint32*)pvB)+3) );
}






