
//
//  PBoolAltivec.cpp - Packed Boolean Logic
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 27 Dec, 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "PBool.h"
#include "PBoolEx.h"


vector signed int vNOT = (vector signed int) (-1, -1, -1, -1);


//______________________________________________________
//
//	Logical NOT 1's complement  D=(!A)
//

void vmp_pnotAltivec0(  void * const pvD, const void * const pvA )
{
	vector signed int vD, vA;

	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pvA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pvA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pvA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pvA)+3);
	
	vD = vec_xor( vA, vNOT );

	*(((uint32 *)pvD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pvD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pvD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pvD)+3) = *(((uint32 *)&vD)+3);
}								



//
//	Logical NOT 1's complement  D=(!A)  (Aligned)
//

void vmp_pnotAltivec(  void * const pvD, const void * const pvA )
{
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvD);

	*(vector signed int *)pvD = vec_xor( (*(vector signed int *)pvA), vNOT );
}		




//______________________________________________________
//
//	Logical AND  D=(A & B) (Unaligned)
//

void vmp_pandAltivec0(  void * const pvD, const void * const pvA, const void * const pvB )
{
	vector unsigned int vD, vA, vB;

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pvA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pvA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pvA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pvA)+3);
	
	*(((uint32 *)&vB)+0) = *(((uint32 *)pvB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pvB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pvB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pvB)+3);
	
	vD = vec_and( vA, vB );				// vand d,a,b

	*(((uint32 *)pvD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pvD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pvD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pvD)+3) = *(((uint32 *)&vD)+3);
}								



//
//	Logical AND  D=(A & B) (Aligned)
//

void vmp_pandAltivec(  void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR16(pvB);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvD);

	*(vector unsigned int *)pvD = vec_and( (*(vector unsigned int *)pvA), (*(vector unsigned int *)pvB) );		// vand d,a,b
}		



//______________________________________________________
//
//	Logical OR  D=(A | B)  (Aligned)
//

void vmp_porAltivec0(  void * const pvD, const void * const pvA, const void * const pvB )
{
	vector unsigned int vD, vA, vB;

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pvA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pvA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pvA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pvA)+3);
	
	*(((uint32 *)&vB)+0) = *(((uint32 *)pvB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pvB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pvB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pvB)+3);
	
	vD = vec_or( vA, vB );				// vor d,a,b

	*(((uint32 *)pvD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pvD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pvD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pvD)+3) = *(((uint32 *)&vD)+3);
}								



//
//	Logical OR  D=(A | B)  (Aligned)
//

void vmp_porAltivec(  void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR16(pvB);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvD);

	*(vector unsigned int *)pvD = vec_or( (*(vector unsigned int *)pvA), (*(vector unsigned int *)pvB) );		// vor d,a,b
}		



//______________________________________________________
//
//	Logical XOR (Exclusive OR)  D=(A^B)  (Unaligned)
//

void vmp_pxorAltivec0(  void * const pvD, const void * const pvA, const void * const pvB )
{
	vector unsigned int vD, vA, vB;

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pvA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pvA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pvA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pvA)+3);
	
	*(((uint32 *)&vB)+0) = *(((uint32 *)pvB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pvB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pvB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pvB)+3);
	
	vD = vec_xor( vA, vB );				// vxor d,a,b

	*(((uint32 *)pvD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pvD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pvD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pvD)+3) = *(((uint32 *)&vD)+3);
}								



//
//	Logical XOR  D=(A | B)  (Aligned)
//

void vmp_pxorAltivec(  void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR16(pvB);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvD);

	*(vector unsigned int *)pvD = vec_xor( (*(vector unsigned int *)pvA), (*(vector unsigned int *)pvB) );		// vxor d,a,b
}		



//______________________________________________________
//
//	Logical And w/1's Complement  D=(A & !B)  (Unaligned)
//

void vmp_pandcAltivec0(  void * const pvD, const void * const pvA, const void * const pvB )
{
	vector unsigned int vD, vA, vB;

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pvA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pvA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pvA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pvA)+3);
	
	*(((uint32 *)&vB)+0) = *(((uint32 *)pvB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pvB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pvB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pvB)+3);
	
	vD = vec_andc( vA, vB );				// vandc d,a,b

	*(((uint32 *)pvD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pvD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pvD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pvD)+3) = *(((uint32 *)&vD)+3);
}								



//
//	Logical And w/1's Complement  D=(A & !B)  (Aligned)
//

void vmp_pandcAltivec(  void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR16(pvB);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvD);

	*(vector unsigned int *)pvD = vec_andc( (*(vector unsigned int *)pvA), (*(vector unsigned int *)pvB) );		// vandc d,a,b
}		



//______________________________________________________
//
//	Logical NOR  1's complement of OR  D=!(A | B)  (Unaligned)
//

void vmp_pnorAltivec0(  void * const pvD, const void * const pvA, const void * const pvB )
{
	vector unsigned int vD, vA, vB;

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pvA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pvA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pvA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pvA)+3);
	
	*(((uint32 *)&vB)+0) = *(((uint32 *)pvB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pvB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pvB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pvB)+3);
	
	vD = vec_nor( vA, vB );				// vnor d,a,b

	*(((uint32 *)pvD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pvD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pvD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pvD)+3) = *(((uint32 *)&vD)+3);
}								



//
//	Logical NOR  1's complement of OR  D=!(A | B)  (Aligned)
//

void vmp_pnorAltivec(  void * const pvD, const void * const pvA, const void * const pvB )
{
	ASSERT_PTR16(pvB);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvD);

	*(vector unsigned int *)pvD = vec_nor( (*(vector unsigned int *)pvA), (*(vector unsigned int *)pvB) );		// vnor d,a,b
}		




//
//	Setup Packed Boolean Logic
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void PBoolSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);
	    
	if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_pand =				vmp_pandAltivec;
		vmp_pand0 =				vmp_pandAltivec0;
		vmp_pandc =				vmp_pandcAltivec;
		vmp_pandc0 =			vmp_pandcAltivec0;
		vmp_por =				vmp_porAltivec;
		vmp_por0 =				vmp_porAltivec0;
		vmp_pxor =				vmp_pxorAltivec;
		vmp_pxor0 =				vmp_pxorAltivec0;
		vmp_pnor =				vmp_pnorAltivec;
		vmp_pnor0 =				vmp_pnorAltivec0;
		vmp_pnot =				vmp_pnotAltivec;
		vmp_pnot0 =				vmp_pnotAltivec0;
      }

	else
	  {
		vmp_pand =				vmp_pandPPC;
		vmp_pand0 =				vmp_pandPPC;
		vmp_pandc =				vmp_pandcPPC;
		vmp_pandc0 =			vmp_pandcPPC;
		vmp_por =				vmp_porPPC;
		vmp_por0 =				vmp_porPPC;
		vmp_pxor =				vmp_pxorPPC;
		vmp_pxor0 =				vmp_pxorPPC;
		vmp_pnor =				vmp_pnorPPC;
		vmp_pnor0 =				vmp_pnorPPC;
		vmp_pnot =				vmp_pnotPPC;
		vmp_pnot0 =				vmp_pnotPPC;
	  }
}




