
//
//  PBoolPPC.cpp - Packed Boolean Logic
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 27 Dec, 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "PBool.h"
#include "PBoolEx.h"




//______________________________________________________
//
//	Logical NOT 1's complement  D=(!A)  (Un)Aligned
//

void vmp_pnotPPC(  void * const pvD, const void * const pvA )
{
	unsigned int register a0, a1, a2, a3;
	
#ifndef USE_ASSERT
#pragma unused(pvD, pvA)
#endif

	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	__asm {
		lwz	a0,0(r4)									// Read A Bits 0...31
		lwz	a1,4(r4)
		lwz	a2,8(r4)
		lwz	a3,12(r4)									//      "     96...127
		
		not	a0,a0
		not	a1,a1
		not	a2,a2
		not	a3,a3
		
		stw	a0,0(r3)									// Write Bits 0...31
		stw	a1,4(r3)
		stw	a2,8(r3)
		stw	a3,12(r3)									//      "     96...127
	}
}		




//______________________________________________________
//
//	Logical AND  D=(A & B) (Un)Aligned
//

void vmp_pandPPC(  void * const pvD, const void * const pvA, const void * const pvB )
{
	unsigned int register a0, a1, a2, a3, b0, b1, b2, b3;
	
#ifndef USE_ASSERT
#pragma unused(pvD, pvA, pvB)
#endif

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	__asm {
		lwz	a0,0(r4)									// Read A Bits 0...31
		lwz	a1,4(r4)
		lwz	a2,8(r4)
		lwz	a3,12(r4)									//      "     96...127
		
		lwz	b0,0(r5)									// Read B Bits 0...31
		lwz	b1,4(r5)
		lwz	b2,8(r5)
		lwz	b3,12(r5)									//      "     96...127
		
		and	a0,a0,b0									// AND   Bits 0...31
		and	a1,a1,b1
		and	a2,a2,b2
		and	a3,a3,b3									//      "     96...127
		
		stw	a0,0(r3)									// Write Bits 0...31
		stw	a1,4(r3)
		stw	a2,8(r3)
		stw	a3,12(r3)									//      "     96...127
	}
}		



//______________________________________________________
//
//	Logical OR  D=(A | B)  (Un)Aligned
//

void vmp_porPPC(  void * const pvD, const void * const pvA, const void * const pvB )
{
	unsigned int register a0, a1, a2, a3, b0, b1, b2, b3;
	
#ifndef USE_ASSERT
#pragma unused(pvD, pvA, pvB)
#endif

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	__asm {
		lwz	a0,0(r4)									// Read A Bits 0...31
		lwz	a1,4(r4)
		lwz	a2,8(r4)
		lwz	a3,12(r4)									//      "     96...127
		
		lwz	b0,0(r5)									// Read B Bits 0...31
		lwz	b1,4(r5)
		lwz	b2,8(r5)
		lwz	b3,12(r5)									//      "     96...127
		
		or	a0,a0,b0									// OR    Bits 0...31
		or	a1,a1,b1
		or	a2,a2,b2
		or	a3,a3,b3									//      "     96...127
		
		stw	a0,0(r3)									// Write Bits 0...31
		stw	a1,4(r3)
		stw	a2,8(r3)
		stw	a3,12(r3)									//      "     96...127
	}
}		



//______________________________________________________
//
//	Logical XOR  D=(A | B)  (Un)Aligned
//

void vmp_pxorPPC(  void * const pvD, const void * const pvA, const void * const pvB )
{
	unsigned int register a0, a1, a2, a3, b0, b1, b2, b3;
	
#ifndef USE_ASSERT
#pragma unused(pvD, pvA, pvB)
#endif

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	__asm {
		lwz	a0,0(r4)									// Read A Bits 0...31
		lwz	a1,4(r4)
		lwz	a2,8(r4)
		lwz	a3,12(r4)									//      "     96...127
		
		lwz	b0,0(r5)									// Read B Bits 0...31
		lwz	b1,4(r5)
		lwz	b2,8(r5)
		lwz	b3,12(r5)									//      "     96...127
		
		xor	a0,a0,b0									// XOR   Bits 0...31
		xor	a1,a1,b1
		xor	a2,a2,b2
		xor	a3,a3,b3									//      "     96...127
		
		stw	a0,0(r3)									// Write Bits 0...31
		stw	a1,4(r3)
		stw	a2,8(r3)
		stw	a3,12(r3)									//      "     96...127
	}
}		



//______________________________________________________
//
//	Logical And w/1's Complement  D=(A & !B)  (Un)Aligned
//

void vmp_pandcPPC(  void * const pvD, const void * const pvA, const void * const pvB )
{
	unsigned int register a0, a1, a2, a3, b0, b1, b2, b3;
	
#ifndef USE_ASSERT
#pragma unused(pvD, pvA, pvB)
#endif

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	__asm {
		lwz	a0,0(r4)									// Read A Bits 0...31
		lwz	a1,4(r4)
		lwz	a2,8(r4)
		lwz	a3,12(r4)									//      "     96...127
		
		lwz	b0,0(r5)									// Read B Bits 0...31
		lwz	b1,4(r5)
		lwz	b2,8(r5)
		lwz	b3,12(r5)									//      "     96...127
		
		andc a0,a0,b0									// ANDC   Bits 0...31
		andc a1,a1,b1
		andc a2,a2,b2
		andc a3,a3,b3									//      "     96...127
			
		stw	a0,0(r3)									// Write Bits 0...31
		stw	a1,4(r3)
		stw	a2,8(r3)
		stw	a3,12(r3)									//      "     96...127
	}
}		



//______________________________________________________
//
//	Logical NOR  1's complement of OR  D=!(A | B)  (Un)Aligned
//

void vmp_pnorPPC(  void * const pvD, const void * const pvA, const void * const pvB )
{
	unsigned int register a0, a1, a2, a3, b0, b1, b2, b3;
	
#ifndef USE_ASSERT
#pragma unused(pvD, pvA, pvB)
#endif

	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvD);

	__asm {
		lwz	a0,0(r4)									// Read A Bits 0...31
		lwz	a1,4(r4)
		lwz	a2,8(r4)
		lwz	a3,12(r4)									//      "     96...127
		
		lwz	b0,0(r5)									// Read B Bits 0...31
		lwz	b1,4(r5)
		lwz	b2,8(r5)
		lwz	b3,12(r5)									//      "     96...127
		
		nor	a0,a0,b0									// NOR     Bits 0...31
		nor	a1,a1,b1
		nor	a2,a2,b2
		nor	a3,a3,b3									//      "     96...127
		
		stw	a0,0(r3)									// Write Bits 0...31
		stw	a1,4(r3)
		stw	a2,8(r3)
		stw	a3,12(r3)									//      "     96...127
	}
}		




