
//
//  PRot.cpp - Packed Rotation 
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      23 Jul 2001 - JCL - Brought up to date.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules
    
#include "PRot.h"
#include "PRotEx.h"

	//	Shift Right Arithmetic

vmp_psraBProc vmp_psraB;
vmp_psraBProc vmp_psraB0;
vmp_psraHProc vmp_psraH;
vmp_psraHProc vmp_psraH0;
vmp_psraWProc vmp_psraW;
vmp_psraWProc vmp_psraW0;

vmp_psra1BProc vmp_psra1B;
vmp_psra1BProc vmp_psra1B0;
vmp_psra1HProc vmp_psra1H;
vmp_psra1HProc vmp_psra1H0;
vmp_psra1WProc vmp_psra1W;
vmp_psra1WProc vmp_psra1W0;

	//	Shift Left Logical

vmp_psllBProc vmp_psllB;
vmp_psllBProc vmp_psllB0;
vmp_psllHProc vmp_psllH;
vmp_psllHProc vmp_psllH0;
vmp_psllWProc vmp_psllW;
vmp_psllWProc vmp_psllW0;

vmp_psll1BProc vmp_psll1B;
vmp_psll1BProc vmp_psll1B0;
vmp_psll1HProc vmp_psll1H;
vmp_psll1HProc vmp_psll1H0;
vmp_psll1WProc vmp_psll1W;
vmp_psll1WProc vmp_psll1W0;

	//	Shift Right Logical

vmp_psrlBProc vmp_psrlB;
vmp_psrlBProc vmp_psrlB0;
vmp_psrlHProc vmp_psrlH;
vmp_psrlHProc vmp_psrlH0;
vmp_psrlWProc vmp_psrlW;
vmp_psrlWProc vmp_psrlW0;

vmp_psrl1BProc vmp_psrl1B;
vmp_psrl1BProc vmp_psrl1B0;
vmp_psrl1HProc vmp_psrl1H;
vmp_psrl1HProc vmp_psrl1H0;
vmp_psrl1WProc vmp_psrl1W;
vmp_psrl1WProc vmp_psrl1W0;

	//	Rotate Left

vmp_prolBProc vmp_prolB;
vmp_prolBProc vmp_prolB0;
vmp_prolHProc vmp_prolH;
vmp_prolHProc vmp_prolH0;
vmp_prolWProc vmp_prolW;
vmp_prolWProc vmp_prolW0;

vmp_prol1BProc vmp_prol1B;
vmp_prol1BProc vmp_prol1B0;
vmp_prol1HProc vmp_prol1H;
vmp_prol1HProc vmp_prol1H0;
vmp_prol1WProc vmp_prol1W;
vmp_prol1WProc vmp_prol1W0;



//	Left Shift 32bit (unsigned) 8bit mask
uint32 llMaskBD[] = { 0xffffffff, 0xfefefefe, 0xfcfcfcfc, 0xf8f8f8f8,
					 0xf0f0f0f0, 0xe0e0e0e0, 0xc0c0c0c0, 0x80808080, 0 };

//	Right Shift 32bit (unsigned) 8bit mask
uint32 rrMaskBD[] = { 0xffffffff, 0x7f7f7f7f, 0x3f3f3f3f, 0x1f1f1f1f,
					 0x0f0f0f0f, 0x07070707, 0x03030303, 0x01010101, 0 };

//	Left Shift 32bit (unsigned) 16bit mask
uint32 llMaskHD[] = { 0xffffffff, 0xfffefffe, 0xfffcfffc, 0xfff8fff8,
					 0xfff0fff0, 0xffe0ffe0, 0xffc0ffc0, 0xff80ff80,
					 0xff00ff00, 0xfe00fe00, 0xfc00fc00, 0xf800f800,
					 0xf000f000, 0xe000e000, 0xc000c000, 0x80008000, 0 };

//	Right Shift 32bit (unsigned) 16bit mask
uint32 rrMaskHD[] = { 0xffffffff, 0x7fff7fff, 0x3fff3fff, 0x1fff1fff,
					 0x0fff0fff, 0x07ff07ff, 0x03ff03ff, 0x01ff01ff,
					 0x00ff00ff, 0x007f007f, 0x003f003f, 0x001f001f,
					 0x000f000f, 0x00070007, 0x00030003, 0x00010001, 0 };

//
//	Packed Shift Right Arithmetic 16x8bit by n:{0...7}
//
void vmp_psraBGeneric( int8 * const pbD, const int8 * const pbA, uint count )
{
	int32 val, *pA;
	uint32 *pD;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	pA = (int32*)pbA;
	pD = (uint32*)pbD;

	count &= (8-1);				// Clip count to data bit size -1

	val = *(pA+0);
	*(pD+0) = (((uint32)((val<<24) >> count)) >> 24)
		| ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		| ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)
		| ((val >> count) & 0xFF000000);

	val = *(pA+1);
	*(pD+1) = (((uint32)((val<<24) >> count)) >> 24)
		| ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		| ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)
		| ((val >> count) & 0xFF000000);

	val = *(pA+2);
	*(pD+2) = (((uint32)((val<<24) >> count)) >> 24)
		| ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		| ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)
		| ((val >> count) & 0xFF000000);

	val = *(pA+3);
	*(pD+3) = (((uint32)((val<<24) >> count)) >> 24)
		| ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		| ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)
		| ((val >> count) & 0xFF000000);

}									



//
//	Packed Shift Right Arithmetic 16x8bit by 1
//
void vmp_psra1BGeneric( int8 * const pbD, const int8 * const pbA )
{
	vmp_psraBGeneric( pbD, pbA, 1 );
}



//
//	Packed Shift Right Arithmetic 8x16bit by n:{0...15}
//
void vmp_psraHGeneric( int16 * const phD, const int16 * const phA, uint count )
{
	int32 val, *pA;
	uint32 *pD;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	pA = (int32*)phA;
	pD = (uint32*)phD;

	count &= (16-1);				// Clip count to data bit size -1

	val = *(pA+0);
	*(pD+0) = (((uint32)((val<<16) >> count)) >> 16)
		| ((val >> count) & 0xFFFF0000);

	val = *(pA+1);
	*(pD+1) = (((uint32)((val<<16) >> count)) >> 16)
		| ((val >> count) & 0xFFFF0000);

	val = *(pA+2);
	*(pD+2) = (((uint32)((val<<16) >> count)) >> 16)
		| ((val >> count) & 0xFFFF0000);

	val = *(pA+3);
	*(pD+3) = (((uint32)((val<<16) >> count)) >> 16)
		| ((val >> count) & 0xFFFF0000);
}									



//
//	Packed Shift Right Arithmetic 8x16bit by 1
//
void vmp_psra1HGeneric( int16 * const phD, const int16 * const phA )
{
	vmp_psraHGeneric( phD, phA, 1 );
}



//
//	Packed Shift Right Arithmetic 4x32bit by n:{0...31}
//
void vmp_psraWGeneric( int32 * const pwD, const int32 * const pwA, uint count )
{
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(pwD+0) = (*(pwA+0)) >> count;
	*(pwD+1) = (*(pwA+1)) >> count;
	*(pwD+2) = (*(pwA+2)) >> count;
	*(pwD+3) = (*(pwA+3)) >> count;
}									



//
//	Packed Shift Right Arithmetic 4x32bit by 1
//
void vmp_psra1WGeneric( int32 * const pwD, const int32 * const pwA )
{
	vmp_psraWGeneric( pwD, pwA, 1 );
}


//______________________________________________________


//
//	Packed Shift Left Logical 16x8bit by n:{0...7}
//
void vmp_psllBGeneric( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	uint32 msk;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	count &= (8-1);				// Clip count to data bit size -1

	msk = rrMaskBD[count];
	*(((uint32*)pbD)+0) = (*(((uint32*)pbA)+0) & msk) << count;
	*(((uint32*)pbD)+1) = (*(((uint32*)pbA)+1) & msk) << count;
	*(((uint32*)pbD)+2) = (*(((uint32*)pbA)+2) & msk) << count;
	*(((uint32*)pbD)+3) = (*(((uint32*)pbA)+3) & msk) << count;
}									



//
//	Packed Shift Left Logical 16x8bit by 1
//
void vmp_psll1BGeneric( uint8 * const pbD, const uint8 * const pbA )
{
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32*)pbD)+0) = (*(((uint32*)pbA)+0) & 0x7f7f7f7f) << 1;
	*(((uint32*)pbD)+1) = (*(((uint32*)pbA)+1) & 0x7f7f7f7f) << 1;
	*(((uint32*)pbD)+2) = (*(((uint32*)pbA)+2) & 0x7f7f7f7f) << 1;
	*(((uint32*)pbD)+3) = (*(((uint32*)pbA)+3) & 0x7f7f7f7f) << 1;
}



//
//	Packed Shift Left Logical 8x16bit by n:{0...15}
//
void vmp_psllHGeneric( uint16 * const phD, const uint16 * const phA, uint count )
{
	uint32 msk;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	count &= (16-1);				// Clip count to data bit size -1

	msk = rrMaskHD[count];
	*(((uint32*)phD)+0) = (*(((uint32*)phA)+0) & msk) << count;
	*(((uint32*)phD)+1) = (*(((uint32*)phA)+1) & msk) << count;
	*(((uint32*)phD)+2) = (*(((uint32*)phA)+2) & msk) << count;
	*(((uint32*)phD)+3) = (*(((uint32*)phA)+3) & msk) << count;
}									



//
//	Packed Shift Left Logical 8x16bit by 1
//
void vmp_psll1HGeneric( uint16 * const phD, const uint16 * const phA )
{
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	*(((uint32*)phD)+0) = (*(((uint32*)phA)+0) & 0x7fff7fff) << 1;
	*(((uint32*)phD)+1) = (*(((uint32*)phA)+1) & 0x7fff7fff) << 1;
	*(((uint32*)phD)+2) = (*(((uint32*)phA)+2) & 0x7fff7fff) << 1;
	*(((uint32*)phD)+3) = (*(((uint32*)phA)+3) & 0x7fff7fff) << 1;
}



//
//	Packed Shift Left Logical 4x32bit by n:{0...31}
//
void vmp_psllWGeneric( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(pwD+0) = (*(pwA+0)) << count;
	*(pwD+1) = (*(pwA+1)) << count;
	*(pwD+2) = (*(pwA+2)) << count;
	*(pwD+3) = (*(pwA+3)) << count;
}									



//
//	Packed Shift Left Logical 4x32bit by 1
//
void vmp_psll1WGeneric( uint32 * const pwD, const uint32 * const pwA )
{
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(pwD+0) = *(pwA+0) << 1;
	*(pwD+1) = *(pwA+1) << 1;
	*(pwD+2) = *(pwA+2) << 1;
	*(pwD+3) = *(pwA+3) << 1;
}



//______________________________________________________


//
//	Packed Shift Right Logical 16x8bit by n:{0...7}
//
void vmp_psrlBGeneric( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	uint32 msk;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	count &= (8-1);				// Clip count to data bit size -1

	msk = llMaskBD[count];
	*(((uint32*)pbD)+0) = (*(((uint32*)pbA)+0) & msk) >> count;
	*(((uint32*)pbD)+1) = (*(((uint32*)pbA)+1) & msk) >> count;
	*(((uint32*)pbD)+2) = (*(((uint32*)pbA)+2) & msk) >> count;
	*(((uint32*)pbD)+3) = (*(((uint32*)pbA)+3) & msk) >> count;
}									



//
//	Packed Shift Right Logical 16x8bit by 1
//
void vmp_psrl1BGeneric( uint8 * const pbD, const uint8 * const pbA )
{
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32*)pbD)+0) = (*(((uint32*)pbA)+0) & 0xfefefefe) >> 1;
	*(((uint32*)pbD)+1) = (*(((uint32*)pbA)+1) & 0xfefefefe) >> 1;
	*(((uint32*)pbD)+2) = (*(((uint32*)pbA)+2) & 0xfefefefe) >> 1;
	*(((uint32*)pbD)+3) = (*(((uint32*)pbA)+3) & 0xfefefefe) >> 1;
}



//
//	Packed Shift Right Logical 8x16bit by n:{0...15}
//
void vmp_psrlHGeneric( uint16 * const phD, const uint16 * const phA, uint count )
{
	uint32 msk;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	count &= (16-1);				// Clip count to data bit size -1

	msk = llMaskHD[count];
	*(((uint32*)phD)+0) = (*(((uint32*)phA)+0) & msk) >> count;
	*(((uint32*)phD)+1) = (*(((uint32*)phA)+1) & msk) >> count;
	*(((uint32*)phD)+2) = (*(((uint32*)phA)+2) & msk) >> count;
	*(((uint32*)phD)+3) = (*(((uint32*)phA)+3) & msk) >> count;
}									



//
//	Packed Shift Right Logical 8x16bit by 1
//
void vmp_psrl1HGeneric( uint16 * const phD, const uint16 * const phA )
{
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	*(((uint32*)phD)+0) = (*(((uint32*)phA)+0) & 0xfffefffe) >> 1;
	*(((uint32*)phD)+1) = (*(((uint32*)phA)+1) & 0xfffefffe) >> 1;
	*(((uint32*)phD)+2) = (*(((uint32*)phA)+2) & 0xfffefffe) >> 1;
	*(((uint32*)phD)+3) = (*(((uint32*)phA)+3) & 0xfffefffe) >> 1;
}



//
//	Packed Shift Right Logical 4x32bit by n:{0...31}
//
void vmp_psrlWGeneric( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(pwD+0) = (*(pwA+0)) >> count;
	*(pwD+1) = (*(pwA+1)) >> count;
	*(pwD+2) = (*(pwA+2)) >> count;
	*(pwD+3) = (*(pwA+3)) >> count;
}									



//
//	Packed Shift Right Logical 4x32bit by 1
//
void vmp_psrl1WGeneric( uint32 * const pwD, const uint32 * const pwA )
{
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(pwD+0) = *(pwA+0) >> 1;
	*(pwD+1) = *(pwA+1) >> 1;
	*(pwD+2) = *(pwA+2) >> 1;
	*(pwD+3) = *(pwA+3) >> 1;
}



//______________________________________________________

//
//	Packed Rotate Left 16x8bit by n:{0...7}
//
void vmp_prolBGeneric( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	uint iCnt;
	uint32 rmsk, lmsk;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	count &= (8-1);				// Clip count to data bit size -1
	if ( count )
	  {
		iCnt = 8-count;

		rmsk = rrMaskBD[count];		// 7f
		lmsk = llMaskBD[iCnt];		// 80

		*(((uint32*)pbD)+0) = ((*(((uint32*)pbA)+0) & rmsk) << count)
							| ((*(((uint32*)pbA)+0) & lmsk) >> iCnt);
		*(((uint32*)pbD)+1) = ((*(((uint32*)pbA)+1) & rmsk) << count)
							| ((*(((uint32*)pbA)+1) & lmsk) >> iCnt);
		*(((uint32*)pbD)+2) = ((*(((uint32*)pbA)+2) & rmsk) << count)
							| ((*(((uint32*)pbA)+2) & lmsk) >> iCnt);
		*(((uint32*)pbD)+3) = ((*(((uint32*)pbA)+3) & rmsk) << count)
							| ((*(((uint32*)pbA)+3) & lmsk) >> iCnt);
	  }
	else
	  {
		*(((uint32*)pbD)+0) = *(((uint32*)pbA)+0);
		*(((uint32*)pbD)+1) = *(((uint32*)pbA)+1);
		*(((uint32*)pbD)+2) = *(((uint32*)pbA)+2);
		*(((uint32*)pbD)+3) = *(((uint32*)pbA)+3);
	  }
}									



//
//	Packed Rotate Left 16x8bit by 1
//
void vmp_prol1BGeneric( uint8 * const pbD, const uint8 * const pbA )
{
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32*)pbD)+0) = ((*(((uint32*)pbA)+0) & 0x7f7f7f7f) << 1 )
						| ((*(((uint32*)pbA)+0) & 0x80808080) >> 7 );
	*(((uint32*)pbD)+1) = ((*(((uint32*)pbA)+1) & 0x7f7f7f7f) << 1 )
						| ((*(((uint32*)pbA)+1) & 0x80808080) >> 7 );
	*(((uint32*)pbD)+2) = ((*(((uint32*)pbA)+2) & 0x7f7f7f7f) << 1 )
						| ((*(((uint32*)pbA)+2) & 0x80808080) >> 7 );
	*(((uint32*)pbD)+3) = ((*(((uint32*)pbA)+3) & 0x7f7f7f7f) << 1 )
						| ((*(((uint32*)pbA)+3) & 0x80808080) >> 7 );
}



//
//	Packed Rotate Left 8x16bit by n:{0...15}
//
void vmp_prolHGeneric( uint16 * const phD, const uint16 * const phA, uint count )
{
	uint iCnt;
	uint32 rmsk, lmsk;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	count &= (16-1);				// Clip count to data bit size -1
	if ( count )
	  {
		iCnt = 16-count;

		rmsk = rrMaskHD[count];		// 7fff
		lmsk = llMaskHD[iCnt];		// 8000

		*(((uint32*)phD)+0) = ((*(((uint32*)phA)+0) & rmsk) << count)
							| ((*(((uint32*)phA)+0) & lmsk) >> iCnt);
		*(((uint32*)phD)+1) = ((*(((uint32*)phA)+1) & rmsk) << count)
							| ((*(((uint32*)phA)+1) & lmsk) >> iCnt);
		*(((uint32*)phD)+2) = ((*(((uint32*)phA)+2) & rmsk) << count)
							| ((*(((uint32*)phA)+2) & lmsk) >> iCnt);
		*(((uint32*)phD)+3) = ((*(((uint32*)phA)+3) & rmsk) << count)
							| ((*(((uint32*)phA)+3) & lmsk) >> iCnt);
	  }
	else
	  {
		*(((uint32*)phD)+0) = *(((uint32*)phA)+0);
		*(((uint32*)phD)+1) = *(((uint32*)phA)+1);
		*(((uint32*)phD)+2) = *(((uint32*)phA)+2);
		*(((uint32*)phD)+3) = *(((uint32*)phA)+3);
	  }
}									



//
//	Packed Rotate Left 8x16bit by 1
//
void vmp_prol1HGeneric( uint16 * const phD, const uint16 * const phA )
{
	uint32 val0, val1, val2, val3;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	val0 = *(((uint32*)phA)+0);
	val1 = *(((uint32*)phA)+1);
	val2 = *(((uint32*)phA)+2);
	val3 = *(((uint32*)phA)+3);

	*(((uint32*)phD)+0) = ((val0 & 0x7fff7fff) << 1 )
						| ((val0 & 0x80008000) >> 15);
	*(((uint32*)phD)+1) = ((val1 & 0x7fff7fff) << 1 )
						| ((val1 & 0x80008000) >> 15);
	*(((uint32*)phD)+2) = ((val2 & 0x7fff7fff) << 1 )
						| ((val2 & 0x80008000) >> 15);
	*(((uint32*)phD)+3) = ((val3 & 0x7fff7fff) << 1 )
						| ((val3 & 0x80008000) >> 15);
}



//
//	Packed Rotate Left 4x32bit by n:{0...31}
//
void vmp_prolWGeneric( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	uint iCnt;

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	if (count)
	  {
		iCnt = 32-count;

		*(pwD+0) = (*(pwA+0) >> iCnt) | (*(pwA+0) << count);
		*(pwD+1) = (*(pwA+1) >> iCnt) | (*(pwA+1) << count);
		*(pwD+2) = (*(pwA+2) >> iCnt) | (*(pwA+2) << count);
		*(pwD+3) = (*(pwA+3) >> iCnt) | (*(pwA+3) << count);
	  }
	else
	  {
		*(pwD+0) = *(pwA+0);
		*(pwD+1) = *(pwA+1);
		*(pwD+2) = *(pwA+2);
		*(pwD+3) = *(pwA+3);
	  }
}									



//
//	Packed Rotate Left 4x32bit by 1
//
void vmp_prol1WGeneric( uint32 * const pwD, const uint32 * const pwA )
{
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(pwD+0) = (*(pwA+0) >> 31) | (*(pwA+0) << 1);
	*(pwD+1) = (*(pwA+1) >> 31) | (*(pwA+1) << 1);
	*(pwD+2) = (*(pwA+2) >> 31) | (*(pwA+2) << 1);
	*(pwD+3) = (*(pwA+3) >> 31) | (*(pwA+3) << 1);
}







