
//
//  PRotAltivec.cpp - Packed Rotation 
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      19 Dec 2001 - JCL - Brought up to date.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "PRot.h"
#include "PRotEx.h"


//
// 8bit Shift values (vector arranged)
//

static vector unsigned char V8Shift[] = {
	{ 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0 },
	{ 1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1 },
	{ 2,2,2,2,2,2,2,2, 2,2,2,2,2,2,2,2 },
	{ 3,3,3,3,3,3,3,3, 3,3,3,3,3,3,3,3 },
	{ 4,4,4,4,4,4,4,4, 4,4,4,4,4,4,4,4 },
	{ 5,5,5,5,5,5,5,5, 5,5,5,5,5,5,5,5 },
	{ 6,6,6,6,6,6,6,6, 6,6,6,6,6,6,6,6 },
	{ 7,7,7,7,7,7,7,7, 7,7,7,7,7,7,7,7 }};


static vector unsigned short V16Shift[] = {
	{ 0,0,0,0,0,0,0,0 },
	{ 1,1,1,1,1,1,1,1 },
	{ 2,2,2,2,2,2,2,2 },
	{ 3,3,3,3,3,3,3,3 },
	{ 4,4,4,4,4,4,4,4 },
	{ 5,5,5,5,5,5,5,5 },
	{ 6,6,6,6,6,6,6,6 },
	{ 7,7,7,7,7,7,7,7 },
	{ 8,8,8,8,8,8,8,8 },
	{ 9,9,9,9,9,9,9,9 },
	{ 10,10,10,10,10,10,10,10 },
	{ 11,11,11,11,11,11,11,11 },
	{ 12,12,12,12,12,12,12,12 },
	{ 13,13,13,13,13,13,13,13 },
	{ 14,14,14,14,14,14,14,14 },
	{ 15,15,15,15,15,15,15,15 }};
	

static vector unsigned int V32Shift[] = {
	{ 0,0,0,0 },     { 1,1,1,1 },     { 2,2,2,2 },     { 3,3,3,3 },
	{ 4,4,4,4 },     { 5,5,5,5 },     { 6,6,6,6 },     { 7,7,7,7 },
	{ 8,8,8,8 },     { 9,9,9,9 },     { 10,10,10,10 }, { 11,11,11,11 },
	{ 12,12,12,12 }, { 13,13,13,13 }, { 14,14,14,14 }, { 15,15,15,15 },
	{ 16,16,16,16 }, { 17,17,17,17 }, { 18,18,18,18 }, { 19,19,19,19 },
	{ 20,20,20,20 }, { 21,21,21,21 }, { 22,22,22,22 }, { 23,23,23,23 },
	{ 24,24,24,24 }, { 25,25,25,25 }, { 26,26,26,26 }, { 27,27,27,27 },
	{ 28,28,28,28 }, { 29,29,29,29 }, { 30,30,30,30 }, { 31,31,31,31 }};

		

//______________________________________________________
//
//	Packed Shift Right Arithmetic 16x8bit by n:{0...7}  {Unaligned}
//
void vmp_psraBAltivec0( int8 * const pbD, const int8 * const pbA, uint count )
{
	vector signed char vbD, vbA;
	
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32 *)&vbA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vbA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vbA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vbA)+3) = *(((uint32 *)pbA)+3);
	
	count &= (8-1);				// Clip count to data bit size -1

	vbD = vec_sra( vbA, V8Shift[count] );				// vsrab d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vbD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vbD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vbD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vbD)+3);
}								



//
//	Packed Shift Right Arithmetic 16x8bit by n:{0...7}  {Aligned}
//
void vmp_psraBAltivec( int8 * const pbD, const int8 * const pbA, uint count )
{
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbD);

	count &= (8-1);				// Clip count to data bit size -1

	*(vector signed char *)pbD = vec_sra( (*(vector signed char *)pbA), V8Shift[count] );		// vsrab d,a,b
}								



//
//	Packed Shift Right Arithmetic 16x8bit by 1  {Unaligned}
//
void vmp_psra1BAltivec0( int8 * const pbD, const int8 * const pbA )
{
	vmp_psraBAltivec0( pbD, pbA, 1 );
}



//
//	Packed Shift Right Arithmetic 16x8bit by 1  {Aligned}
//
void vmp_psra1BAltivec( int8 * const pbD, const int8 * const pbA )
{
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbD);

	*(vector signed char *)pbD = vec_sra( (*(vector signed char *)pbA), V8Shift[1] );			// vsrab d,a,b
}



//______________________________________________________


//
//	Packed Shift Right Arithmetic 8x16bit by n:{0...15}  {Unaligned}
//
void vmp_psraHAltivec0( int16 * const phD, const int16 * const phA, uint count )
{
	vector signed short vhD, vhA;
	
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	*(((uint32 *)&vhA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vhA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vhA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vhA)+3) = *(((uint32 *)phA)+3);
	
	count &= (16-1);				// Clip count to data bit size -1

	vhD = vec_sra( vhA, V16Shift[count] );	// vsrah d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vhD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vhD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vhD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vhD)+3);
}


//
//	Packed Shift Right Arithmetic 8x16bit by n:{0...15}  {Aligned}
//
void vmp_psraHAltivec( int16 * const phD, const int16 * const phA, uint count )
{
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phD);

	count &= (16-1);				// Clip count to data bit size -1

	*(vector signed short *)phD = vec_sra( (*(vector signed short *)phA), V16Shift[count] );	// vsrah d,a,b
}



//
//	Packed Shift Right Arithmetic 8x16bit by 1  {Unaligned}
//
void vmp_psra1HAltivec0( int16 * const phD, const int16 * const phA )
{
	vmp_psraHAltivec0( phD, phA, 1 );
}



//
//	Packed Shift Right Arithmetic 8x16bit by 1  {Aligned}
//
void vmp_psra1HAltivec( int16 * const phD, const int16 * const phA )
{
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phD);

	*(vector signed short *)phD = vec_sra( (*(vector signed short *)phA), V16Shift[1] );		// vsrah d,a,b
}



//______________________________________________________



//
//	Packed Shift Right Arithmetic 4x32bit by n:{0...31}  {Unaligned}
//
void vmp_psraWAltivec0( int32 * const pwD, const int32 * const pwA, uint count )
{
	vector signed int vwD, vwA;
	
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(((uint32 *)&vwA)+0) = *(((uint32 *)pwA)+0);
	*(((uint32 *)&vwA)+1) = *(((uint32 *)pwA)+1);
	*(((uint32 *)&vwA)+2) = *(((uint32 *)pwA)+2);
	*(((uint32 *)&vwA)+3) = *(((uint32 *)pwA)+3);
	
	count &= (32-1);				// Clip count to data bit size -1

	vwD = vec_sra( vwA, V32Shift[count] );		// vsraw d,a,b

	*(((uint32 *)pwD)+0) = *(((uint32 *)&vwD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&vwD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&vwD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&vwD)+3);
}									



//
//	Packed Shift Right Arithmetic 4x32bit by n:{0...31}  {Aligned}
//
void vmp_psraWAltivec( int32 * const pwD, const int32 * const pwA, uint count )
{
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(vector signed int *)pwD = vec_sra( (*(vector signed int *)pwA), V32Shift[count] );		// vsraw d,a,b
}									



//
//	Packed Shift Right Arithmetic 4x32bit by 1  {Unaligned}
//
void vmp_psra1WAltivec0( int32 * const pwD, const int32 * const pwA )
{
	vmp_psraWAltivec0( pwD, pwA, 1 );
}



//
//	Packed Shift Right Arithmetic 4x32bit by 1  {Aligned}
//
void vmp_psra1WAltivec( int32 * const pwD, const int32 * const pwA )
{
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwD);

	*(vector signed int *)pwD = vec_sra( (*(vector signed int *)pwA), V32Shift[1] );			// vsraw d,a,b
}


//______________________________________________________
//------------------------------------------------------


//
//	Packed Shift Left Logical 16x8bit by n:{0...7} {Unaligned}
//
void vmp_psllBAltivec0( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	vector unsigned char vbD, vbA;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32 *)&vbA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vbA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vbA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vbA)+3) = *(((uint32 *)pbA)+3);
	
	count &= (8-1);				// Clip count to data bit size -1

	vbD = vec_sl( vbA, V8Shift[count] );		// vslb d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vbD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vbD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vbD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vbD)+3);
}									



//
//	Packed Shift Left Logical 16x8bit by n:{0...7} {Aligned}
//
void vmp_psllBAltivec( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbD);

	count &= (8-1);				// Clip count to data bit size -1

	*(vector unsigned char *)pbD = vec_sl( (*(vector unsigned char *)pbA), V8Shift[count] );		// vslb d,a,b
}									



//
//	Packed Shift Left Logical 16x8bit by 1 {Unaligned}
//
void vmp_psll1BAltivec0( uint8 * const pbD, const uint8 * const pbA )
{
	vmp_psllBAltivec0( pbD, pbA, 1 );
}



//
//	Packed Shift Left Logical 16x8bit by 1 {Aligned}
//
void vmp_psll1BAltivec( uint8 * const pbD, const uint8 * const pbA )
{
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbD);

	*(vector unsigned char *)pbD = vec_sl( (*(vector unsigned char *)pbA), V8Shift[1] );			// vslb d,a,b
}


//______________________________________________________


//
//	Packed Shift Left Logical 8x16bit by n:{0...15}  {Unaligned}
//
void vmp_psllHAltivec0( uint16 * const phD, const uint16 * const phA, uint count )
{
	vector unsigned short vhD, vhA;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	*(((uint32 *)&vhA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vhA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vhA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vhA)+3) = *(((uint32 *)phA)+3);
	
	count &= (16-1);				// Clip count to data bit size -1

	vhD = vec_sl( vhA, V16Shift[count] );		// vslh d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vhD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vhD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vhD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vhD)+3);
}									



//
//	Packed Shift Left Logical 8x16bit by n:{0...15}  {Aligned}
//
void vmp_psllHAltivec( uint16 * const phD, const uint16 * const phA, uint count )
{
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phD);

	count &= (16-1);				// Clip count to data bit size -1

	*(vector unsigned short *)phD = vec_sl( (*(vector unsigned short *)phA), V16Shift[count] );		// vslh d,a,b
}									



//
//	Packed Shift Left Logical 8x16bit by 1  {Unaligned}
//
void vmp_psll1HAltivec0( uint16 * const phD, const uint16 * const phA )
{
	vmp_psllHAltivec0( phD, phA, 1 );
}



//
//	Packed Shift Left Logical 8x16bit by 1  {Aligned}
//
void vmp_psll1HAltivec( uint16 * const phD, const uint16 * const phA )
{
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phD);

	*(vector unsigned short *)phD = vec_sl( (*(vector unsigned short *)phA), V16Shift[1] );			// vslh d,a,b
}


//______________________________________________________


//
//	Packed Shift Left Logical 4x32bit by n:{0...31}  {Unaligned}
//
void vmp_psllWAltivec0( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	vector unsigned int vwD, vwA;

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(((uint32 *)&vwA)+0) = *(((uint32 *)pwA)+0);
	*(((uint32 *)&vwA)+1) = *(((uint32 *)pwA)+1);
	*(((uint32 *)&vwA)+2) = *(((uint32 *)pwA)+2);
	*(((uint32 *)&vwA)+3) = *(((uint32 *)pwA)+3);
	
	count &= (32-1);				// Clip count to data bit size -1

	vwD = vec_sl( vwA, V32Shift[count] );			// vslw d,a,b

	*(((uint32 *)pwD)+0) = *(((uint32 *)&vwD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&vwD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&vwD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&vwD)+3);
}									



//
//	Packed Shift Left Logical 4x32bit by n:{0...31}  {Aligned}
//
void vmp_psllWAltivec( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(vector unsigned int *)pwD = vec_sl( (*(vector unsigned int *)pwA), V32Shift[count] );			// vslw d,a,b
}									



//
//	Packed Shift Left Logical 4x32bit by 1  {Unaligned}
//
void vmp_psll1WAltivec0( uint32 * const pwD, const uint32 * const pwA )
{
	vmp_psllWAltivec0( pwD, pwA, 1 );
}



//
//	Packed Shift Left Logical 4x32bit by 1  {Aligned}
//
void vmp_psll1WAltivec( uint32 * const pwD, const uint32 * const pwA )
{
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwD);

	*(vector unsigned int *)pwD = vec_sl( (*(vector unsigned int *)pwA), V32Shift[1] );				// vslw d,a,b
}


//______________________________________________________
//------------------------------------------------------


//
//	Packed Shift Right Logical 16x8bit by n:{0...7} {Unaligned}
//
void vmp_psrlBAltivec0( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	vector unsigned char vbD, vbA;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32 *)&vbA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vbA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vbA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vbA)+3) = *(((uint32 *)pbA)+3);
	
	count &= (8-1);				// Clip count to data bit size -1

	vbD = vec_sr( vbA, V8Shift[count] );		// vsrb d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vbD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vbD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vbD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vbD)+3);
}									



//
//	Packed Shift Right Logical 16x8bit by n:{0...7} {Aligned}
//
void vmp_psrlBAltivec( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbD);

	count &= (8-1);				// Clip count to data bit size -1

	*(vector unsigned char *)pbD = vec_sr( (*(vector unsigned char *)pbA), V8Shift[count] );		// vsrb d,a,b
}									



//
//	Packed Shift Right Logical 16x8bit by 1 {Unaligned}
//
void vmp_psrl1BAltivec0( uint8 * const pbD, const uint8 * const pbA )
{
	vmp_psrlBAltivec0( pbD, pbA, 1 );
}



//
//	Packed Shift Right Logical 16x8bit by 1 {Aligned}
//
void vmp_psrl1BAltivec( uint8 * const pbD, const uint8 * const pbA )
{
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbD);

	*(vector unsigned char *)pbD = vec_sr( (*(vector unsigned char *)pbA), V8Shift[1] );			// vsrb d,a,b
}


//______________________________________________________


//
//	Packed Shift Right Logical 8x16bit by n:{0...15}  {Unaligned}
//
void vmp_psrlHAltivec0( uint16 * const phD, const uint16 * const phA, uint count )
{
	vector unsigned short vhD, vhA;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	*(((uint32 *)&vhA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vhA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vhA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vhA)+3) = *(((uint32 *)phA)+3);
	
	count &= (16-1);				// Clip count to data bit size -1

	vhD = vec_sr( vhA, V16Shift[count] );		// vsrh d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vhD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vhD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vhD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vhD)+3);
}									



//
//	Packed Shift Right Logical 8x16bit by n:{0...15}  {Aligned}
//
void vmp_psrlHAltivec( uint16 * const phD, const uint16 * const phA, uint count )
{
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phD);

	count &= (16-1);				// Clip count to data bit size -1

	*(vector unsigned short *)phD = vec_sr( (*(vector unsigned short *)phA), V16Shift[count] );		// vsrh d,a,b
}									



//
//	Packed Shift Right Logical 8x16bit by 1  {Unaligned}
//
void vmp_psrl1HAltivec0( uint16 * const phD, const uint16 * const phA )
{
	vmp_psrlHAltivec0( phD, phA, 1 );
}



//
//	Packed Shift Right Logical 8x16bit by 1  {Aligned}
//
void vmp_psrl1HAltivec( uint16 * const phD, const uint16 * const phA )
{
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phD);

	*(vector unsigned short *)phD = vec_sr( (*(vector unsigned short *)phA), V16Shift[1] );			// vsrh d,a,b
}


//______________________________________________________


//
//	Packed Shift Right Logical 4x32bit by n:{0...31}  {Unaligned}
//
void vmp_psrlWAltivec0( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	vector unsigned int vwD, vwA;

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(((uint32 *)&vwA)+0) = *(((uint32 *)pwA)+0);
	*(((uint32 *)&vwA)+1) = *(((uint32 *)pwA)+1);
	*(((uint32 *)&vwA)+2) = *(((uint32 *)pwA)+2);
	*(((uint32 *)&vwA)+3) = *(((uint32 *)pwA)+3);
	
	count &= (32-1);				// Clip count to data bit size -1

	vwD = vec_sr( vwA, V32Shift[count] );			// vsrw d,a,b

	*(((uint32 *)pwD)+0) = *(((uint32 *)&vwD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&vwD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&vwD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&vwD)+3);
}									



//
//	Packed Shift Right Logical 4x32bit by n:{0...31}  {Aligned}
//
void vmp_psrlWAltivec( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(vector unsigned int *)pwD = vec_sr( (*(vector unsigned int *)pwA), V32Shift[count] );			// vsrw d,a,b
}									



//
//	Packed Shift Right Logical 4x32bit by 1  {Unaligned}
//
void vmp_psrl1WAltivec0( uint32 * const pwD, const uint32 * const pwA )
{
	vmp_psrlWAltivec0( pwD, pwA, 1 );
}



//
//	Packed Shift Right Logical 4x32bit by 1  {Aligned}
//
void vmp_psrl1WAltivec( uint32 * const pwD, const uint32 * const pwA )
{
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwD);

	*(vector unsigned int *)pwD = vec_sr( (*(vector unsigned int *)pwA), V32Shift[1] );				// vsrw d,a,b
}


//______________________________________________________
//------------------------------------------------------



//
//	Packed Rotate Left Logical 16x8bit by n:{0...7} {Unaligned}
//
void vmp_prolBAltivec0( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	vector unsigned char vbD, vbA;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32 *)&vbA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vbA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vbA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vbA)+3) = *(((uint32 *)pbA)+3);
	
	count &= (8-1);				// Clip count to data bit size -1

	vbD = vec_rl( vbA, V8Shift[count] );		// vrlb d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vbD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vbD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vbD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vbD)+3);
}									



//
//	Packed Rotate Left Logical 16x8bit by n:{0...7} {Aligned}
//
void vmp_prolBAltivec( uint8 * const pbD, const uint8 * const pbA, uint count )
{
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbD);

	count &= (8-1);				// Clip count to data bit size -1

	*(vector unsigned char *)pbD = vec_rl( (*(vector unsigned char *)pbA), V8Shift[count] );		// vrlb d,a,b
}									



//
//	Packed Rotate Left Logical 16x8bit by 1 {Unaligned}
//
void vmp_prol1BAltivec0( uint8 * const pbD, const uint8 * const pbA )
{
	vmp_prolBAltivec0( pbD, pbA, 1 );
}



//
//	Packed Rotate Left Logical 16x8bit by 1 {Aligned}
//
void vmp_prol1BAltivec( uint8 * const pbD, const uint8 * const pbA )
{
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbD);

	*(vector unsigned char *)pbD = vec_rl( (*(vector unsigned char *)pbA), V8Shift[1] );			// vrlb d,a,b
}


//______________________________________________________


//
//	Packed Rotate Left Logical 8x16bit by n:{0...15}  {Unaligned}
//
void vmp_prolHAltivec0( uint16 * const phD, const uint16 * const phA, uint count )
{
	vector unsigned short vhD, vhA;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	*(((uint32 *)&vhA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vhA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vhA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vhA)+3) = *(((uint32 *)phA)+3);
	
	count &= (16-1);				// Clip count to data bit size -1

	vhD = vec_rl( vhA, V16Shift[count] );		// vrlh d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vhD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vhD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vhD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vhD)+3);
}									



//
//	Packed Rotate Left Logical 8x16bit by n:{0...15}  {Aligned}
//
void vmp_prolHAltivec( uint16 * const phD, const uint16 * const phA, uint count )
{
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phD);

	count &= (16-1);				// Clip count to data bit size -1

	*(vector unsigned short *)phD = vec_rl( (*(vector unsigned short *)phA), V16Shift[count] );		// vrlh d,a,b
}									



//
//	Packed Rotate Left Logical 8x16bit by 1  {Unaligned}
//
void vmp_prol1HAltivec0( uint16 * const phD, const uint16 * const phA )
{
	vmp_prolHAltivec0( phD, phA, 1 );
}


//
//	Packed Rotate Left Logical 8x16bit by 1  {Aligned}
//
void vmp_prol1HAltivec( uint16 * const phD, const uint16 * const phA )
{
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phD);

	*(vector unsigned short *)phD = vec_rl( (*(vector unsigned short *)phA), V16Shift[1] );			// vrlh d,a,b
}


//______________________________________________________


//
//	Packed Rotate Left Logical 4x32bit by n:{0...31}  {Unaligned}
//
void vmp_prolWAltivec0( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	vector unsigned int vwD, vwA;

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(((uint32 *)&vwA)+0) = *(((uint32 *)pwA)+0);
	*(((uint32 *)&vwA)+1) = *(((uint32 *)pwA)+1);
	*(((uint32 *)&vwA)+2) = *(((uint32 *)pwA)+2);
	*(((uint32 *)&vwA)+3) = *(((uint32 *)pwA)+3);
	
	count &= (32-1);				// Clip count to data bit size -1

	vwD = vec_rl( vwA, V32Shift[count] );			// vrlw d,a,b

	*(((uint32 *)pwD)+0) = *(((uint32 *)&vwD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&vwD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&vwD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&vwD)+3);
}									



//
//	Packed Rotate Left Logical 4x32bit by n:{0...31}  {Aligned}
//
void vmp_prolWAltivec( uint32 * const pwD, const uint32 * const pwA, uint count )
{
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(vector unsigned int *)pwD = vec_rl( (*(vector unsigned int *)pwA), V32Shift[count] );			// vrlw d,a,b
}									



//
//	Packed Rotate Left Logical 4x32bit by 1  {Unaligned}
//
void vmp_prol1WAltivec0( uint32 * const pwD, const uint32 * const pwA )
{
	vmp_prolWAltivec0( pwD, pwA, 1 );
}



//
//	Packed Rotate Left Logical 4x32bit by 1  {Aligned}
//
void vmp_prol1WAltivec( uint32 * const pwD, const uint32 * const pwA )
{
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwD);

	*(vector unsigned int *)pwD = vec_rl( (*(vector unsigned int *)pwA), V32Shift[1] );				// vrlw d,a,b
}


//______________________________________________________


//
//	Setup Vec3D module
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void PRotSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);
	    
	if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_psraB =				vmp_psraBAltivec;
		vmp_psraB0 =			vmp_psraBAltivec0;
		vmp_psraH =				vmp_psraHAltivec;
		vmp_psraH0 =			vmp_psraHAltivec0;
		vmp_psraW =				vmp_psraWAltivec;
		vmp_psraW0 =			vmp_psraWAltivec0;

		vmp_psra1B =			vmp_psra1BAltivec;
		vmp_psra1B0 =			vmp_psra1BAltivec0;
		vmp_psra1H =			vmp_psra1HAltivec;
		vmp_psra1H0 =			vmp_psra1HAltivec0;
		vmp_psra1W =			vmp_psra1WAltivec;
		vmp_psra1W0 =			vmp_psra1WAltivec0;

		vmp_psllB =				vmp_psllBAltivec;
		vmp_psllB0 =			vmp_psllBAltivec0;
		vmp_psllH =				vmp_psllHAltivec;
		vmp_psllH0 =			vmp_psllHAltivec0;
		vmp_psllW =				vmp_psllWAltivec;
		vmp_psllW0 =			vmp_psllWAltivec0;

		vmp_psll1B =			vmp_psll1BAltivec;
		vmp_psll1B0 =			vmp_psll1BAltivec0;
		vmp_psll1H =			vmp_psll1HAltivec;
		vmp_psll1H0 =			vmp_psll1HAltivec0;
		vmp_psll1W =			vmp_psll1WAltivec;
		vmp_psll1W0 =			vmp_psll1WAltivec0;

		vmp_psrlB =				vmp_psrlBAltivec;
		vmp_psrlB0 =			vmp_psrlBAltivec0;
		vmp_psrlH =				vmp_psrlHAltivec;
		vmp_psrlH0 =			vmp_psrlHAltivec0;
		vmp_psrlW =				vmp_psrlWAltivec;
		vmp_psrlW0 =			vmp_psrlWAltivec0;

		vmp_psrl1B =			vmp_psrl1BAltivec;
		vmp_psrl1B0 =			vmp_psrl1BAltivec0;
		vmp_psrl1H =			vmp_psrl1HAltivec;
		vmp_psrl1H0 =			vmp_psrl1HAltivec0;
		vmp_psrl1W =			vmp_psrl1WAltivec;
		vmp_psrl1W0 =			vmp_psrl1WAltivec0;

		vmp_prolB =				vmp_prolBAltivec;
		vmp_prolB0 =			vmp_prolBAltivec0;
		vmp_prolH =				vmp_prolHAltivec;
		vmp_prolH0 =			vmp_prolHAltivec0;
		vmp_prolW =				vmp_prolWAltivec;
		vmp_prolW0 =			vmp_prolWAltivec0;

		vmp_prol1B =			vmp_prol1BAltivec;
		vmp_prol1B0 =			vmp_prol1BAltivec0;
		vmp_prol1H =			vmp_prol1HAltivec;
		vmp_prol1H0 =			vmp_prol1HAltivec0;
		vmp_prol1W =			vmp_prol1WAltivec;
		vmp_prol1W0 =			vmp_prol1WAltivec0;
      }

	else
	  {
		vmp_psraB =				vmp_psraBPPC;
		vmp_psraB0 =			vmp_psraBPPC;
		vmp_psraH =				vmp_psraHPPC;
		vmp_psraH0 =			vmp_psraHPPC;
		vmp_psraW =				vmp_psraWPPC;
		vmp_psraW0 =			vmp_psraWPPC;

		vmp_psra1B =			vmp_psra1BPPC;
		vmp_psra1B0 =			vmp_psra1BPPC;
		vmp_psra1H =			vmp_psra1HPPC;
		vmp_psra1H0 =			vmp_psra1HPPC;
		vmp_psra1W =			vmp_psra1WPPC;
		vmp_psra1W0 =			vmp_psra1WPPC;

		vmp_psllB =				vmp_psllBPPC;
		vmp_psllB0 =			vmp_psllBPPC;
		vmp_psllH =				vmp_psllHPPC;
		vmp_psllH0 =			vmp_psllHPPC;
		vmp_psllW =				vmp_psllWPPC;
		vmp_psllW0 =			vmp_psllWPPC;

		vmp_psll1B =			vmp_psll1BPPC;
		vmp_psll1B0 =			vmp_psll1BPPC;
		vmp_psll1H =			vmp_psll1HPPC;
		vmp_psll1H0 =			vmp_psll1HPPC;
		vmp_psll1W =			vmp_psll1WPPC;
		vmp_psll1W0 =			vmp_psll1WPPC;
		
		vmp_psrlB =				vmp_psrlBPPC;
		vmp_psrlB0 =			vmp_psrlBPPC;
		vmp_psrlH =				vmp_psrlHPPC;
		vmp_psrlH0 =			vmp_psrlHPPC;
		vmp_psrlW =				vmp_psrlWPPC;
		vmp_psrlW0 =			vmp_psrlWPPC;

		vmp_psrl1B =			vmp_psrl1BPPC;
		vmp_psrl1B0 =			vmp_psrl1BPPC;
		vmp_psrl1H =			vmp_psrl1HPPC;
		vmp_psrl1H0 =			vmp_psrl1HPPC;
		vmp_psrl1W =			vmp_psrl1WPPC;
		vmp_psrl1W0 =			vmp_psrl1WPPC;

		vmp_prolB =				vmp_prolBPPC;
		vmp_prolB0 =			vmp_prolBPPC;
		vmp_prolH =				vmp_prolHPPC;
		vmp_prolH0 =			vmp_prolHPPC;
		vmp_prolW =				vmp_prolWPPC;
		vmp_prolW0 =			vmp_prolWPPC;

		vmp_prol1B =			vmp_prol1BPPC;
		vmp_prol1B0 =			vmp_prol1BPPC;
		vmp_prol1H =			vmp_prol1HPPC;
		vmp_prol1H0 =			vmp_prol1HPPC;
		vmp_prol1W =			vmp_prol1WPPC;
		vmp_prol1W0 =			vmp_prol1WPPC;
	  }
}




