
//
//  prolPPC.cpp - Packed Rotate Left (PowerPC) 
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      19 Dec 2001 - JCL - Brought up to date.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "PRot.h"
#include "PRotEx.h"


extern uint32 llMaskBD[];
extern uint32 rrMaskBD[];
extern uint32 llMaskHD[];
extern uint32 rrMaskHD[];



//
//	Packed Rotate Left 16x8bit by n:{0...7}
//
void vmp_prolBPPC( uint8 * const pbD, const uint8 * const pbA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _r27, _r26, _r25, _r24;
	unsigned int register _lMsk, _rMsk, _iCnt;

#ifdef USE_ASSERT
#pragma unused(count)
#else
#pragma unused(pbD, pbA, count)
#endif

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	__asm {
		li		_iCnt,8
		clrlwi	r5,r5,29			// Clear upper 29 bits so value is clipped at (0...7)
		lwz		_r27,rrMaskBD
		lwz		_r26,llMaskBD+(8*4)

		slwi	_lMsk,r5,2			// x4 32bit table ref.
		sub		_iCnt,_iCnt,r5		// iCnt=8-count

		add		_rMsk,_r27,_lMsk	// +(count*4)
		sub		_lMsk,_r26,_lMsk	// -(count*4)
		
		lwz		_rMsk,0(_rMsk)		// right Mask
		lwz		_lMsk,0(_lMsk)		// left mask

		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		and		_r27,_r31,_rMsk		// (*(((uint32*)phA)+0) & rmsk)
		and		_r31,_r31,_lMsk		// (*(((uint32*)phA)+0) & lmsk)
		and		_r26,_r30,_rMsk
		and		_r30,_r30,_lMsk
		and		_r25,_r29,_rMsk
		and		_r29,_r29,_lMsk
		and 	_r24,_r28,_rMsk
		and		_r28,_r28,_lMsk

		srw		_r31,_r31,_iCnt		// ((*(((uint32*)phA)+0) & lmsk) >> iCnt)
		slw		_r27,_r27,r5		// ((*(((uint32*)phA)+0) & rmsk) << count)
		srw		_r30,_r30,_iCnt
		slw		_r26,_r26,r5
		srw		_r29,_r29,_iCnt
		slw		_r25,_r25,r5
		srw		_r28,_r28,_iCnt
		slw		_r24,_r24,r5

		or		_r31,_r31,_r27		// Blend bits
		or		_r30,_r30,_r26
		or		_r29,_r29,_r25
		or		_r28,_r28,_r24
			
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	uint iCnt;
	uint32 rmsk, lmsk;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	count &= (8-1);				// Clip count to data bit size -1
	if ( count )
	  {
		iCnt = 8-count;

		rmsk = rrMaskBD[count];		// 7f
		lmsk = llMaskBD[iCnt];		// 80

		*(((uint32*)pbD)+0) = ((*(((uint32*)pbA)+0) & rmsk) << count)
							| ((*(((uint32*)pbA)+0) & lmsk) >> iCnt);
		*(((uint32*)pbD)+1) = ((*(((uint32*)pbA)+1) & rmsk) << count)
							| ((*(((uint32*)pbA)+1) & lmsk) >> iCnt);
		*(((uint32*)pbD)+2) = ((*(((uint32*)pbA)+2) & rmsk) << count)
							| ((*(((uint32*)pbA)+2) & lmsk) >> iCnt);
		*(((uint32*)pbD)+3) = ((*(((uint32*)pbA)+3) & rmsk) << count)
							| ((*(((uint32*)pbA)+3) & lmsk) >> iCnt);
	  }
	else
	  {
		*(((uint32*)pbD)+0) = *(((uint32*)pbA)+0);
		*(((uint32*)pbD)+1) = *(((uint32*)pbA)+1);
		*(((uint32*)pbD)+2) = *(((uint32*)pbA)+2);
		*(((uint32*)pbD)+3) = *(((uint32*)pbA)+3);
	  }
#endif
}									



//
//	Packed Rotate Left 16x8bit by 1
//
void vmp_prol1BPPC( uint8 * const pbD, const uint8 * const pbA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _r27, _r26, _r25, _r24, _lMsk, _rMsk;

#ifndef USE_ASSERT
#pragma unused(pbD, pbA)
#endif

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	__asm {
		lis		_rMsk,0x7f7f		// =7F7F0000
		lis		_lMsk,0x8080		// =80800000
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		addi	_rMsk,_rMsk,0x7f7f	// =7F7F7F7F
		ori		_lMsk,_lMsk,0x8080	// =80808080
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		and		_r27,_r31,_rMsk		// (*(((uint32*)pbA)+0) & 0x7f7f7f7f)
		and		_r31,_r31,_lMsk		// (*(((uint32*)pbA)+0) & 0x80808080)
		and		_r26,_r30,_rMsk
		and		_r30,_r30,_lMsk
		and		_r25,_r29,_rMsk
		and		_r29,_r29,_lMsk
		and 	_r24,_r28,_rMsk
		and		_r28,_r28,_lMsk

		srwi	_r31,_r31,7			// ((*(((uint32*)pbA)+0) & 0x80808080) >> 7 )
		slwi	_r27,_r27,1			// ((*(((uint32*)pbA)+0) & 0x7f7f7f7f) << 1 )
		srwi	_r30,_r30,7
		slwi	_r26,_r26,1
		srwi	_r29,_r29,7
		slwi	_r25,_r25,1
		srwi	_r28,_r28,7
		slwi	_r24,_r24,1

		or		_r31,_r31,_r27		// Blend bits
		or		_r30,_r30,_r26
		or		_r29,_r29,_r25
		or		_r28,_r28,_r24
			
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
	
#else
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32*)pbD)+0) = ((*(((uint32*)pbA)+0) & 0x7f7f7f7f) << 1 )
						| ((*(((uint32*)pbA)+0) & 0x80808080) >> 7 );
	*(((uint32*)pbD)+1) = ((*(((uint32*)pbA)+1) & 0x7f7f7f7f) << 1 )
						| ((*(((uint32*)pbA)+1) & 0x80808080) >> 7 );
	*(((uint32*)pbD)+2) = ((*(((uint32*)pbA)+2) & 0x7f7f7f7f) << 1 )
						| ((*(((uint32*)pbA)+2) & 0x80808080) >> 7 );
	*(((uint32*)pbD)+3) = ((*(((uint32*)pbA)+3) & 0x7f7f7f7f) << 1 )
						| ((*(((uint32*)pbA)+3) & 0x80808080) >> 7 );
#endif
}



//
//	Packed Rotate Left 8x16bit by n:{0...15}
//
void vmp_prolHPPC( uint16 * const phD, const uint16 * const phA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _r27, _r26, _r25, _r24;
	unsigned int register _lMsk, _rMsk, _iCnt;

#ifdef USE_ASSERT
#pragma unused(count)
#else
#pragma unused(phD, phA, count)
#endif

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	__asm {
		li		_iCnt,16
		clrlwi	r5,r5,28			// Clear upper 28 bits so value is clipped at (0...15)
		lwz		_r27,rrMaskHD
		lwz		_r26,llMaskHD+(16*4)

		slwi	_lMsk,r5,2			// x4 32bit table ref.
		sub		_iCnt,_iCnt,r5		// iCnt=16-count

		add		_rMsk,_r27,_lMsk	// +(count*4)
		sub		_lMsk,_r26,_lMsk	// -(count*4)
		
		lwz		_rMsk,0(_rMsk)		// right Mask
		lwz		_lMsk,0(_lMsk)		// left mask

		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		and		_r27,_r31,_rMsk		// (*(((uint32*)phA)+0) & rmsk)
		and		_r31,_r31,_lMsk		// (*(((uint32*)phA)+0) & lmsk)
		and		_r26,_r30,_rMsk
		and		_r30,_r30,_lMsk
		and		_r25,_r29,_rMsk
		and		_r29,_r29,_lMsk
		and 	_r24,_r28,_rMsk
		and		_r28,_r28,_lMsk

		srw		_r31,_r31,_iCnt		// ((*(((uint32*)phA)+0) & lmsk) >> iCnt)
		slw		_r27,_r27,r5		// ((*(((uint32*)phA)+0) & rmsk) << count)
		srw		_r30,_r30,_iCnt
		slw		_r26,_r26,r5
		srw		_r29,_r29,_iCnt
		slw		_r25,_r25,r5
		srw		_r28,_r28,_iCnt
		slw		_r24,_r24,r5

		or		_r31,_r31,_r27		// Blend bits
		or		_r30,_r30,_r26
		or		_r29,_r29,_r25
		or		_r28,_r28,_r24
			
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}

#else
	uint iCnt;
	uint32 rmsk, lmsk;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	count &= (16-1);				// Clip count to data bit size -1
	if ( count )
	  {
		iCnt = 16-count;

		rmsk = rrMaskHD[count];		// 7fff
		lmsk = llMaskHD[iCnt];		// 8000

		*(((uint32*)phD)+0) = ((*(((uint32*)phA)+0) & rmsk) << count)
							| ((*(((uint32*)phA)+0) & lmsk) >> iCnt);
		*(((uint32*)phD)+1) = ((*(((uint32*)phA)+1) & rmsk) << count)
							| ((*(((uint32*)phA)+1) & lmsk) >> iCnt);
		*(((uint32*)phD)+2) = ((*(((uint32*)phA)+2) & rmsk) << count)
							| ((*(((uint32*)phA)+2) & lmsk) >> iCnt);
		*(((uint32*)phD)+3) = ((*(((uint32*)phA)+3) & rmsk) << count)
							| ((*(((uint32*)phA)+3) & lmsk) >> iCnt);
	  }
	else
	  {
		*(((uint32*)phD)+0) = *(((uint32*)phA)+0);
		*(((uint32*)phD)+1) = *(((uint32*)phA)+1);
		*(((uint32*)phD)+2) = *(((uint32*)phA)+2);
		*(((uint32*)phD)+3) = *(((uint32*)phA)+3);
	  }
#endif
}									



//
//	Packed Rotate Left 8x16bit by 1
//
void vmp_prol1HPPC( uint16 * const phD, const uint16 * const phA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _r27, _r26, _r25, _r24, _lMsk, _rMsk;

#ifndef USE_ASSERT
#pragma unused(phD, phA)
#endif

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	__asm {
		lis		_rMsk,0x7fff		// =7FFF0000
		lis		_lMsk,0x8000		// =80000000
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		addi	_rMsk,_rMsk,0x7fff	// =7FFF7FFF
		ori		_lMsk,_lMsk,0x8000	// =80008000
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		and		_r27,_r31,_rMsk		// (val0 & 0x7fff7fff)
		and		_r31,_r31,_lMsk		// (val0 & 0x80008000)
		and		_r26,_r30,_rMsk
		and		_r30,_r30,_lMsk
		and		_r25,_r29,_rMsk
		and		_r29,_r29,_lMsk
		and 	_r24,_r28,_rMsk
		and		_r28,_r28,_lMsk

		srwi	_r31,_r31,15		// ((val0 & 0x80008000) >> 15)
		slwi	_r27,_r27,1			// ((val0 & 0x7fff7fff) << 1 )
		srwi	_r30,_r30,15
		slwi	_r26,_r26,1
		srwi	_r29,_r29,15
		slwi	_r25,_r25,1
		srwi	_r28,_r28,15
		slwi	_r24,_r24,1

		or		_r31,_r31,_r27		// Blend bits
		or		_r30,_r30,_r26
		or		_r29,_r29,_r25
		or		_r28,_r28,_r24
			
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
	
#else
	uint32 val0, val1, val2, val3;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	val0 = *(((uint32*)phA)+0);
	val1 = *(((uint32*)phA)+1);
	val2 = *(((uint32*)phA)+2);
	val3 = *(((uint32*)phA)+3);

	*(((uint32*)phD)+0) = ((val0 & 0x7fff7fff) << 1 )
						| ((val0 & 0x80008000) >> 15);
	*(((uint32*)phD)+1) = ((val1 & 0x7fff7fff) << 1 )
						| ((val1 & 0x80008000) >> 15);
	*(((uint32*)phD)+2) = ((val2 & 0x7fff7fff) << 1 )
						| ((val2 & 0x80008000) >> 15);
	*(((uint32*)phD)+3) = ((val3 & 0x7fff7fff) << 1 )
						| ((val3 & 0x80008000) >> 15);
#endif
}



//
//	Packed Rotate Left 4x32bit by n:{0...31}
//
void vmp_prolWPPC( uint32 * const pwD, const uint32 * const pwA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28;

#ifdef USE_ASSERT
#pragma unused(count)
#else
#pragma unused(pwD, pwA, count)
#endif

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	__asm {				// Note: any wrap on count is clipped by processor! (0...31)
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		rlwnm	_r31,_r31,r5,0,31	// rotate and set ALL bits
		rlwnm	_r30,_r30,r5,0,31
		rlwnm	_r29,_r29,r5,0,31
		rlwnm	_r28,_r28,r5,0,31
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	uint iCnt;

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	if (count)
	  {
		iCnt = 32-count;

		*(pwD+0) = (*(pwA+0) >> iCnt) | (*(pwA+0) << count);
		*(pwD+1) = (*(pwA+1) >> iCnt) | (*(pwA+1) << count);
		*(pwD+2) = (*(pwA+2) >> iCnt) | (*(pwA+2) << count);
		*(pwD+3) = (*(pwA+3) >> iCnt) | (*(pwA+3) << count);
	  }
	else
	  {
		*(pwD+0) = *(pwA+0);
		*(pwD+1) = *(pwA+1);
		*(pwD+2) = *(pwA+2);
		*(pwD+3) = *(pwA+3);
	  }
#endif
}									



//
//	Packed Rotate Left 4x32bit by 1
//
void vmp_prol1WPPC( uint32 * const pwD, const uint32 * const pwA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28;

#ifndef USE_ASSERT
#pragma unused(pwD, pwA)
#endif

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	__asm {
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		rlwinm	_r31,_r31,1,0,31
		rlwinm	_r30,_r30,1,0,31
		rlwinm	_r29,_r29,1,0,31
		rlwinm	_r28,_r28,1,0,31
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
	
#else
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(pwD+0) = (*(pwA+0) >> 31) | (*(pwA+0) << 1);
	*(pwD+1) = (*(pwA+1) >> 31) | (*(pwA+1) << 1);
	*(pwD+2) = (*(pwA+2) >> 31) | (*(pwA+2) << 1);
	*(pwD+3) = (*(pwA+3) >> 31) | (*(pwA+3) << 1);
#endif
}

