
//
//  psraPPC.cpp - Packed Shift Right Arithmetic (PowerPC) 
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      19 Dec 2001 - JCL - Brought up to date.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "PRot.h"
#include "PRotEx.h"


extern uint32 llMaskBD[];
extern uint32 rrMaskBD[];
extern uint32 llMaskHD[];
extern uint32 rrMaskHD[];




//
//	Packed Shift Right Arithmetic 16x8bit by n:{0...7}
//
void vmp_psraBPPC( int8 * const pbD, const int8 * const pbA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _r27, _r26, _r25, _r24;
	unsigned int register _r23, _r22, _r21, _r20, _r19, _r18, _r17, _r16;
	unsigned int register _lMsk, _mMsk, _hMsk;

#ifdef USE_ASSERT	
#pragma unused(count)
#else
#pragma unused(pbD, pbA, count)
#endif

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	__asm {
		clrlwi	r5,r5,29		// Clear upper 29 bits so value is clipped at (0...7)
		lis		_hMsk,0xFF00	// >>0 =  FF000000
		lis		_mMsk,0x00FF	// >>8 =  00FF0000
		li		_lMsk,0xFF00	// >>16 = 0000FF00
		
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)

		slwi	_r27,_r31,16	// (val << 16)
		slwi	_r23,_r31,24	// (val << 24)
		slwi	_r19,_r31,8		// (val << 8)
		sraw	_r31,_r31,r5	// (val >> count)
		
		slwi	_r26,_r30,16
		slwi	_r22,_r30,24
		slwi	_r18,_r30,8
		sraw	_r30,_r30,r5
		
		slwi	_r25,_r29,16
		slwi	_r21,_r29,24
		slwi	_r17,_r29,8
		sraw	_r29,_r29,r5
		
		slwi	_r24,_r28,16
		slwi	_r20,_r28,24
		slwi	_r16,_r28,8
		sraw	_r28,_r28,r5

		sraw	_r27,_r27,r5	// ((val<<16) >> count)
		sraw	_r23,_r23,r5	// ((val<<24) >> count)
		sraw	_r19,_r19,r5	// ((val<<8)  >> count)
		
		sraw	_r26,_r26,r5
		sraw	_r22,_r22,r5
		sraw	_r18,_r18,r5
		
		sraw	_r25,_r25,r5
		sraw	_r21,_r21,r5
		sraw	_r17,_r17,r5
		
		sraw	_r24,_r24,r5
		sraw	_r20,_r20,r5
		sraw	_r16,_r16,r5
		
		srwi	_r27,_r27,16	// (((uint32)((val<<16) >> count)) >> 16)
		srwi	_r23,_r23,24	// (((uint32)((val<<24) >> count)) >> 24)
		srwi	_r19,_r19,8		// (((uint32)((val<< 8) >> count)) >>  8)
		
		srwi	_r26,_r26,16
		srwi	_r22,_r22,24
		srwi	_r18,_r18,8
		
		srwi	_r25,_r25,16
		srwi	_r21,_r21,24
		srwi	_r17,_r17,8
		
		srwi	_r24,_r24,16
		srwi	_r20,_r20,24
		srwi	_r16,_r16,8
		
		and		_r31,_r31,_hMsk	// ((val >> count) & 0xFF000000)
		and		_r27,_r27,_lMsk	// ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		and		_r19,_r19,_mMsk	// ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)

		and		_r30,_r30,_hMsk
		and		_r26,_r26,_lMsk
		and		_r18,_r18,_mMsk

		and		_r29,_r29,_hMsk
		and		_r25,_r25,_lMsk
		and		_r17,_r17,_mMsk

		and		_r28,_r28,_hMsk
		and		_r24,_r24,_lMsk
		and		_r16,_r16,_mMsk
		
		or		_r31,_r31,_r27	// Blend Bits
		or		_r23,_r23,_r19
		
		or		_r30,_r30,_r26
		or		_r22,_r22,_r18
		
		or		_r29,_r29,_r25
		or		_r21,_r21,_r17
		
		or		_r28,_r28,_r24
		or		_r20,_r20,_r16
		
		or		_r31,_r31,_r23
		or		_r30,_r30,_r22
		or		_r29,_r29,_r21
		or		_r28,_r28,_r20
		
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	int32 val, *pA;
	uint32 *pD;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	pA = (int32*)pbA;
	pD = (uint32*)pbD;

	count &= (8-1);				// Clip count to data bit size -1

	val = *(pA+0);
	*(pD+0) = (((uint32)((val<<24) >> count)) >> 24)
		| ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		| ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)
		| ((val >> count) & 0xFF000000);

	val = *(pA+1);
	*(pD+1) = (((uint32)((val<<24) >> count)) >> 24)
		| ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		| ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)
		| ((val >> count) & 0xFF000000);

	val = *(pA+2);
	*(pD+2) = (((uint32)((val<<24) >> count)) >> 24)
		| ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		| ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)
		| ((val >> count) & 0xFF000000);

	val = *(pA+3);
	*(pD+3) = (((uint32)((val<<24) >> count)) >> 24)
		| ((((uint32)((val<<16) >> count)) >> 16) & 0x0000FF00)
		| ((((uint32)((val<< 8) >> count)) >>  8) & 0x00FF0000)
		| ((val >> count) & 0xFF000000);
#endif
}									



//
//	Packed Shift Right Arithmetic 16x8bit by 1
//
void vmp_psra1BPPC( int8 * const pbD, const int8 * const pbA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _r27, _r26, _r25, _r24, _lMsk, _rMsk;

#ifndef USE_ASSERT
#pragma unused(pbD, pbA)
#endif

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	__asm {
		lis		_lMsk,0xfefe		// =FEFE0000
		lis		_rMsk,0x8080		// =80800000
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		ori		_lMsk,_lMsk,0xfefe	// =FEFEFEFE
		ori		_rMsk,_rMsk,0x8080	// =80808080
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)

		and		_r27,_r31,_lMsk		// & FEFEFEFE
		and		_r31,_r31,_rMsk		// & 80808080
		and		_r26,_r30,_lMsk
		and		_r30,_r30,_rMsk
		and		_r25,_r29,_lMsk
		and		_r29,_r29,_rMsk
		and 	_r24,_r28,_lMsk
		and		_r28,_r28,_rMsk
		
		srwi	_r27,_r27,1
		srwi	_r26,_r26,1
		srwi	_r25,_r25,1
		srwi	_r24,_r24,1
	
		or		_r31,_r31,_r27		// Blend Sticky (sign) bit
		or		_r30,_r30,_r26
		or		_r29,_r29,_r25
		or		_r28,_r28,_r24
		
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	vmp_psraBPPC( pbD, pbA, 1 );
#endif
}



//
//	Packed Shift Right Arithmetic 8x16bit by n:{0...15}
//
void vmp_psraHPPC( int16 * const phD, const int16 * const phA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _r27, _r26, _r25, _r24, _eMsk;

#ifdef USE_ASSERT
#pragma unused(count)
#else
#pragma unused(phD, phA, count)
#endif

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	__asm {
		clrlwi	r5,r5,28		// Clear upper 28 bits so value is clipped at (0...15)
		lis		_eMsk,0xFFFF

		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)

		slwi	_r27,_r31,16	// (val <<16)
		sraw	_r31,_r31,r5	// (val >> count)
		slwi	_r26,_r30,16
		sraw	_r30,_r30,r5
		slwi	_r25,_r29,16
		sraw	_r29,_r29,r5
		slwi	_r24,_r28,16
		sraw	_r28,_r28,r5

		sraw	_r27,_r27,r5	// ((val<<16) >> count)
		and		_r31,_r31,_eMsk	// ((val >> count) & 0xFFFF0000)
		sraw	_r26,_r26,r5
		and		_r30,_r30,_eMsk
		sraw	_r25,_r25,r5
		and		_r29,_r29,_eMsk
		sraw	_r24,_r24,r5
		and		_r28,_r28,_eMsk
		
		srwi	_r27,_r27,16
		srwi	_r26,_r26,16
		srwi	_r25,_r25,16
		srwi	_r24,_r24,16
	
		or		_r31,_r31,_r27	// Blend Bits
		or		_r30,_r30,_r26
		or		_r29,_r29,_r25
		or		_r28,_r28,_r24
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	int32 val, *pA;
	uint32 *pD;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	pA = (int32*)phA;
	pD = (uint32*)phD;

	count &= (16-1);				// Clip count to data bit size -1

	val = *(pA+0);
	*(pD+0) = (((uint32)((val<<16) >> count)) >> 16)
		| ((val >> count) & 0xFFFF0000);

	val = *(pA+1);
	*(pD+1) = (((uint32)((val<<16) >> count)) >> 16)
		| ((val >> count) & 0xFFFF0000);

	val = *(pA+2);
	*(pD+2) = (((uint32)((val<<16) >> count)) >> 16)
		| ((val >> count) & 0xFFFF0000);

	val = *(pA+3);
	*(pD+3) = (((uint32)((val<<16) >> count)) >> 16)
		| ((val >> count) & 0xFFFF0000);
#endif
}									



//
//	Packed Shift Right Arithmetic 8x16bit by 1
//
void vmp_psra1HPPC( int16 * const phD, const int16 * const phA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28;

#ifndef USE_ASSERT
#pragma unused(phD, phA)
#endif

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	__asm {
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)

		rlwimi	_r31,_r31,1,15,15			// reverse shift (lower) sign bit then insert
		rlwimi	_r30,_r30,1,15,15
		rlwimi	_r29,_r29,1,15,15
		rlwimi	_r28,_r28,1,15,15
		
		srawi	_r31,_r31,1
		srawi	_r30,_r30,1
		srawi	_r29,_r29,1
		srawi	_r28,_r28,1

		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	vmp_psraHPPC( phD, phA, 1 );
#endif
}



//
//	Packed Shift Right Arithmetic 4x32bit by n:{0...31}
//
void vmp_psraWPPC( int32 * const pwD, const int32 * const pwA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28;

#ifdef USE_ASSERT
#pragma unused(count)
#else
#pragma unused(pwD, pwA, count)
#endif

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	__asm {				// Note: any wrap on count is clipped by processor! (0...31)
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		sraw	_r31,_r31,r5
		sraw	_r30,_r30,r5
		sraw	_r29,_r29,r5
		sraw	_r28,_r28,r5
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(pwD+0) = (*(pwA+0)) >> count;
	*(pwD+1) = (*(pwA+1)) >> count;
	*(pwD+2) = (*(pwA+2)) >> count;
	*(pwD+3) = (*(pwA+3)) >> count;
#endif
}									



//
//	Packed Shift Right Arithmetic 4x32bit by 1
//
void vmp_psra1WPPC( int32 * const pwD, const int32 * const pwA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28;

#ifndef USE_ASSERT
#pragma unused(pwD, pwA)
#endif

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	__asm {
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		srawi	_r31,_r31,1
		srawi	_r30,_r30,1
		srawi	_r29,_r29,1
		srawi	_r28,_r28,1
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	vmp_psraWPPC( pwD, pwA, 1 );
#endif
}

