
//
//  psrlPPC.cpp - Packed Shift Right Logical (PowerPC) 
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//      19 Dec 2001 - JCL - Brought up to date.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "PRot.h"
#include "PRotEx.h"


extern uint32 llMaskBD[];
extern uint32 rrMaskBD[];
extern uint32 llMaskHD[];
extern uint32 rrMaskHD[];



//
//	Packed Shift Right Logical 16x8bit by n:{0...7}
//
void vmp_psrlBPPC( uint8 * const pbD, const uint8 * const pbA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _eMsk, _rTbl;

#ifdef USE_ASSERT
#pragma unused(count)
#else
#pragma unused(pbD, pbA, count)
#endif

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	__asm {
		clrlwi	r5,r5,29		// Clear upper 29 bits so value is clipped at (0...7)
		lwz		_rTbl,llMaskBD

		lwz		_r31,0(r4)
		slwi	_eMsk,r5,2					// x4
		lwz		_r30,4(r4)
		add		_rTbl,_rTbl,_eMsk
		lwz		_r29,8(r4)
		lwz		_eMsk,0(_rTbl)
		lwz		_r28,12(r4)
	
		and		_r31,_r31,_eMsk
		and		_r30,_r30,_eMsk
		and		_r29,_r29,_eMsk
		and 	_r28,_r28,_eMsk
	
		srw		_r31,_r31,r5
		srw		_r30,_r30,r5
		srw		_r29,_r29,r5
		srw		_r28,_r28,r5
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	uint32 msk;

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	count &= (8-1);				// Clip count to data bit size -1

	msk = llMaskBD[count];
	*(((uint32*)pbD)+0) = (*(((uint32*)pbA)+0) & msk) >> count;
	*(((uint32*)pbD)+1) = (*(((uint32*)pbA)+1) & msk) >> count;
	*(((uint32*)pbD)+2) = (*(((uint32*)pbA)+2) & msk) >> count;
	*(((uint32*)pbD)+3) = (*(((uint32*)pbA)+3) & msk) >> count;
#endif
}									



//
//	Packed Shift Right Logical 16x8bit by 1
//
void vmp_psrl1BPPC( uint8 * const pbD, const uint8 * const pbA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _eMsk;

#ifndef USE_ASSERT
#pragma unused(pbD, pbA)
#endif

	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	__asm {
		lis		_eMsk,0xfefe		// =FEFE0000
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		ori		_eMsk,_eMsk,0xfefe	// =FEFEFEFE
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		and		_r31,_r31,_eMsk
		and		_r30,_r30,_eMsk
		and		_r29,_r29,_eMsk
		and 	_r28,_r28,_eMsk
	
		srwi	_r31,_r31,1
		srwi	_r30,_r30,1
		srwi	_r29,_r29,1
		srwi	_r28,_r28,1
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
	
#else
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbD);

	*(((uint32*)pbD)+0) = (*(((uint32*)pbA)+0) & 0xfefefefe) >> 1;
	*(((uint32*)pbD)+1) = (*(((uint32*)pbA)+1) & 0xfefefefe) >> 1;
	*(((uint32*)pbD)+2) = (*(((uint32*)pbA)+2) & 0xfefefefe) >> 1;
	*(((uint32*)pbD)+3) = (*(((uint32*)pbA)+3) & 0xfefefefe) >> 1;
#endif
}



//
//	Packed Shift Right Logical 8x16bit by n:{0...15}
//
void vmp_psrlHPPC( uint16 * const phD, const uint16 * const phA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _eMsk, _rTbl;

#ifdef USE_ASSERT
#pragma unused(count)
#else
#pragma unused(phD, phA, count)
#endif

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	__asm {
		clrlwi	r5,r5,28		// Clear upper 28 bits so value is clipped at (0...15)
		lwz		_rTbl,llMaskHD

		lwz		_r31,0(r4)
		slwi	_eMsk,r5,2					// x4
		lwz		_r30,4(r4)
		add		_rTbl,_rTbl,_eMsk
		lwz		_r29,8(r4)
		lwz		_eMsk,0(_rTbl)
		lwz		_r28,12(r4)
	
		and		_r31,_r31,_eMsk
		and		_r30,_r30,_eMsk
		and		_r29,_r29,_eMsk
		and 	_r28,_r28,_eMsk
	
		srw		_r31,_r31,r5
		srw		_r30,_r30,r5
		srw		_r29,_r29,r5
		srw		_r28,_r28,r5
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	uint32 msk;

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	count &= (16-1);				// Clip count to data bit size -1

	msk = llMaskHD[count];
	*(((uint32*)phD)+0) = (*(((uint32*)phA)+0) & msk) >> count;
	*(((uint32*)phD)+1) = (*(((uint32*)phA)+1) & msk) >> count;
	*(((uint32*)phD)+2) = (*(((uint32*)phA)+2) & msk) >> count;
	*(((uint32*)phD)+3) = (*(((uint32*)phA)+3) & msk) >> count;
#endif
}									



//
//	Packed Shift Right Logical 8x16bit by 1
//
void vmp_psrl1HPPC( uint16 * const phD, const uint16 * const phA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28, _eMsk;

#ifndef USE_ASSERT
#pragma unused(phD, phA)
#endif

	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	__asm {
		lis		_eMsk,0xfffe		// =FFFE0000
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		ori		_eMsk,_eMsk,0xfffe	// =FFFEFFFE
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		and		_r31,_r31,_eMsk
		and		_r30,_r30,_eMsk
		and		_r29,_r29,_eMsk
		and 	_r28,_r28,_eMsk
	
		srwi	_r31,_r31,1
		srwi	_r30,_r30,1
		srwi	_r29,_r29,1
		srwi	_r28,_r28,1
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phD);

	*(((uint32*)phD)+0) = (*(((uint32*)phA)+0) & 0xfffefffe) >> 1;
	*(((uint32*)phD)+1) = (*(((uint32*)phA)+1) & 0xfffefffe) >> 1;
	*(((uint32*)phD)+2) = (*(((uint32*)phA)+2) & 0xfffefffe) >> 1;
	*(((uint32*)phD)+3) = (*(((uint32*)phA)+3) & 0xfffefffe) >> 1;
#endif
}



//
//	Packed Shift Right Logical 4x32bit by n:{0...31}
//
void vmp_psrlWPPC( uint32 * const pwD, const uint32 * const pwA, uint count )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28;

#ifdef USE_ASSERT
#pragma unused(count)
#else
#pragma unused(pwD, pwA, count)
#endif

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	__asm {				// Note: any wrap on count is clipped by processor! (0...31)
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		srw		_r31,_r31,r5
		srw		_r30,_r30,r5
		srw		_r29,_r29,r5
		srw		_r28,_r28,r5
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
#else
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	count &= (32-1);				// Clip count to data bit size -1

	*(pwD+0) = (*(pwA+0)) >> count;
	*(pwD+1) = (*(pwA+1)) >> count;
	*(pwD+2) = (*(pwA+2)) >> count;
	*(pwD+3) = (*(pwA+3)) >> count;
#endif
}									



//
//	Packed Shift Right Logical 4x32bit by 1
//
void vmp_psrl1WPPC( uint32 * const pwD, const uint32 * const pwA )
{
#if 01
	unsigned int register _r31, _r30, _r29, _r28;

#ifndef USE_ASSERT
#pragma unused(pwD, pwA)
#endif

	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	__asm {
		lwz		_r31,0(r4)
		lwz		_r30,4(r4)
		lwz		_r29,8(r4)
		lwz		_r28,12(r4)
	
		srwi	_r31,_r31,1
		srwi	_r30,_r30,1
		srwi	_r29,_r29,1
		srwi	_r28,_r28,1
	
		stw		_r31,0(r3)
		stw		_r30,4(r3)
		stw		_r29,8(r3)
		stw		_r28,12(r3)
	}
	
#else
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwD);

	*(pwD+0) = *(pwA+0) >> 1;
	*(pwD+1) = *(pwA+1) >> 1;
	*(pwD+2) = *(pwA+2) >> 1;
	*(pwD+3) = *(pwA+3) >> 1;
#endif
}


