
//
//  Bench.cpp - Packed Addition/Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 31 Dec 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include <iostream.h>
#include <string.h>
#include <assert.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"
#include "Pas.h"
#include "PasEx.h"



typedef enum
{
	iTEST_PADDB,			// add 8bit
	iTEST_PADDH,			// add 16bit
	iTEST_PADDW,			// add 32bit

	iTEST_PSUBB,			// sub 8bit
	iTEST_PSUBH,			// sub 16bit
	iTEST_PSUBW,			// sub 32bit

	iTEST_PADDSB,			// add 8bit signed w/Saturation
	iTEST_PADDSH,			// add 16bit

	iTEST_PSUBSB,			// sub 8bit
	iTEST_PSUBSH,			// sub 16bit

	iTEST_PADDUSB,			// add 8bit unsigned w/Saturation
	iTEST_PADDUSH,			// add 16bit

	iTEST_PSUBUSB,			// sub 8bit
	iTEST_PSUBUSH			// sub 16bit
} iTEST;




//
//	Compare two 128bit bit blocks
//

bool vmp_IsPVEqual( void * const pvA, void * const pvB )
{
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	if ( memcmp( pvA, pvB, 16 ) )
	  {
	  	return false;
	  }

  	return true;
}												



//
//	Test Packed Shift 128bit
//
static void test_pV( bool bAlign, iTEST iFunc )
{
	int8	v;
    uint8	buf[4*16 +15], *bt, *bu, *bv, *bw;

	if (bAlign)
	  {
		bt = (uint8*) ALIGN16((int)buf);
	  }
	else
	  {
		bt = buf;
	  }

	bu = bt + (128/8);
	bv = bu + (128/8);
	bw = bv + (128/8);

	v = 0;
	do {
		bt[0]=v;     bt[1]=v+1;   bt[2]=v+2;   bt[3]=v+3;
		bt[4]=v+4;   bt[5]=v+5;   bt[6]=v+6;   bt[7]=v+7;
		bt[8]=v+8;   bt[9]=v+9;   bt[10]=v+10; bt[11]=v+11;
		bt[12]=v+12; bt[13]=v+13; bt[14]=v+14; bt[15]=v+15;

		bw[0]=v+15;  bw[1]=v+14;  bw[2]=v+13;  bw[3]=v+12;
		bw[4]=v+11;  bw[5]=v+10;  bw[6]=v+9;   bw[7]=v+8;
		bw[8]=v+7;   bw[9]=v+6;   bw[10]=v+5;  bw[11]=v+4;
		bw[12]=v+3;  bw[13]=v+2;  bw[14]=v+1;  bw[15]=v+0;

		switch (iFunc)
		  {
  			case iTEST_PADDB:						// add 8bit
				vmp_paddBGeneric(bu, bt, bw);
				if ( bAlign )
					vmp_paddB(   bv, bt, bw);
				else
					vmp_paddB0(  bv, bt, bw);
				break;

			case iTEST_PADDH:						// add 16bit
				vmp_paddHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
				if ( bAlign )
					vmp_paddH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
				else
					vmp_paddH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
				break;

			case iTEST_PADDW:						// add 32bit
				vmp_paddWGeneric((uint32*)bu, (uint32*)bt, (uint32*)bw);
				if ( bAlign )
					vmp_paddW(   (uint32*)bv, (uint32*)bt, (uint32*)bw);
				else
					vmp_paddW0(  (uint32*)bv, (uint32*)bt, (uint32*)bw);
				break;

  			case iTEST_PADDUSB:						// add 8bit unsigned w/Saturation
				vmp_paddusBGeneric(bu, bt, bw);
				if ( bAlign )
					vmp_paddusB(   bv, bt, bw);
				else
					vmp_paddusB0(  bv, bt, bw);
				break;

			case iTEST_PADDUSH:						// add 16bit unsigned w/Saturation
				vmp_paddusHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
				if ( bAlign )
					vmp_paddusH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
				else
					vmp_paddusH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
				break;

  			case iTEST_PADDSB:						// add 8bit signed w/Saturation
				vmp_paddsBGeneric((int8*)bu, (int8*)bt, (int8*)bw);
				if ( bAlign )
					vmp_paddsB(   (int8*)bv, (int8*)bt, (int8*)bw);
				else
					vmp_paddsB0(  (int8*)bv, (int8*)bt, (int8*)bw);
				break;

			case iTEST_PADDSH:						// add 16bit signed w/Saturation
				vmp_paddsHGeneric((int16*)bu, (int16*)bt, (int16*)bw);
				if ( bAlign )
					vmp_paddsH(   (int16*)bv, (int16*)bt, (int16*)bw);
				else
					vmp_paddsH0(  (int16*)bv, (int16*)bt, (int16*)bw);
				break;

  			case iTEST_PSUBB:						// sub 8bit
				vmp_psubBGeneric(bu, bt, bw);
				if ( bAlign )
					vmp_psubB(   bv, bt, bw);
				else
					vmp_psubB0(  bv, bt, bw);
				break;

			case iTEST_PSUBH:						// sub 16bit
				vmp_psubHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
				if ( bAlign )
					vmp_psubH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
				else
					vmp_psubH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
				break;

			case iTEST_PSUBW:						// sub 32bit
				vmp_psubWGeneric((uint32*)bu, (uint32*)bt, (uint32*)bw);
				if ( bAlign )
					vmp_psubW(   (uint32*)bv, (uint32*)bt, (uint32*)bw);
				else
					vmp_psubW0(  (uint32*)bv, (uint32*)bt, (uint32*)bw);
				break;

  			case iTEST_PSUBUSB:						// sub 8bit unsigned w/Saturation
				vmp_psubusBGeneric(bu, bt, bw);
				if ( bAlign )
					vmp_psubusB(   bv, bt, bw);
				else
					vmp_psubusB0(  bv, bt, bw);
				break;

			case iTEST_PSUBUSH:						// sub 16bit unsigned w/Saturation
				vmp_psubusHGeneric((uint16*)bu, (uint16*)bt, (uint16*)bw);
				if ( bAlign )
					vmp_psubusH(   (uint16*)bv, (uint16*)bt, (uint16*)bw);
				else
					vmp_psubusH0(  (uint16*)bv, (uint16*)bt, (uint16*)bw);
				break;

  			case iTEST_PSUBSB:						// sub 8bit signed w/Saturation
				vmp_psubsBGeneric((int8*)bu, (int8*)bt, (int8*)bw);
				if ( bAlign )
					vmp_psubsB(   (int8*)bv, (int8*)bt, (int8*)bw);
				else
					vmp_psubsB0(  (int8*)bv, (int8*)bt, (int8*)bw);
				break;

			case iTEST_PSUBSH:						// sub 16bit signed w/Saturation
				vmp_psubsHGeneric((int16*)bu, (int16*)bt, (int16*)bw);
				if ( bAlign )
					vmp_psubsH(   (int16*)bv, (int16*)bt, (int16*)bw);
				else
					vmp_psubsH0(  (int16*)bv, (int16*)bt, (int16*)bw);
				break;

		  }

		if ( !vmp_IsPVEqual( bu, bv ) )
		  {
			cout << "Mismatch on " << v << endl;
		  }
	  } while( 0 != ++v);
}



int main( int argc, const char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

#pragma unused(argc, argv)

    CpuDetect( &cinfo );								// Detect CPU
    
    cout << "\nPacked (Vector) Addition/Subtraction Test\n\n";
    cout << CpuInfoStr( szBuf, &cinfo ) << endl;
    
	PasSetup( &cinfo );								// Setup our 3D Vector stuff

	cout << "Testing (Vector) Addition 16x8bit" << endl;
	test_pV( false, iTEST_PADDB );

	cout << "Testing (Vector) Addition 16x8bit (Aligned)" << endl;
	test_pV( true,  iTEST_PADDB );

	cout << "Testing (Vector) Addition 8x16bit" << endl;
	test_pV( false, iTEST_PADDH );

	cout << "Testing (Vector) Addition 8x16bit (Aligned)" << endl;
	test_pV( true,  iTEST_PADDH );

	cout << "Testing (Vector) Addition 4x32bit" << endl;
	test_pV( false, iTEST_PADDW );

	cout << "Testing (Vector) Addition 4x32bit (Aligned)" << endl;
	test_pV( true,  iTEST_PADDW );
	cout << endl;


	cout << "Testing (Vector) Subtraction 16x8bit" << endl;
	test_pV( false, iTEST_PSUBB );

	cout << "Testing (Vector) Subtraction 16x8bit (Aligned)" << endl;
	test_pV( true,  iTEST_PSUBB );

	cout << "Testing (Vector) Subtraction 8x16bit" << endl;
	test_pV( false, iTEST_PSUBH );

	cout << "Testing (Vector) Subtraction 8x16bit (Aligned)" << endl;
	test_pV( true,  iTEST_PSUBH );

	cout << "Testing (Vector) Subtraction 4x32bit" << endl;
	test_pV( false, iTEST_PSUBW );

	cout << "Testing (Vector) Subtraction 4x32bit (Aligned)" << endl;
	test_pV( true,  iTEST_PSUBW );
	cout << endl;


	cout << "Testing (Vector) Addition 16x8bit unsigned w/Saturation" << endl;
	test_pV( false, iTEST_PADDUSB );

	cout << "Testing (Vector) Addition 16x8bit unsigned w/Saturation (Aligned)" << endl;
	test_pV( true,  iTEST_PADDUSB );

	cout << "Testing (Vector) Addition 8x16bit unsigned w/Saturation" << endl;
	test_pV( false, iTEST_PADDUSH );

	cout << "Testing (Vector) Addition 8x16bit unsigned w/Saturation (Aligned)" << endl;
	test_pV( true,  iTEST_PADDUSH );
	cout << endl;

	cout << "Testing (Vector) Subtraction 16x8bit unsigned w/Saturation" << endl;
	test_pV( false, iTEST_PSUBUSB );

	cout << "Testing (Vector) Subtraction 16x8bit unsigned w/Saturation (Aligned)" << endl;
	test_pV( true,  iTEST_PSUBUSB );

	cout << "Testing (Vector) Subtraction 8x16bit unsigned w/Saturation" << endl;
	test_pV( false, iTEST_PSUBUSH );

	cout << "Testing (Vector) Subtraction 8x16bit unsigned w/Saturation (Aligned)" << endl;
	test_pV( true,  iTEST_PSUBUSH );
	cout << endl;


	cout << "Testing (Vector) Addition 16x8bit signed w/Saturation" << endl;
	test_pV( false, iTEST_PADDSB );

	cout << "Testing (Vector) Addition 16x8bit signed w/Saturation (Aligned)" << endl;
	test_pV( true,  iTEST_PADDSB );

	cout << "Testing (Vector) Addition 8x16bit signed w/Saturation" << endl;
	test_pV( false, iTEST_PADDSH );

	cout << "Testing (Vector) Addition 8x16bit signed w/Saturation (Aligned)" << endl;
	test_pV( true,  iTEST_PADDSH );
	cout << endl;

	cout << "Testing (Vector) Subtraction 16x8bit signed w/Saturation" << endl;
	test_pV( false, iTEST_PSUBSB );

	cout << "Testing (Vector) Subtraction 16x8bit signed w/Saturation (Aligned)" << endl;
	test_pV( true,  iTEST_PSUBSB );

	cout << "Testing (Vector) Subtraction 8x16bit signed w/Saturation" << endl;
	test_pV( false, iTEST_PSUBSH );

	cout << "Testing (Vector) Subtraction 8x16bit signed w/Saturation (Aligned)" << endl;
	test_pV( true,  iTEST_PSUBSH );
	cout << endl;

	cout << "Fin\n\n" << flush;

    return 0;
}

