
//
//  PasAltivec.cpp - Packed Addition/Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 31 Dec 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "Pas.h"
#include "PasEx.h"



//
//	(Quad) Vector Summation (Addition) - Unaligned
//

void vmp_paddBAltivec0( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned char vD, vA, vB;

	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pbA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pbB)+3);

	vD = vec_add( vA, vB );		// vaddubm d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vD)+3);
}



//
//	(Quad) Vector Summation (Addition) - Aligned
//

void vmp_paddBAltivec( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector unsigned char *)pbD = vec_add( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) ); // vaddubm d,a,b
}



//
//	Packed (un)signed Addition 8x16bit - Unaligned
//
void vmp_paddHAltivec0( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned short vD, vA, vB;

	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)phA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)phB)+3);

	vD = vec_add( vA, vB );		// vadduhm d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed (un)signed Addition 8x16bit - Aligned
//
void vmp_paddHAltivec( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector unsigned short *)phD = vec_add( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) ); // vadduhm d,a,b
}



//
//	Packed (un)signed Addition 4x32bit - Unaligned
//
void vmp_paddWAltivec0( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
{
	vector unsigned int vD, vA, vB;

	ASSERT_PTR4(pwD);
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pwA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pwA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pwA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pwA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)pwB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pwB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pwB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pwB)+3);

	vD = vec_add( vA, vB );		// vadduwm d,a,b

	*(((uint32 *)pwD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed (un)signed Addition 4x32bit - Aligned
//
void vmp_paddWAltivec( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
{
	ASSERT_PTR16(pwD);
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwB);

	*(vector unsigned int *)pwD = vec_add( (*(vector unsigned int *)pwA), (*(vector unsigned int *)pwB) ); // vadduwm d,a,b
}



//______________________________________________________



//
//	Packed unsigned Addition 16x8bit w/Saturation {0...255} - unaligned
//
void vmp_paddusBAltivec0( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned char vD, vA, vB;

	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pbA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pbB)+3);

	vD = vec_adds( vA, vB );		// vaddubs d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vD)+3);
}


//
//	Packed unsigned Addition 16x8bit w/Saturation {0...255} - aligned
//
void vmp_paddusBAltivec( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector unsigned char *)pbD = vec_adds( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) ); // vaddubs d,a,b
}



//
//	Packed unsigned Addition 8x16bit w/Saturation {0...65535} - unaligned
//
void vmp_paddusHAltivec0( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned short vD, vA, vB;

	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)phA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)phB)+3);

	vD = vec_adds( vA, vB );		// vadduhs d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vD)+3);
}


//
//	Packed unsigned Addition 8x16bit w/Saturation {0...65535} - aligned
//
void vmp_paddusHAltivec( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector unsigned short *)phD = vec_adds( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) ); // vadduhs d,a,b
}



//______________________________________________________



//
//	Packed signed Addition 16x8bit w/Saturation {-128...127} - unaligned
//
void vmp_paddsBAltivec0( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
{
	vector signed char vD, vA, vB;

	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pbA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pbB)+3);

	vD = vec_adds( vA, vB );		// vaddsbs d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed signed Addition 16x8bit w/Saturation {-128...127} - aligned
//
void vmp_paddsBAltivec( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
{
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector signed char *)pbD = vec_adds( (*(vector signed char *)pbA), (*(vector signed char *)pbB) ); // vaddsbs d,a,b
}



//
//	Packed signed Addition 8x16bit w/Saturation {-32768...32767} - unaligned
//
void vmp_paddsHAltivec0( int16 * const phD, const int16 * const phA, const int16 * const phB )
{
	vector signed short vD, vA, vB;

	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)phA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)phB)+3);

	vD = vec_adds( vA, vB );		// vaddshs d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed signed Addition 8x16bit w/Saturation {-32768...32767} - aligned
//
void vmp_paddsHAltivec( int16 * const phD, const int16 * const phA, const int16 * const phB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector signed short *)phD = vec_adds( (*(vector signed short *)phA), (*(vector signed short *)phB) ); // vaddshs d,a,b
}



//______________________________________________________



//
//	Packed (un)signed Subtraction 16x8bit - unaligned
//
void vmp_psubBAltivec0( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned char vD, vA, vB;

	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pbA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pbB)+3);

	vD = vec_sub( vA, vB );		// vsububm d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed (un)signed Subtraction 16x8bit - aligned
//
void vmp_psubBAltivec( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector unsigned char *)pbD = vec_sub( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) ); // vsububm d,a,b
}



//
//	Packed (un)signed Subraction 8x16bit - unaligned
//
void vmp_psubHAltivec0( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned short vD, vA, vB;

	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)phA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)phB)+3);

	vD = vec_sub( vA, vB );		// vsubuhm d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed (un)signed Subraction 8x16bit - aligned
//
void vmp_psubHAltivec( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector unsigned short *)phD = vec_sub( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) ); // vsubuhm d,a,b
}



//
//	Packed (un)signed Subtraction 4x32bit - unaligned
//
void vmp_psubWAltivec0( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
{
	vector unsigned int vD, vA, vB;

	ASSERT_PTR4(pwD);
	ASSERT_PTR4(pwA);
	ASSERT_PTR4(pwB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pwA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pwA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pwA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pwA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)pwB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pwB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pwB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pwB)+3);

	vD = vec_sub( vA, vB );		// vsubuwm d,a,b

	*(((uint32 *)pwD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed (un)signed Subtraction 4x32bit - aligned
//
void vmp_psubWAltivec( uint32 * const pwD, const uint32 * const pwA, const uint32 * const pwB )
{
	ASSERT_PTR16(pwD);
	ASSERT_PTR16(pwA);
	ASSERT_PTR16(pwB);

	*(vector unsigned int *)pwD = vec_sub( (*(vector unsigned int *)pwA), (*(vector unsigned int *)pwB) ); // vsubuwm d,a,b
}



//______________________________________________________



//
//	Packed unsigned Subtraction 16x8bit w/Saturation {0...255} - unaligned
//
void vmp_psubusBAltivec0( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned char vD, vA, vB;

	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pbA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pbB)+3);

	vD = vec_subs( vA, vB );		// vsububs d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed unsigned Subtraction 16x8bit w/Saturation {0...255} - aligned
//
void vmp_psubusBAltivec( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector unsigned char *)pbD = vec_subs( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) ); // vsububs d,a,b
}



//
//	Packed unsigned Subtraction 8x16bit w/Saturation {0...65535} - unaligned
//
void vmp_psubusHAltivec0( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned short vD, vA, vB;

	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)phA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)phB)+3);

	vD = vec_subs( vA, vB );		// vsubuhs d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed unsigned Subtraction 8x16bit w/Saturation {0...65535} - aligned
//
void vmp_psubusHAltivec( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector unsigned short *)phD = vec_subs( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) ); // vsubuhs d,a,b
}



//______________________________________________________



//
//	Packed signed Subtraction 16x8bit w/Saturation {-128...127} - unaligned
//
void vmp_psubsBAltivec0( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
{
	vector signed char vD, vA, vB;

	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)pbA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)pbB)+3);

	vD = vec_subs( vA, vB );		// vsubsbs d,a,b

	*(((uint32 *)pbD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed signed Subtraction 16x8bit w/Saturation {-128...127} - aligned
//
void vmp_psubsBAltivec( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
{
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector signed char *)pbD = vec_subs( (*(vector signed char *)pbA), (*(vector signed char *)pbB) ); // vsubsbs d,a,b
}



//
//	Packed signed Subtraction 8x16bit w/Saturation {-32768...32767} - unaligned
//
void vmp_psubsHAltivec0( int16 * const phD, const int16 * const phA, const int16 * const phB )
{
	vector signed short vD, vA, vB;

	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&vA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&vA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&vA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&vA)+3) = *(((uint32 *)phA)+3);

	*(((uint32 *)&vB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&vB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&vB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&vB)+3) = *(((uint32 *)phB)+3);

	vD = vec_subs( vA, vB );		// vsubshs d,a,b

	*(((uint32 *)phD)+0) = *(((uint32 *)&vD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&vD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&vD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&vD)+3);
}



//
//	Packed signed Subtraction 8x16bit w/Saturation {-32768...32767} - aligned
//
void vmp_psubsHAltivec( int16 * const phD, const int16 * const phA, const int16 * const phB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector signed short *)phD = vec_subs( (*(vector signed short *)phA), (*(vector signed short *)phB) ); // vsubshs d,a,b
}



//
//	Setup Packed Addition/Subtraction
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void PasSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);
	    
	if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_paddB =				vmp_paddBAltivec;
		vmp_paddB0 =			vmp_paddBAltivec0;
		vmp_paddH =				vmp_paddHAltivec;
		vmp_paddH0 =			vmp_paddHAltivec0;
		vmp_paddW =				vmp_paddWAltivec;
		vmp_paddW0 =			vmp_paddWAltivec0;

		vmp_psubB =				vmp_psubBAltivec;
		vmp_psubB0 =			vmp_psubBAltivec0;
		vmp_psubH =				vmp_psubHAltivec;
		vmp_psubH0 =			vmp_psubHAltivec0;
		vmp_psubW =				vmp_psubWAltivec;
		vmp_psubW0 =			vmp_psubWAltivec0;

		vmp_paddusB =			vmp_paddusBAltivec;
		vmp_paddusB0 =			vmp_paddusBAltivec0;
		vmp_paddusH =			vmp_paddusHAltivec;
		vmp_paddusH0 =			vmp_paddusHAltivec0;

		vmp_psubusB =			vmp_psubusBAltivec;
		vmp_psubusB0 =			vmp_psubusBAltivec0;
		vmp_psubusH =			vmp_psubusHAltivec;
		vmp_psubusH0 =			vmp_psubusHAltivec0;

		vmp_paddsB =			vmp_paddsBAltivec;
		vmp_paddsB0 =			vmp_paddsBAltivec0;
		vmp_paddsH =			vmp_paddsHAltivec;
		vmp_paddsH0 =			vmp_paddsHAltivec0;

		vmp_psubsB =			vmp_psubsBAltivec;
		vmp_psubsB0 =			vmp_psubsBAltivec0;
		vmp_psubsH =			vmp_psubsHAltivec;
		vmp_psubsH0 =			vmp_psubsHAltivec0;
      }

	else
	  {
		vmp_paddB =				vmp_paddBGeneric;
		vmp_paddB0 =			vmp_paddBGeneric;
		vmp_paddH =				vmp_paddHGeneric;
		vmp_paddH0 =			vmp_paddHGeneric;
		vmp_paddW =				vmp_paddWGeneric;
		vmp_paddW0 =			vmp_paddWGeneric;

		vmp_psubB =				vmp_psubBGeneric;
		vmp_psubB0 =			vmp_psubBGeneric;
		vmp_psubH =				vmp_psubHGeneric;
		vmp_psubH0 =			vmp_psubHGeneric;
		vmp_psubW =				vmp_psubWGeneric;
		vmp_psubW0 =			vmp_psubWGeneric;

		vmp_paddusB =			vmp_paddusBGeneric;
		vmp_paddusB0 =			vmp_paddusBGeneric;
		vmp_paddusH =			vmp_paddusHGeneric;
		vmp_paddusH0 =			vmp_paddusHGeneric;

		vmp_psubusB =			vmp_psubusBGeneric;
		vmp_psubusB0 =			vmp_psubusBGeneric;
		vmp_psubusH =			vmp_psubusHGeneric;
		vmp_psubusH0 =			vmp_psubusHGeneric;

		vmp_paddsB =			vmp_paddsBGeneric;
		vmp_paddsB0 =			vmp_paddsBGeneric;
		vmp_paddsH =			vmp_paddsHGeneric;
		vmp_paddsH0 =			vmp_paddsHGeneric;

		vmp_psubsB =			vmp_psubsBGeneric;
		vmp_psubsB0 =			vmp_psubsBGeneric;
		vmp_psubsH =			vmp_psubsHGeneric;
		vmp_psubsH0 =			vmp_psubsHGeneric;
	  }
}




