
//
//  Bench.cpp - (Quad) Vector 3D Addition & Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include <iostream.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"
#include "QVas3D.h"										// Vector Addition & Subtraction
#include "QVas3DEx.h"									//	"	" (Private)



typedef enum
{
	iTEST_ADD,
	iTEST_SUB,
	iTEST_ADD_SCALAR,
	iTEST_SUB_SCALAR
} iTEST;	





//
//	Test Vector Math
//

static void testQVec( bool bAlign, iTEST iFunc )
{
	vmp3DQVector *pvA, *pvB, *pvTmp, *pvTmp2;
	int x, y, z, w, dif;
	float fPrecision, fScalar;
	float buf[4*4  +3];			// enough space + 12 bytes (3 floats) extra

		// Force proper alignment (Crude just for this demo)

	if ( bAlign )
	  {
		pvA    = (vmp3DQVector*) ALIGN16((int)(buf));
		pvB    = (vmp3DQVector*) ALIGN16((int)(pvA+1));
		pvTmp  = (vmp3DQVector*) ALIGN16((int)(pvB+1));
		pvTmp2 = (vmp3DQVector*) ALIGN16((int)(pvTmp+1));
	  }
	else
	  {
		pvA    = (vmp3DQVector*) buf;
		pvB    = pvA + 1;
		pvTmp  = pvB + 1;
		pvTmp2 = pvTmp + 1;
	  }

	dif = 4;
	fPrecision = 0.0000001f;
	
	for ( w = -dif; w <= dif; w++ )
	  {
	  	pvA->w = 0.123f *(float)w;
		pvB->z = 0.123f *(float)w;
					
	  	for ( z = -dif; z <= dif; z++ )
	  	  {
			pvA->z = 0.123f *(float)z;
			pvB->y = 0.123f *(float)z;

			for ( y = -dif; y <= dif; y++ )
			  {
				pvA->y = 0.123f *(float)y;
				pvB->x = 0.123f *(float)y;

				for ( x = -dif; x <= dif; x++ )
				  {
				    pvB->w = 0.123f *(float)x;
					pvA->x = 0.123f *(float)x;

					fScalar = pvA->x;

					switch (iFunc)
					  {
						case iTEST_ADD:
							vmp_QVecAddGeneric( pvTmp2, pvA, pvB );
							if ( bAlign )
								vmp_QVecAdd( pvTmp, pvA, pvB );	// Aligned
							else
								vmp_QVecAdd0( pvTmp, pvA, pvB );	// Aligned
							break;

						case iTEST_SUB:
							vmp_QVecSubGeneric( pvTmp2, pvA, pvB );
							if ( bAlign )
								vmp_QVecSub( pvTmp, pvA, pvB );	// Aligned
							else
								vmp_QVecSub0( pvTmp, pvA, pvB );
							break;

						case iTEST_ADD_SCALAR:
							vmp_QVecAddScalarGeneric( pvTmp2, pvA, fScalar );
							if ( bAlign )
								vmp_QVecAddScalar( pvTmp, pvA, fScalar );	// Aligned
							else
								vmp_QVecAddScalar0( pvTmp, pvA, fScalar );
							break;

						case iTEST_SUB_SCALAR:
							vmp_QVecSubScalarGeneric( pvTmp2, pvA, fScalar );
							if ( bAlign )
								vmp_QVecSubScalar( pvTmp, pvA, fScalar );	// Aligned
							else
								vmp_QVecSubScalar0( pvTmp, pvA, fScalar );	// Aligned
							break;
					  }

					if (!vmp_IsQVEqual( pvTmp, pvTmp2, fPrecision)) 
					  {
						cout << "TMP x:" << pvTmp2->x << " y:" << pvTmp2->y << " z:" << pvTmp2->z << " w:" << pvTmp2->w << " =\n";
						cout << " vA.x:" << pvA->x << " vA.y:" << pvA->y << " vA.z:" << pvA->z << " vA.w:" << pvA->w << "\n";
						cout << " vB.x:" << pvB->x << " vB.y:" << pvB->y << " vB.z:" << pvB->z << " vB.w:" << pvB->w << "\n"; 
						cout << "    x:" << pvTmp->x << " y:" << pvTmp->y << " z:" << pvTmp->z << " w:" << pvTmp->w << "   *** MISMATCH!!! ***\n" << endl;
					  }
				  }
			  }
		  }
	  }
}




int main( int argc, const char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

#pragma unused(argc, argv)

    CpuDetect( &cinfo );								// Detect CPU
    
    cout << "\n(Quad) Vector Addition & Subtraction Test\n\n";
    cout << CpuInfoStr( szBuf, &cinfo ) << endl;
    
	QVas3DSetup( &cinfo );								// Setup our 3D Vector stuff

	cout << "Testing (Quad) Addition Scalar" << endl;
	testQVec( false, iTEST_ADD_SCALAR );

	cout << "Testing (Quad) Addition Scalar (Aligned)" << endl;
	testQVec( true,  iTEST_ADD_SCALAR );

	cout << "Testing (Quad) Subtraction Scalar" << endl;
	testQVec( false, iTEST_SUB_SCALAR );

	cout << "Testing (Quad) Subtraction Scalar (Aligned)" << endl;
	testQVec( true,  iTEST_SUB_SCALAR );

	cout << "Testing (Quad) Addition" << endl;
	testQVec( false, iTEST_ADD );

	cout << "Testing (Quad) Addition (Aligned)" << endl;
	testQVec( true,  iTEST_ADD );

	cout << "Testing (Quad) Subtraction" << endl;
	testQVec( false, iTEST_SUB );

	cout << "Testing (Quad) Subtraction (Aligned)" << endl;
	testQVec( true,  iTEST_SUB );

	cout << "Fin\n\n" << flush;

    return 0;
}

