
// 
//	QVas3DAltivec.cpp - 3D (Quad) Vector Addition & Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include	"CpuAsm.h"									// CPU
#include	"vmp3D.h"
#include	"vmp3DEx.h"
#include	"QVas3D.h"									// 3D Math
#include	"QVas3DEx.h"



//
//	(Quad) Vector Summation (Addition) - Unaligned
//

void vmp_QVecAddAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = pvB->w;
	
	vD = vec_add( vA, vB );					// vaddfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}



//
//	(Quad) Vector Summation (Addition) - Aligned
//

void vmp_QVecAddAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);

	*(vector float *)pvD = vec_add( (*(vector float *)pvA), (*(vector float *)pvB) );		// vaddfp d,a,b
}



//______________________________________________________
//
//	(Quad) Vector Difference (Subtraction) - Unaligned
//

void vmp_QVecSubAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = pvB->w;
	
	vD = vec_sub( vA, vB );					// vsubfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}



//
//	(Quad) Vector Difference (Subtraction) - Aligned
//

void vmp_QVecSubAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);

	*(vector float *)pvD = vec_sub( (*(vector float *)pvA), (*(vector float *)pvB) );		// vsubfp d,a,b
}



//______________________________________________________
//
//	(Quad) Vector Scalar Summation (Addition) - Unaligned
//

void vmp_QVecAddScalarAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = ((vmp3DQVector *) &vB)->w = fScalar;

	vD = vec_add( vA, vB );					// vaddfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}



//
//	(Quad) Vector Scalar Summation (Addition) - Aligned
//

void vmp_QVecAddScalarAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{
	vector float vB;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = ((vmp3DQVector *) &vB)->w = fScalar;
	
	*(vector float *)pvD = vec_add( (*(vector float *)pvA), vB );	// vaddfp d,a,b
}


//______________________________________________________
//
//	(Quad) Vector Scalar Difference (Subtraction) - Unaligned
//

void vmp_QVecSubScalarAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{	
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = ((vmp3DQVector *) &vB)->w = fScalar;
	
	vD = vec_sub( vA, vB );					// vsubfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}



//
//	(Quad) Vector Scalar Difference (Subtraction) - Aligned
//

void vmp_QVecSubScalarAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar)
{	
	vector float vB;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = ((vmp3DQVector *) &vB)->w = fScalar;
	
	*(vector float *)pvD = vec_sub( (*(vector float *)pvA), vB );	// vsubfp d,a,b
}



//______________________________________________________
//
//	Setup Vec3D module
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void QVas3DSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);
	
	if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_QVecAdd =			vmp_QVecAddAltivec;
		vmp_QVecAdd0 =			vmp_QVecAddAltivec0;
		vmp_QVecSub =			vmp_QVecSubAltivec;
		vmp_QVecSub0 =			vmp_QVecSubAltivec0;
		vmp_QVecAddScalar =		vmp_QVecAddScalarAltivec;
		vmp_QVecAddScalar0 =	vmp_QVecAddScalarAltivec0;
		vmp_QVecSubScalar =		vmp_QVecSubScalarAltivec;
		vmp_QVecSubScalar0 =	vmp_QVecSubScalarAltivec0;
      }

	else
	  {
		vmp_QVecAdd =			vmp_QVecAddGeneric;
		vmp_QVecAdd0 =			vmp_QVecAddGeneric;
		vmp_QVecSub =			vmp_QVecSubGeneric;
		vmp_QVecSub0 =			vmp_QVecSubGeneric;
		vmp_QVecAddScalar =		vmp_QVecAddScalarGeneric;
		vmp_QVecAddScalar0 =	vmp_QVecAddScalarGeneric;
		vmp_QVecSubScalar =		vmp_QVecSubScalarGeneric;
		vmp_QVecSubScalar0 =	vmp_QVecSubScalarGeneric;
	  }
}




