
//
//	Vas3DAltivec.cpp - 3D Vector Addition & Subtraction
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//	NOTE: Because these a 3 float Vectors the 4th float has to be handled as 
//		a special case thus slowing the code down. This is the same as an AoS 
//		orientation. Instead a SoA should be used in conjunction with the quad
//		set of functions.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include	"CpuAsm.h"									// CPU
#include	"vmp3D.h"
#include	"vmp3DEx.h"
#include	"Vas3D.h"									// 3D Math
#include	"Vas3DEx.h"



//
//	Vector Summation (Addition) - Unaligned
//

void vmp_VecAddAltivec0(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0f;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0f;
	
	vD = vec_add( vA, vB );					// vaddfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
}


//
//	Vector Summation (Addition) - Aligned
//	NOTE: Only difference is assertion!
//

void vmp_VecAddAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0f;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0f;
	
	vD = vec_add( vA, vB );					// vaddfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
}



//______________________________________________________
//
//	Vector Difference (Subtraction) - Unaligned
//

void vmp_VecSubAltivec0(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0f;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0f;
	
	vD = vec_sub( vA, vB );					// vsubfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
}


//
//	Vector Difference (Subtraction) - Aligned
//	NOTE: Only difference is assertion!
//

void vmp_VecSubAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0f;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0f;
	
	vD = vec_sub( vA, vB );					// vsubfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
}


//______________________________________________________
//
//	Vector Scalar Summation (Addition) - Unaligned
//

void vmp_VecAddScalarAltivec0(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0f;

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = fScalar;
	((vmp3DQVector *) &vB)->w = 0.0f;

	vD = vec_add( vA, vB );					// vaddfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
}


//
//	Vector Scalar Summation (Addition) - Aligned
//	NOTE: Only difference is assertion!
//

void vmp_VecAddScalarAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0f;

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = fScalar;
	((vmp3DQVector *) &vB)->w = 0.0f;

	vD = vec_add( vA, vB );					// vaddfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
}


//______________________________________________________
//
//	Vector Scalar Difference (Subtraction) - Unaligned
//

void vmp_VecSubScalarAltivec0(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
{	
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0f;

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = fScalar;
	((vmp3DQVector *) &vB)->w = 0.0f;

	vD = vec_sub( vA, vB );					// vsubfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
}




//
//	Vector Scalar Difference (Subtraction) - Aligned
//	NOTE: Only difference is assertion!
//

void vmp_VecSubScalarAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar)
{	
	vector float vD, vA, vB;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0f;

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = fScalar;
	((vmp3DQVector *) &vB)->w = 0.0f;

	vD = vec_sub( vA, vB );					// vsubfp d,a,b

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
}




//______________________________________________________
//
//	Setup Vec3D module
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void Vas3DSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);

    if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_VecAdd =			vmp_VecAddAltivec;
		vmp_VecAdd0 =			vmp_VecAddAltivec0;
		vmp_VecSub =			vmp_VecSubAltivec;
		vmp_VecSub0 =			vmp_VecSubAltivec0;
		vmp_VecAddScalar =		vmp_VecAddScalarAltivec;
		vmp_VecAddScalar0 =		vmp_VecAddScalarAltivec0;
		vmp_VecSubScalar =		vmp_VecSubScalarAltivec;
		vmp_VecSubScalar0 =		vmp_VecSubScalarAltivec0;
      }
	else
	  {
		vmp_VecAdd =			vmp_VecAddGeneric;
		vmp_VecAdd0 =			vmp_VecAddGeneric;
		vmp_VecSub =			vmp_VecSubGeneric;
		vmp_VecSub0 =			vmp_VecSubGeneric;
		vmp_VecAddScalar =		vmp_VecAddScalarGeneric;
		vmp_VecAddScalar0 =		vmp_VecAddScalarGeneric;
		vmp_VecSubScalar =		vmp_VecSubScalarGeneric;
		vmp_VecSubScalar0 =		vmp_VecSubScalarGeneric;
	  }
}





