
//
//  Bench.cpp - (Scalar) Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include <iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"
#include "Fmd.h"
#include "FmdEx.h"



typedef enum
{
	iTEST_FLT_MUL,
	iTEST_FLT_DIV,
	iTEST_FLT_DIV_FAST
} iTEST;



//
//	Test 
//

static void testVal( iTEST iFunc )
{
	int x, z, dif;
	float f, f2, fPrecision, fA, fB;

	dif = 4;

	fPrecision = 0.0000001f;
	
	for ( z = -dif; z <= dif; z++ )
	  {
		fB = 0.123f *(float)z;

		for ( x = -dif; x <= dif; x++ )
		  {	
			fA = 0.123f *(float)x;

			if ( 0.0f == fA ) continue;
			if ( 0.0f == fB ) continue;

			switch(iFunc)
			  {
				case iTEST_FLT_MUL:
					fPrecision = 0.0000001f;
					vmp_FMulGeneric( &f2, fA, fB );
					vmp_FMul( &f, fA, fB );
					break;

				case iTEST_FLT_DIV:
					fPrecision = 0.0000001f;
					vmp_FDivGeneric( &f2, fA, fB );
					vmp_FDiv( &f, fA, fB );
					break;

				case iTEST_FLT_DIV_FAST:
					fPrecision = 0.0001f;
					vmp_FDivGeneric( &f2, fA, fB );
					vmp_FDivFast( &f, fA, fB );
					break;
			  }

			if (!vmp_IsFEqual(f, f2, fPrecision))
			  {
				cout << " " << f2 << " vs " << f << "    fA:" << fA << " fB:" << fB << "  *** MISMATCH!!! ***" << endl;
			  }
		  }
	  }
}



int main( int argc, const char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

#pragma unused(argc, argv)

    CpuDetect( &cinfo );								// Detect CPU

	cout << "\nScalar Multiplication & Division\n\n";
	cout << CpuInfoStr( szBuf, &cinfo ) << endl;

	FmdSetup( &cinfo );								// Setup (float) stuff

	cout << "Testing (Scalar) Multiplication " << endl;
	testVal( iTEST_FLT_MUL );

	cout << "Testing (Scalar) Division " << endl;
	testVal( iTEST_FLT_DIV );

	cout << "Testing (Scalar) Division *FAST (Low Precision)*" << endl;
	testVal( iTEST_FLT_DIV_FAST );

	cout << "Fin!\n\n" << flush;

    return 0;
}

