
//
//  PmdAltivec.cpp - Packed Integer Multiplication
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 31 Dec 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "Pmd.h"
#include "PmdEx.h"



static vector unsigned short V16Shift[] = {
	{ 0,0,0,0,0,0,0,0 },
	{ 1,1,1,1,1,1,1,1 },
	{ 2,2,2,2,2,2,2,2 },
	{ 3,3,3,3,3,3,3,3 },
	{ 4,4,4,4,4,4,4,4 },
	{ 5,5,5,5,5,5,5,5 },
	{ 6,6,6,6,6,6,6,6 },
	{ 7,7,7,7,7,7,7,7 },
	{ 8,8,8,8,8,8,8,8 },
	{ 9,9,9,9,9,9,9,9 },
	{ 10,10,10,10,10,10,10,10 },
	{ 11,11,11,11,11,11,11,11 },
	{ 12,12,12,12,12,12,12,12 },
	{ 13,13,13,13,13,13,13,13 },
	{ 14,14,14,14,14,14,14,14 },
	{ 15,15,15,15,15,15,15,15 }};
	
	
static vector unsigned int V32Shift[] = {
	{ 0,0,0,0 },     { 1,1,1,1 },     { 2,2,2,2 },     { 3,3,3,3 },
	{ 4,4,4,4 },     { 5,5,5,5 },     { 6,6,6,6 },     { 7,7,7,7 },
	{ 8,8,8,8 },     { 9,9,9,9 },     { 10,10,10,10 }, { 11,11,11,11 },
	{ 12,12,12,12 }, { 13,13,13,13 }, { 14,14,14,14 }, { 15,15,15,15 },
	{ 16,16,16,16 }, { 17,17,17,17 }, { 18,18,18,18 }, { 19,19,19,19 },
	{ 20,20,20,20 }, { 21,21,21,21 }, { 22,22,22,22 }, { 23,23,23,23 },
	{ 24,24,24,24 }, { 25,25,25,25 }, { 26,26,26,26 }, { 27,27,27,27 },
	{ 28,28,28,28 }, { 29,29,29,29 }, { 30,30,30,30 }, { 31,31,31,31 }};




//
//	Packed signed Multiplication High(16bit) = 8x16bit  (Aligned)
//
void vmp_pmulhHAltivec( int16 * const phD, const int16 * const phA, const int16 * const phB )
{
	vector signed int ie, io, ihi, ilo;
	  
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	ie = vec_mule( (*(vector signed short *)phA), (*(vector signed short *)phB) );	// [e0 e2 e4 e6]
	io = vec_mulo( (*(vector signed short *)phA), (*(vector signed short *)phB) );	// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            			// [e0 e2 e4 e6] [ab0 ab1 ab2 ab3]
	ilo = vec_mergel(ie, io);	            			// [o1 o3 o5 o7] [ab4 ab5 ab6 ab7]

  	ihi = vec_sr(ihi, V32Shift[16]);					// [Ue0, Ue2, Ue4, Ue6]
  	ilo = vec_sr(ilo, V32Shift[16]);					// [Uo1, Uo3, Uo5, Uo7]

    *((vector signed short *)phD) = vec_pack(ihi, ilo); // [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7]
}									



//
//	Packed signed Multiplication High(16bit) = 8x16bit  (Unaligned)
//
void vmp_pmulhHAltivec0( int16 * const phD, const int16 * const phA, const int16 * const phB )
{
	vector signed int ie, io, ihi, ilo;
	vector signed short hA, hB;
	  
	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);

	ie = vec_mule( hA, hB );				// [e0 e2 e4 e6]
	io = vec_mulo( hA, hB );				// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

  	ihi = vec_sr(ihi, V32Shift[16]);		// [Ue0, Ue2, Ue4, Ue6]
  	ilo = vec_sr(ilo, V32Shift[16]);		// [Uo1, Uo3, Uo5, Uo7]

    hA = vec_pack(ihi, ilo); 				// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7]

	*(((uint32 *)phD)+0) = *(((uint32 *)&hA)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&hA)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&hA)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&hA)+3);								
}



//
//	Packed unsigned Multiplication High(16bit) = 8x16bit  (Aligned)
//
void vmp_pmulhuHAltivec( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned int ie, io, ihi, ilo;
	  
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	ie = vec_mule( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) );	// [e0 e2 e4 e6]
	io = vec_mulo( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) );	// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

  	ihi = vec_sr(ihi, V32Shift[16]);		// [Ue0, Ue2, Ue4, Ue6]
  	ilo = vec_sr(ilo, V32Shift[16]);		// [Uo1, Uo3, Uo5, Uo7]

    *((vector unsigned short *)phD) = vec_pack(ihi, ilo); // [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7]
}



//
//	Packed unsigned Multiplication High(16bit) = 8x16bit  (Unsligned)
//
void vmp_pmulhuHAltivec0( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned int ie, io, ihi, ilo;
	vector unsigned short hA, hB;
	  
	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);

	ie = vec_mule( hA, hB );				// [e0 e2 e4 e6]
	io = vec_mulo( hA, hB );				// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

  	ihi = vec_sr(ihi, V32Shift[16]);		// [Ue0, Ue2, Ue4, Ue6]
  	ilo = vec_sr(ilo, V32Shift[16]);		// [Uo1, Uo3, Uo5, Uo7]

    hA = vec_pack(ihi, ilo); 				// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7]

	*(((uint32 *)phD)+0) = *(((uint32 *)&hA)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&hA)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&hA)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&hA)+3);								
}



//
//	Packed (un)signed Multiplication Low(16bit) = 8x16bit  (Aligned)
//
void vmp_pmulluHAltivec( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned int ie, io, ihi, ilo;
	  
	ASSERT_PTR16(phD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	ie = vec_mule( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) );	// [e0 e2 e4 e6]
	io = vec_mulo( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) );	// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

    *((vector unsigned short *)phD) = vec_pack(ihi, ilo); // [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7]
}



//
//	Packed (un)signed Multiplication Low(16bit) = 8x16bit  (Unaligned)
//
void vmp_pmulluHAltivec0( uint16 * const phD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned int ie, io, ihi, ilo;
	vector unsigned short hA, hB;
	  
	ASSERT_PTR4(phD);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);

	ie = vec_mule( hA, hB );				// [e0 e2 e4 e6]
	io = vec_mulo( hA, hB );				// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

    hA = vec_pack(ihi, ilo); 				// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7]

	*(((uint32 *)phD)+0) = *(((uint32 *)&hA)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&hA)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&hA)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&hA)+3);								
}

//______________________________________________________


//
//	Packed signed Multiplication High(8bit) = 16x8bit  (Aligned)
//
void vmp_pmulhBAltivec( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
{
	vector signed short ie, io, ihi, ilo;
	  
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	ie = vec_mule( (*(vector signed char *)pbA), (*(vector signed char *)pbB) );	// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( (*(vector signed char *)pbA), (*(vector signed char *)pbB) );	// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            			// [e0 e2 e4 e6 e8 eA eC eE] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            			// [o1 o3 o5 o7 o9 oB oD oF] [AB4 AB5 AB6 AB7]

  	ihi = vec_sr(ihi, V16Shift[8]);						// [Ue0, Ue2, Ue4, Ue6, Ue8, UeA, UeC, UeE]
  	ilo = vec_sr(ilo, V16Shift[8]);						// [Uo1, Uo3, Uo5, Uo7, Uo9, UoB, UoD, UoF]

    *((vector signed char *)pbD) = vec_pack(ihi, ilo); // [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7 Ue8 Uo9 UeA UoB UeC UoD UeE UoF]
}									



//
//	Packed signed Multiplication High(8bit) = 16x8bit  (Unaligned)
//
void vmp_pmulhBAltivec0( int8 * const pbD, const int8 * const pbA, const int8 * const pbB )
{
	vector signed short ie, io, ihi, ilo;
	vector signed char bA, bB;
	  
	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&bA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&bA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&bA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&bA)+3) = *(((uint32 *)pbA)+3);
	
	*(((uint32 *)&bB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&bB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&bB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&bB)+3) = *(((uint32 *)pbB)+3);

	ie = vec_mule( bA, bB );				// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( bA, bB );				// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6 e8 eA eC eE] [ab0 ab1 ab2 ab3 ab4 ab5 ab6 ab7]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7 o9 oB oD oF] [ab8 ab9 abA abB abC abD abE abF]

  	ihi = vec_sr(ihi, V16Shift[8]);			// [Ue0, Ue2, Ue4, Ue6, Ue8, UeA, UeC, UeE]
  	ilo = vec_sr(ilo, V16Shift[8]);			// [Uo1, Uo3, Uo5, Uo7, Uo9, UoB, UoD, UoF]

    bA = vec_pack(ihi, ilo); 				// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7 Ue8 Uo9 UeA UoB UeC UoD UeE UoF]

	*(((uint32 *)pbD)+0) = *(((uint32 *)&bA)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&bA)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&bA)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&bA)+3);								
}



//
//	Packed unsigned Multiplication High(8bit) = 16x8bit  (Aligned)
//
void vmp_pmulhuBAltivec( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned short ie, io, ihi, ilo;
	  
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	ie = vec_mule( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) );	// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) );	// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            			// [e0 e2 e4 e6 e8 eA eC eE] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            			// [o1 o3 o5 o7 o9 oB oD oF] [AB4 AB5 AB6 AB7]

  	ihi = vec_sr(ihi, V16Shift[8]);						// [Ue0, Ue2, Ue4, Ue6, Ue8, UeA, UeC, UeE]
  	ilo = vec_sr(ilo, V16Shift[8]);						// [Uo1, Uo3, Uo5, Uo7, Uo9, UoB, UoD, UoF]

    *((vector unsigned char *)pbD) = vec_pack(ihi, ilo); // [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7 Ue8 Uo9 UeA UoB UeC UoD UeE UoF]
}									



//
//	Packed unsigned Multiplication High(8bit) = 16x8bit  (Unaligned)
//
void vmp_pmulhuBAltivec0( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned short ie, io, ihi, ilo;
	vector unsigned char bA, bB;
	  
	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&bA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&bA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&bA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&bA)+3) = *(((uint32 *)pbA)+3);
	
	*(((uint32 *)&bB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&bB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&bB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&bB)+3) = *(((uint32 *)pbB)+3);

	ie = vec_mule( bA, bB );				// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( bA, bB );				// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6 e8 eA eC eE] [ab0 ab1 ab2 ab3 ab4 ab5 ab6 ab7]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7 o9 oB oD oF] [ab8 ab9 abA abB abC abD abE abF]

  	ihi = vec_sr(ihi, V16Shift[8]);			// [Ue0, Ue2, Ue4, Ue6, Ue8, UeA, UeC, UeE]
  	ilo = vec_sr(ilo, V16Shift[8]);			// [Uo1, Uo3, Uo5, Uo7, Uo9, UoB, UoD, UoF]

    bA = vec_pack(ihi, ilo); 				// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7 Ue8 Uo9 UeA UoB UeC UoD UeE UoF]

	*(((uint32 *)pbD)+0) = *(((uint32 *)&bA)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&bA)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&bA)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&bA)+3);								
}



//
//	Packed (un)signed Multiplication Low(8bit) = 16x8bit  (Aligned)
//
void vmp_pmulluBAltivec( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned short ie, io, ihi, ilo;
	  
	ASSERT_PTR16(pbD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	ie = vec_mule( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) );	// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) );	// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6 e8 eA eC eE] [ab0 ab1 ab2 ab3 ab4 ab5 ab6 ab7]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7 o9 oB oD oF] [ab8 ab9 abA abB abC abD abE abF]

    *((vector unsigned char *)pbD) = vec_pack(ihi, ilo); // [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7 Le8 Lo9 LeA LoB LeC LoD LeE LoF]
}



//
//	Packed (un)signed Multiplication Low(8bit) = 16x8bit  (Unaligned)
//
void vmp_pmulluBAltivec0( uint8 * const pbD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned short ie, io, ihi, ilo;
	vector unsigned char bA, bB;
	  
	ASSERT_PTR4(pbD);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(pbB);

	*(((uint32 *)&bA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&bA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&bA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&bA)+3) = *(((uint32 *)pbA)+3);
	
	*(((uint32 *)&bB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&bB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&bB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&bB)+3) = *(((uint32 *)pbB)+3);

	ie = vec_mule( bA, bB );				// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( bA, bB );				// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6 e8 eA eC eE] [ab0 ab1 ab2 ab3 ab4 ab5 ab6 ab7]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7 o9 oB oD oF] [ab8 ab9 abA abB abC abD abE abF]

    bA = vec_pack(ihi, ilo); 				// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7 Le8 Lo9 LeA LoB LeC LoD LeE LoF]

	*(((uint32 *)pbD)+0) = *(((uint32 *)&bA)+0);
	*(((uint32 *)pbD)+1) = *(((uint32 *)&bA)+1);
	*(((uint32 *)pbD)+2) = *(((uint32 *)&bA)+2);
	*(((uint32 *)pbD)+3) = *(((uint32 *)&bA)+3);								
}


//______________________________________________________


//
//	Packed signed Multiplication Hi/Lo(2x8x16bit) = 8x16bit  (Aligned)
//
void vmp_pmulhlHAltivec( int16 * const phHi, int16 * const phLo, const int16 * const phA, const int16 * const phB )
{
	vector signed int ie, io, ihi, ilo;
	  
	ASSERT_PTR16(phHi);
	ASSERT_PTR16(phLo);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	ie = vec_mule( (*(vector signed short *)phA), (*(vector signed short *)phB) );	// [e0 e2 e4 e6]
	io = vec_mulo( (*(vector signed short *)phA), (*(vector signed short *)phB) );	// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            				// [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            				// [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

    *((vector signed short *)phLo) = vec_pack(ihi, ilo);	// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7]

  	ihi = vec_sr(ihi, V32Shift[16]);						// [Ue0, Ue2, Ue4, Ue6]
  	ilo = vec_sr(ilo, V32Shift[16]);						// [Uo1, Uo3, Uo5, Uo7]

    *((vector signed short *)phHi) = vec_pack(ihi, ilo); 	// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7]
}									



//
//	Packed signed Multiplication Hi/Lo(2x8x16bit) = 8x16bit  (Unaligned)
//
void vmp_pmulhlHAltivec0( int16 * const phHi, int16 * const phLo, const int16 * const phA, const int16 * const phB )
{
	vector signed int ie, io, ihi, ilo;
	vector signed short hA, hB;
	  
	ASSERT_PTR4(phHi);
	ASSERT_PTR4(phLo);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);

	ie = vec_mule( hA, hB );				// [e0 e2 e4 e6]
	io = vec_mulo( hA, hB );				// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

    hA = vec_pack(ihi, ilo); 				// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7]

  	ihi = vec_sr(ihi, V32Shift[16]);		// [Ue0, Ue2, Ue4, Ue6]
  	ilo = vec_sr(ilo, V32Shift[16]);		// [Uo1, Uo3, Uo5, Uo7]

    hB = vec_pack(ihi, ilo); 				// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7]

	*(((uint32 *)phLo)+0) = *(((uint32 *)&hA)+0);
	*(((uint32 *)phLo)+1) = *(((uint32 *)&hA)+1);
	*(((uint32 *)phLo)+2) = *(((uint32 *)&hA)+2);
	*(((uint32 *)phLo)+3) = *(((uint32 *)&hA)+3);								

	*(((uint32 *)phHi)+0) = *(((uint32 *)&hB)+0);
	*(((uint32 *)phHi)+1) = *(((uint32 *)&hB)+1);
	*(((uint32 *)phHi)+2) = *(((uint32 *)&hB)+2);
	*(((uint32 *)phHi)+3) = *(((uint32 *)&hB)+3);
}



//
//	Packed unsigned Multiplication Hi/Lo(2x8x16bit) = 8x16bit  (Aligned)
//
void vmp_pmulhluHAltivec( uint16 * const phHi, uint16 * const phLo, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned int ie, io, ihi, ilo;
	  
	ASSERT_PTR16(phHi);
	ASSERT_PTR16(phLo);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	ie = vec_mule( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) );	// [e0 e2 e4 e6]
	io = vec_mulo( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) );	// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            				// [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            				// [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

    *((vector unsigned short *)phLo) = vec_pack(ihi, ilo);	// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7]

  	ihi = vec_sr(ihi, V32Shift[16]);						// [Ue0, Ue2, Ue4, Ue6]
  	ilo = vec_sr(ilo, V32Shift[16]);						// [Uo1, Uo3, Uo5, Uo7]

    *((vector unsigned short *)phHi) = vec_pack(ihi, ilo); 	// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7]
}									



//
//	Packed unsigned Multiplication Hi/Lo(2x8x16bit) = 8x16bit  (Unaligned)
//
void vmp_pmulhluHAltivec0( uint16 * const phHi, uint16 * const phLo, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned int ie, io, ihi, ilo;
	vector unsigned short hA, hB;
	  
	ASSERT_PTR4(phHi);
	ASSERT_PTR4(phLo);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);

	ie = vec_mule( hA, hB );				// [e0 e2 e4 e6]
	io = vec_mulo( hA, hB );				// [o1 o3 o5 o7]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7] [AB4 AB5 AB6 AB7]

    hA = vec_pack(ihi, ilo); 				// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7]

  	ihi = vec_sr(ihi, V32Shift[16]);		// [Ue0, Ue2, Ue4, Ue6]
  	ilo = vec_sr(ilo, V32Shift[16]);		// [Uo1, Uo3, Uo5, Uo7]

    hB = vec_pack(ihi, ilo); 				// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7]

	*(((uint32 *)phLo)+0) = *(((uint32 *)&hA)+0);
	*(((uint32 *)phLo)+1) = *(((uint32 *)&hA)+1);
	*(((uint32 *)phLo)+2) = *(((uint32 *)&hA)+2);
	*(((uint32 *)phLo)+3) = *(((uint32 *)&hA)+3);								

	*(((uint32 *)phHi)+0) = *(((uint32 *)&hB)+0);
	*(((uint32 *)phHi)+1) = *(((uint32 *)&hB)+1);
	*(((uint32 *)phHi)+2) = *(((uint32 *)&hB)+2);
	*(((uint32 *)phHi)+3) = *(((uint32 *)&hB)+3);
}


//______________________________________________________


//
//	Packed signed Multiplication Hi/Lo(2x16x8bit) = 16x8bit  (Aligned)
//
void vmp_pmulhlBAltivec( int8 * const phHi, int8 * const phLo, const int8 * const phA, const int8 * const phB )
{
	vector signed short ie, io, ihi, ilo;
	  
	ASSERT_PTR16(phHi);
	ASSERT_PTR16(phLo);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	ie = vec_mule( (*(vector signed char *)phA), (*(vector signed char *)phB) );	// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( (*(vector signed char *)phA), (*(vector signed char *)phB) );	// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            				// [e0 e2 e4 e6 e8 eA eC eE] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            				// [o1 o3 o5 o7 o9 oB oD oF] [AB4 AB5 AB6 AB7]

    *((vector signed char *)phLo) = vec_pack(ihi, ilo);		// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7 Le8 Lo9 LeA LoB LeC LoD LeE LoF]

  	ihi = vec_sr(ihi, V16Shift[8]);							// [Ue0, Ue2, Ue4, Ue6, Ue8, UeA, UeC, UeE]
  	ilo = vec_sr(ilo, V16Shift[8]);							// [Uo1, Uo3, Uo5, Uo7, Uo9, UoB, UoD, UoF]

    *((vector signed char *)phHi) = vec_pack(ihi, ilo); 	// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7 Ue8 Uo9 UeA UoB UeC UoD UeE UoF]
}									



//
//	Packed signed Multiplication Hi/Lo(2x16x8bit) = 16x8bit  (Unaligned)
//
void vmp_pmulhlBAltivec0( int8 * const phHi, int8 * const phLo, const int8 * const phA, const int8 * const phB )
{
	vector signed short ie, io, ihi, ilo;
	vector signed char hA, hB;
	  
	ASSERT_PTR4(phHi);
	ASSERT_PTR4(phLo);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);

	ie = vec_mule( hA, hB );				// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( hA, hB );				// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6 e8 eA eC eE] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7 o9 oB oD oF] [AB4 AB5 AB6 AB7]

    hA = vec_pack(ihi, ilo); 				// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7 Le8 Lo9 LeA LoB LeC LoD LeE LoF]

  	ihi = vec_sr(ihi, V16Shift[8]);			// [Ue0, Ue2, Ue4, Ue6, Ue8, UeA, UeC, UeE]
  	ilo = vec_sr(ilo, V16Shift[8]);			// [Uo1, Uo3, Uo5, Uo7, Uo9, UoB, UoD, UoF]

    hB = vec_pack(ihi, ilo); 				// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7 Ue8 Uo9 UeA UoB UeC UoD UeE UoF]

	*(((uint32 *)phLo)+0) = *(((uint32 *)&hA)+0);
	*(((uint32 *)phLo)+1) = *(((uint32 *)&hA)+1);
	*(((uint32 *)phLo)+2) = *(((uint32 *)&hA)+2);
	*(((uint32 *)phLo)+3) = *(((uint32 *)&hA)+3);								

	*(((uint32 *)phHi)+0) = *(((uint32 *)&hB)+0);
	*(((uint32 *)phHi)+1) = *(((uint32 *)&hB)+1);
	*(((uint32 *)phHi)+2) = *(((uint32 *)&hB)+2);
	*(((uint32 *)phHi)+3) = *(((uint32 *)&hB)+3);
}



//
//	Packed unsigned Multiplication Hi/Lo(2x16x8bit) = 16x8bit  (Aligned)
//
void vmp_pmulhluBAltivec( uint8 * const phHi, uint8 * const phLo, const uint8 * const phA, const uint8 * const phB )
{
	vector unsigned short ie, io, ihi, ilo;
	  
	ASSERT_PTR16(phHi);
	ASSERT_PTR16(phLo);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	ie = vec_mule( (*(vector unsigned char *)phA), (*(vector unsigned char *)phB) );	// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( (*(vector unsigned char *)phA), (*(vector unsigned char *)phB) );	// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            				// [e0 e2 e4 e6 e8 eA eC eE] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            				// [o1 o3 o5 o7 o9 oB oD oF] [AB4 AB5 AB6 AB7]

    *((vector unsigned char *)phLo) = vec_pack(ihi, ilo);		// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7 Le8 Lo9 LeA LoB LeC LoD LeE LoF]

  	ihi = vec_sr(ihi, V16Shift[8]);							// [Ue0, Ue2, Ue4, Ue6, Ue8, UeA, UeC, UeE]
  	ilo = vec_sr(ilo, V16Shift[8]);							// [Uo1, Uo3, Uo5, Uo7, Uo9, UoB, UoD, UoF]

    *((vector unsigned char *)phHi) = vec_pack(ihi, ilo); 	// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7 Ue8 Uo9 UeA UoB UeC UoD UeE UoF]
}									



//
//	Packed unsigned Multiplication Hi/Lo(2x16x8bit) = 16x8bit  (Unaligned)
//
void vmp_pmulhluBAltivec0( uint8 * const phHi, uint8 * const phLo, const uint8 * const phA, const uint8 * const phB )
{
	vector unsigned short ie, io, ihi, ilo;
	vector unsigned char hA, hB;
	  
	ASSERT_PTR4(phHi);
	ASSERT_PTR4(phLo);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(phB);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);

	ie = vec_mule( hA, hB );				// [e0 e2 e4 e6 e8 eA eC eE]
	io = vec_mulo( hA, hB );				// [o1 o3 o5 o7 o9 oB oD oF]

	ihi = vec_mergeh(ie, io);	            // [e0 e2 e4 e6 e8 eA eC eE] [AB0 AB1 AB2 AB3]
	ilo = vec_mergel(ie, io);	            // [o1 o3 o5 o7 o9 oB oD oF] [AB4 AB5 AB6 AB7]

    hA = vec_pack(ihi, ilo); 				// [Le0 Lo1 Le2 Lo3 Le4 Lo5 Le6 Lo7 Le8 Lo9 LeA LoB LeC LoD LeE LoF]

  	ihi = vec_sr(ihi, V16Shift[8]);			// [Ue0, Ue2, Ue4, Ue6, Ue8, UeA, UeC, UeE]
  	ilo = vec_sr(ilo, V16Shift[8]);			// [Uo1, Uo3, Uo5, Uo7, Uo9, UoB, UoD, UoF]

    hB = vec_pack(ihi, ilo); 				// [Ue0 Uo1 Ue2 Uo3 Ue4 Uo5 Ue6 Uo7 Ue8 Uo9 UeA UoB UeC UoD UeE UoF]

	*(((uint32 *)phLo)+0) = *(((uint32 *)&hA)+0);
	*(((uint32 *)phLo)+1) = *(((uint32 *)&hA)+1);
	*(((uint32 *)phLo)+2) = *(((uint32 *)&hA)+2);
	*(((uint32 *)phLo)+3) = *(((uint32 *)&hA)+3);								

	*(((uint32 *)phHi)+0) = *(((uint32 *)&hB)+0);
	*(((uint32 *)phHi)+1) = *(((uint32 *)&hB)+1);
	*(((uint32 *)phHi)+2) = *(((uint32 *)&hB)+2);
	*(((uint32 *)phHi)+3) = *(((uint32 *)&hB)+3);
}



//______________________________________________________


//
//	Packed unsigned Multiplication 16bit Even = 16x8bit  (Aligned)
//
void vmp_pmullueBAltivec( uint16 * const phD, const uint8 * const pbA, const uint8 * const pbB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector unsigned short *)phD = vec_mule( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) );
}									



//
//	Packed unsigned Multiplication 16bit Even = 16x8bit  (Unaligned)
//
void vmp_pmullueBAltivec0( uint16 * const phD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned short hD;
	vector unsigned char bA, bB;

	ASSERT_PTR4(pbB);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(phD);

	*(((uint32 *)&bA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&bA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&bA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&bA)+3) = *(((uint32 *)pbA)+3);
	
	*(((uint32 *)&bB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&bB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&bB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&bB)+3) = *(((uint32 *)pbB)+3);
	
	hD = vec_mule( bA, bB );

	*(((uint32 *)phD)+0) = *(((uint32 *)&hD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&hD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&hD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&hD)+3);
}



//
//	Packed unsigned Multiplication 16bit Odd = 16x8bit  (Aligned)
//
void vmp_pmulluoBAltivec( uint16 * const phD, const uint8 * const pbA, const uint8 * const pbB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector unsigned short *)phD = vec_mulo( (*(vector unsigned char *)pbA), (*(vector unsigned char *)pbB) );

}									



//
//	Packed unsigned Multiplication 16bit Odd = 16x8bit  (Unaligned)
//
void vmp_pmulluoBAltivec0( uint16 * const phD, const uint8 * const pbA, const uint8 * const pbB )
{
	vector unsigned short hD;
	vector unsigned char bA, bB;

	ASSERT_PTR4(pbB);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(phD);

	*(((uint32 *)&bA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&bA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&bA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&bA)+3) = *(((uint32 *)pbA)+3);
	
	*(((uint32 *)&bB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&bB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&bB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&bB)+3) = *(((uint32 *)pbB)+3);
	
	hD = vec_mulo( bA, bB );

	*(((uint32 *)phD)+0) = *(((uint32 *)&hD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&hD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&hD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&hD)+3);
}



//
//	Packed signed Multiplication 16bit Even = 16x8bit  (Aligned)
//
void vmp_pmulleBAltivec( int16 * const phD, const int8 * const pbA, const int8 * const pbB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector signed short *)phD = vec_mule( (*(vector signed char *)pbA), (*(vector signed char *)pbB) );
}		



//
//	Packed signed Multiplication 16bit Even = 16x8bit  (Unaligned)
//
void vmp_pmulleBAltivec0( int16 * const phD, const int8 * const pbA, const int8 * const pbB )
{
	vector signed short hD;
	vector signed char bA, bB;

	ASSERT_PTR4(pbB);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(phD);

	*(((uint32 *)&bA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&bA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&bA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&bA)+3) = *(((uint32 *)pbA)+3);
	
	*(((uint32 *)&bB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&bB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&bB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&bB)+3) = *(((uint32 *)pbB)+3);
	
	hD = vec_mule( bA, bB );

	*(((uint32 *)phD)+0) = *(((uint32 *)&hD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&hD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&hD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&hD)+3);
}



//
//	Packed signed Multiplication 16bit Odd = 16x8bit  (Aligned)
//
void vmp_pmulloBAltivec( int16 * const phD, const int8 * const pbA, const int8 * const pbB )
{
	ASSERT_PTR16(phD);
	ASSERT_PTR16(pbA);
	ASSERT_PTR16(pbB);

	*(vector signed short *)phD = vec_mulo( (*(vector signed char *)pbA), (*(vector signed char *)pbB) );
}	


//
//	Packed signed Multiplication 16bit Odd = 16x8bit  (Unaligned)
//
void vmp_pmulloBAltivec0( int16 * const phD, const int8 * const pbA, const int8 * const pbB )
{
	vector signed short hD;
	vector signed char bA, bB;

	ASSERT_PTR4(pbB);
	ASSERT_PTR4(pbA);
	ASSERT_PTR4(phD);

	*(((uint32 *)&bA)+0) = *(((uint32 *)pbA)+0);
	*(((uint32 *)&bA)+1) = *(((uint32 *)pbA)+1);
	*(((uint32 *)&bA)+2) = *(((uint32 *)pbA)+2);
	*(((uint32 *)&bA)+3) = *(((uint32 *)pbA)+3);
	
	*(((uint32 *)&bB)+0) = *(((uint32 *)pbB)+0);
	*(((uint32 *)&bB)+1) = *(((uint32 *)pbB)+1);
	*(((uint32 *)&bB)+2) = *(((uint32 *)pbB)+2);
	*(((uint32 *)&bB)+3) = *(((uint32 *)pbB)+3);
	
	hD = vec_mulo( bA, bB );

	*(((uint32 *)phD)+0) = *(((uint32 *)&hD)+0);
	*(((uint32 *)phD)+1) = *(((uint32 *)&hD)+1);
	*(((uint32 *)phD)+2) = *(((uint32 *)&hD)+2);
	*(((uint32 *)phD)+3) = *(((uint32 *)&hD)+3);
}



//______________________________________________________



//
//	Packed unsigned Multiplication 32bit Even = 8x16bit  (Aligned)
//
void vmp_pmullueHAltivec( uint32 * const pwD, const uint16 * const phA, const uint16 * const phB )
{
	ASSERT_PTR16(pwD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector unsigned int *)pwD = vec_mule( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) );
}									



//
//	Packed unsigned Multiplication 32bit Even = 8x16bit  (Unaligned)
//
void vmp_pmullueHAltivec0( uint32 * const pwD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned int wD;
	vector unsigned short hA, hB;

	ASSERT_PTR4(phB);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(pwD);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);
	
	wD = vec_mule( hA, hB );

	*(((uint32 *)pwD)+0) = *(((uint32 *)&wD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&wD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&wD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&wD)+3);
}



//
//	Packed unsigned Multiplication 32bit Odd = 8x16bit  (Aligned)
//
void vmp_pmulluoHAltivec( uint32 * const pwD, const uint16 * const phA, const uint16 * const phB )
{
	ASSERT_PTR16(pwD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector unsigned int *)pwD = vec_mulo( (*(vector unsigned short *)phA), (*(vector unsigned short *)phB) );
}									



//
//	Packed unsigned Multiplication 32bit Odd = 8x16bit  (Unaligned)
//
void vmp_pmulluoHAltivec0( uint32 * const pwD, const uint16 * const phA, const uint16 * const phB )
{
	vector unsigned int wD;
	vector unsigned short hA, hB;

	ASSERT_PTR4(phB);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(pwD);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);
	
	wD = vec_mulo( hA, hB );

	*(((uint32 *)pwD)+0) = *(((uint32 *)&wD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&wD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&wD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&wD)+3);
}



//
//	Packed signed Multiplication 32bit Even = 8x16bit  (Aligned)
//
void vmp_pmulleHAltivec( int32 * const pwD, const int16 * const phA, const int16 * const phB )
{
	ASSERT_PTR16(pwD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector signed int *)pwD = vec_mule( (*(vector signed short *)phA), (*(vector signed short *)phB) );
}									



//
//	Packed signed Multiplication 32bit Even = 8x16bit  (Unaligned)
//
void vmp_pmulleHAltivec0( int32 * const pwD, const int16 * const phA, const int16 * const phB )
{
	vector signed int wD;
	vector signed short hA, hB;

	ASSERT_PTR4(phB);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(pwD);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);
	
	wD = vec_mule( hA, hB );

	*(((uint32 *)pwD)+0) = *(((uint32 *)&wD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&wD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&wD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&wD)+3);
}



//
//	Packed signed Multiplication 32bit Odd = 8x16bit  (Aligned)
//
void vmp_pmulloHAltivec( int32 * const pwD, const int16 * const phA, const int16 * const phB )
{
	ASSERT_PTR16(pwD);
	ASSERT_PTR16(phA);
	ASSERT_PTR16(phB);

	*(vector signed int *)pwD = vec_mulo( (*(vector signed short *)phA), (*(vector signed short *)phB) );
}									



//
//	Packed signed Multiplication 32bit Odd = 8x16bit  (Unaligned)
//
void vmp_pmulloHAltivec0( int32 * const pwD, const int16 * const phA, const int16 * const phB )
{
	vector signed int wD;
	vector signed short hA, hB;

	ASSERT_PTR4(phB);
	ASSERT_PTR4(phA);
	ASSERT_PTR4(pwD);

	*(((uint32 *)&hA)+0) = *(((uint32 *)phA)+0);
	*(((uint32 *)&hA)+1) = *(((uint32 *)phA)+1);
	*(((uint32 *)&hA)+2) = *(((uint32 *)phA)+2);
	*(((uint32 *)&hA)+3) = *(((uint32 *)phA)+3);
	
	*(((uint32 *)&hB)+0) = *(((uint32 *)phB)+0);
	*(((uint32 *)&hB)+1) = *(((uint32 *)phB)+1);
	*(((uint32 *)&hB)+2) = *(((uint32 *)phB)+2);
	*(((uint32 *)&hB)+3) = *(((uint32 *)phB)+3);
	
	wD = vec_mulo( hA, hB );

	*(((uint32 *)pwD)+0) = *(((uint32 *)&wD)+0);
	*(((uint32 *)pwD)+1) = *(((uint32 *)&wD)+1);
	*(((uint32 *)pwD)+2) = *(((uint32 *)&wD)+2);
	*(((uint32 *)pwD)+3) = *(((uint32 *)&wD)+3);
}



//
//	Setup Packed Multiplication/Division
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void PmdSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);
	    
	if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_pmulhH =			vmp_pmulhHAltivec;
		vmp_pmulhH0 =			vmp_pmulhHAltivec0;
		vmp_pmulhuH =			vmp_pmulhuHAltivec;
		vmp_pmulhuH0 =			vmp_pmulhuHAltivec0;
		vmp_pmulluH =			vmp_pmulluHAltivec;
		vmp_pmulluH0 =			vmp_pmulluHAltivec0;

	//	Multiplication Hi/Lo

		vmp_pmulhlH =			vmp_pmulhlHAltivec;
		vmp_pmulhlH0 =			vmp_pmulhlHAltivec0;
		vmp_pmulhluH =			vmp_pmulhluHAltivec;
		vmp_pmulhluH0 =			vmp_pmulhluHAltivec0;

	//	Multiplication 16bit Odd/Even

		vmp_pmullueH =			vmp_pmullueHAltivec;
		vmp_pmullueH0 =			vmp_pmullueHAltivec0;
		vmp_pmulluoH =			vmp_pmulluoHAltivec;
		vmp_pmulluoH0 =			vmp_pmulluoHAltivec0;
		vmp_pmulleH =			vmp_pmulleHAltivec;
		vmp_pmulleH0 =			vmp_pmulleHAltivec0;
		vmp_pmulloH =			vmp_pmulloHAltivec;
		vmp_pmulloH0 =			vmp_pmulloHAltivec0;

	//	Multiplication 8bit Odd/Even

		vmp_pmullueB =			vmp_pmullueBAltivec;
		vmp_pmullueB0 =			vmp_pmullueBAltivec0;
		vmp_pmulluoB =			vmp_pmulluoBAltivec;
		vmp_pmulluoB0 =			vmp_pmulluoBAltivec0;
		vmp_pmulleB =			vmp_pmulleBAltivec;
		vmp_pmulleB0 =			vmp_pmulleBAltivec0;
		vmp_pmulloB =			vmp_pmulloBAltivec;
		vmp_pmulloB0 =			vmp_pmulloBAltivec0;

	//	Multiplication Hi/Lo

		vmp_pmulhlB =			vmp_pmulhlBAltivec;
		vmp_pmulhlB0 =			vmp_pmulhlBAltivec0;
		vmp_pmulhluB =			vmp_pmulhluBAltivec;
		vmp_pmulhluB0 =			vmp_pmulhluBAltivec0;

		vmp_pmulhB =			vmp_pmulhBAltivec;
		vmp_pmulhB0 =			vmp_pmulhBAltivec0;
		vmp_pmulhuB =			vmp_pmulhuBAltivec;
		vmp_pmulhuB0 =			vmp_pmulhuBAltivec0;
		vmp_pmulluB =			vmp_pmulluBAltivec;
		vmp_pmulluB0 =			vmp_pmulluBAltivec0;
      }

	else
	  {
		vmp_pmulhH =			vmp_pmulhHGeneric;
		vmp_pmulhH0 =			vmp_pmulhHGeneric;
		vmp_pmulhuH =			vmp_pmulhuHGeneric;
		vmp_pmulhuH0 =			vmp_pmulhuHGeneric;
		vmp_pmulluH =			vmp_pmulluHGeneric;
		vmp_pmulluH0 =			vmp_pmulluHGeneric;

	//	Multiplication Hi/Lo

		vmp_pmulhlH =			vmp_pmulhlHGeneric;
		vmp_pmulhlH0 =			vmp_pmulhlHGeneric;
		vmp_pmulhluH =			vmp_pmulhluHGeneric;
		vmp_pmulhluH0 =			vmp_pmulhluHGeneric;

	//	Multiplication Odd/Even

		vmp_pmullueH =			vmp_pmullueHGeneric;
		vmp_pmullueH0 =			vmp_pmullueHGeneric;
		vmp_pmulluoH =			vmp_pmulluoHGeneric;
		vmp_pmulluoH0 =			vmp_pmulluoHGeneric;
		vmp_pmulleH =			vmp_pmulleHGeneric;
		vmp_pmulleH0 =			vmp_pmulleHGeneric;
		vmp_pmulloH =			vmp_pmulloHGeneric;
		vmp_pmulloH0 =			vmp_pmulloHGeneric;

	//	Multiplication Odd/Even

		vmp_pmullueB =			vmp_pmullueBGeneric;
		vmp_pmullueB0 =			vmp_pmullueBGeneric;
		vmp_pmulluoB =			vmp_pmulluoBGeneric;
		vmp_pmulluoB0 =			vmp_pmulluoBGeneric;
		vmp_pmulleB =			vmp_pmulleBGeneric;
		vmp_pmulleB0 =			vmp_pmulleBGeneric;
		vmp_pmulloB =			vmp_pmulloBGeneric;
		vmp_pmulloB0 =			vmp_pmulloBGeneric;

	//	Multiplication Hi/Lo

		vmp_pmulhlB =			vmp_pmulhlBGeneric;
		vmp_pmulhlB0 =			vmp_pmulhlBGeneric;
		vmp_pmulhluB =			vmp_pmulhluBGeneric;
		vmp_pmulhluB0 =			vmp_pmulhluBGeneric;

		vmp_pmulhB =			vmp_pmulhBGeneric;
		vmp_pmulhB0 =			vmp_pmulhBGeneric;
		vmp_pmulhuB =			vmp_pmulhuBGeneric;
		vmp_pmulhuB0 =			vmp_pmulhuBGeneric;
		vmp_pmulluB =			vmp_pmulluBGeneric;
		vmp_pmulluB0 =			vmp_pmulluBGeneric;
	  }
	  
	vmp_pmulluW =			vmp_pmulluWGeneric;
	vmp_pmulluW0 =			vmp_pmulluWGeneric;
}




