
//
//  Bench.cpp - (Quad) Vector 3D Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries

#include <iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"
#include "QVmd3D.h"										// (Quad) Vector Multiplication & Division
#include "QVmd3DEx.h"									// (Quad)  "   " (Private)



typedef enum
{
	iTEST_DOT_PRODUCT,
	iTEST_CROSS_PRODUCT,
	iTEST_VEC_MUL,
	iTEST_VEC_MADD,
	iTEST_VEC_MSUB,
	iTEST_VEC_SCALE,
	iTEST_VEC_DIV,
	iTEST_VEC_DIV_FAST
} iTEST;



//
//	Test 
//

static void testQVec( bool bAlign, iTEST iFunc )
{
	vmp3DQVector *pvA, *pvB, *pvC, *pvTmp, *pvTmp2;
	int w, x, y, z, dif;
	float f, f2, fPrecision, fScalar;
	float buf[4*4  +3];			// enough space + 12 bytes (3 floats) extra
	bool bF;

		// Force proper alignment (Crude just for this demo)

	bF = false;

	if (bAlign)
	  {
		pvA    = (vmp3DQVector*) ALIGN16((int)(buf));
		pvB    = (vmp3DQVector*) ALIGN16((int)(pvA+1));
		pvTmp  = (vmp3DQVector*) ALIGN16((int)(pvB+1));
		pvTmp2 = (vmp3DQVector*) ALIGN16((int)(pvTmp+1));
	  }
	else
	  {
		pvA    = (vmp3DQVector*) buf;
		pvB    = pvA + 1;
		pvTmp  = pvB + 1;
		pvTmp2 = pvTmp + 1;
	  }

	pvC = pvA;
	dif = 4;
	fPrecision = 0.0000001f;
	
	for ( w = -dif; w <= dif; w++ )
	  {
	  	pvA->w = 0.123f *(float)w;
		pvB->z = 0.123f *(float)w;
		
		for ( z = -dif; z <= dif; z++ )
		  {
			pvA->z = 0.123f *(float)z;
			pvB->y = 0.123f *(float)z;

			for ( y = -dif; y <= dif; y++ )
			  {
				pvA->y = 0.123f *(float)y;
				pvB->x = 0.123f *(float)y;

				for ( x = -dif; x <= dif; x++ )
				  {	
				    pvB->w = 0.123f *(float)x;
				    pvA->x = 0.123f *(float)x;

					fScalar = pvA->x;

					switch(iFunc)
					  {
						case iTEST_DOT_PRODUCT:
							fPrecision = 0.0000001f;
							bF = true;
							vmp_QDotProductGeneric( &f2, pvA, pvB );
							if (bAlign)
								vmp_QDotProduct( &f, pvA, pvB );		// Aligned
							else
								vmp_QDotProduct0( &f, pvA, pvB );
							break;

						case iTEST_CROSS_PRODUCT:
							fPrecision = 0.0000001f;
							bF = false;
							vmp_QCrossProductGeneric( pvTmp2, pvA, pvB );
							if (bAlign)
								vmp_QCrossProduct( pvTmp, pvA, pvB );	// Aligned
							else
								vmp_QCrossProduct0( pvTmp, pvA, pvB );
							break;

						case iTEST_VEC_MUL:
							fPrecision = 0.0000001f;
							bF = false;
							vmp_QVecMulGeneric( pvTmp2, pvA, pvB );
							if (bAlign)
								vmp_QVecMul( pvTmp, pvA, pvB );			// Aligned
							else
								vmp_QVecMul0( pvTmp, pvA, pvB );
							break;

						case iTEST_VEC_MADD:
							fPrecision = 0.0000001f;
							bF = false;
							vmp_QVecMAddGeneric( pvTmp2, pvA, pvB, pvC );
							if (bAlign)
								vmp_QVecMAdd( pvTmp, pvA, pvB, pvC );			// Aligned
							else
								vmp_QVecMAdd0( pvTmp, pvA, pvB, pvC );
							break;

						case iTEST_VEC_MSUB:
							fPrecision = 0.0000001f;
							bF = false;
							vmp_QVecMSubGeneric( pvTmp2, pvA, pvB, pvC );
							if (bAlign)
								vmp_QVecMSub( pvTmp, pvA, pvB, pvC );			// Aligned
							else
								vmp_QVecMSub0( pvTmp, pvA, pvB, pvC );
							break;
						case iTEST_VEC_SCALE:
							fPrecision = 0.0000001f;
							bF = false;
							vmp_QVecScaleGeneric( pvTmp2, pvA, fScalar );
							if (bAlign)
								vmp_QVecScale( pvTmp, pvA, fScalar );	// Aligned
							else
								vmp_QVecScale0( pvTmp, pvA, fScalar );
							break;

						case iTEST_VEC_DIV:
							fPrecision = 0.0000001f;
							bF = false;
							if (0.0==fScalar) continue;

							vmp_QVecDivGeneric( pvTmp2, pvA, fScalar );
							if (bAlign)
								vmp_QVecDiv( pvTmp, pvA, fScalar );			// Aligned
							else
								vmp_QVecDiv0( pvTmp, pvA, fScalar );
							break;

						case iTEST_VEC_DIV_FAST:						// Lower Precision
							fPrecision = 0.001f;
							bF = false;
							if (0.0==fScalar) continue;

							vmp_QVecDivGeneric( pvTmp2, pvA, fScalar );
							if (bAlign)
								vmp_QVecDivFast( pvTmp, pvA, fScalar );			// Aligned
							else
								vmp_QVecDivFast0( pvTmp, pvA, fScalar );
							break;
					  }

					if (bF)
					  {
						if (!vmp_IsFEqual(f, f2, fPrecision))
						  {
							cout << " " << f2 << "   vA.x:" << pvA->x << " vA.y:" << pvA->y << " vA.z:" << pvA->z << "\n";
							cout << " vB.x:" << pvB->x << " vB.y:" << pvB->y << " vB.z:" << pvB->z << "\n"; 
							cout << "    " << f << "  *** MISMATCH!!! ***\n" << endl;
						  }
					  }
					else
					  {
						if (!vmp_IsQVEqual(pvTmp, pvTmp2, fPrecision)) 
						  {
							cout << "TMP x:" << pvTmp2->x << " y:" << pvTmp2->y << " z:" << pvTmp2->z << " w:" << pvTmp2->w << " =\n";
							cout << " vA.x:" << pvA->x << " vA.y:" << pvA->y << " vA.z:" << pvA->z << " vA.w:" << pvA->w << "\n";
							cout << " vB.x:" << pvB->x << " vB.y:" << pvB->y << " vB.z:" << pvB->z << " vB.w:" << pvB->w << "\n"; 
							cout << "    x:" << pvTmp->x << " y:" << pvTmp->y << " z:" << pvTmp->z << " w:" << pvTmp->w << "   *** MISMATCH!!! ***\n" << endl;
						  }
					  }
				  }
			  }
		  }
	  }
}



int main( int argc, const char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

#pragma unused(argc, argv)

    CpuDetect( &cinfo );								// Detect CPU
    
    cout << "\n(Quad) Vector Scaling and Multiplication Test\n\n";
    cout << CpuInfoStr( szBuf, &cinfo ) << endl;
    
    QVmd3DSetup( &cinfo );								// Setup Quad (Vector) stuff
    
	cout << "Testing (Quad) Scaling" << endl; 
	testQVec( false, iTEST_VEC_SCALE );
	
	cout << "Testing (Quad) Scaling (Aligned)" << endl;
	testQVec( true,  iTEST_VEC_SCALE );

	cout << "Testing (Quad) Vector Multiplication" << endl;
	testQVec( false, iTEST_VEC_MUL );

	cout << "Testing (Quad) Vector Multiplication (Aligned)" << endl;
	testQVec( true,  iTEST_VEC_MUL );

	cout << "Testing (Quad) Vector Multiplication-Add" << endl;
	testQVec( false, iTEST_VEC_MADD );

	cout << "Testing (Quad) Vector Multiplication-Add (Aligned)" << endl;
	testQVec( true,  iTEST_VEC_MADD );

	cout << "Testing (Quad) Vector Multiplication-Sub" << endl;
	testQVec( false, iTEST_VEC_MSUB );

	cout << "Testing (Quad) Vector Multiplication-Sub (Aligned)" << endl;
	testQVec( true,  iTEST_VEC_MSUB );

	cout << "Testing (Quad) Vector Division" << endl;
	testQVec( false, iTEST_VEC_DIV );

	cout << "Testing (Quad) Vector Division (Aligned)" << endl;
	testQVec( true,  iTEST_VEC_DIV );

	cout << "Testing (Quad) Vector Division **FAST**" << endl;
	testQVec( false, iTEST_VEC_DIV_FAST );

	cout << "Testing (Quad) Vector Division (Aligned) **FAST**" << endl;
	testQVec( true,  iTEST_VEC_DIV_FAST );

	cout << "Testing (Quad) Dot Product" << endl;
	testQVec( false, iTEST_DOT_PRODUCT );

	cout << "Testing (Quad) Dot Product (Aligned)" << endl;
	testQVec( true,  iTEST_DOT_PRODUCT );

	cout << "Testing (Quad) Cross Product" << endl;
	testQVec( false, iTEST_CROSS_PRODUCT );

	cout << "Testing (Quad) Cross Product (Aligned)" << endl;
	testQVec( true,  iTEST_CROSS_PRODUCT );

	cout << "Fin\n\n" << flush;

    return 0;
}

