
//
//	Vmd3DAltivec.cpp - (Quad) 3D Vector Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include	"CpuAsm.h"									// CPU
#include	"vmp3D.h"
#include	"vmp3DEx.h"
#include	"QVmd3D.h"									// 3D Math
#include	"QVmd3DEx.h"




//
//	(Quad) Vector Product (Multiplication) - Unaligned
//

void vmp_QVecMulAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{	
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = pvB->w;
	
	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}



//
//	(Quad) Vector Product (Multiplication) - Aligned
//

void vmp_QVecMulAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);

	*(vector float *)pvD = vec_madd( (*(vector float *)pvA), (*(vector float *)pvB), vZero );		// vmaddfp d,a,b,c
}



//______________________________________________________
//
//	(Quad) Vector Product Sum (Multiplication-Add) - Unaligned
//

void vmp_QVecMAddAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA,
                   const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
{	
	vector float vD, vA, vB, vC;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvC);
		
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = pvB->w;
	
	((vmp3DQVector *) &vC)->x = pvC->x;
	((vmp3DQVector *) &vC)->y = pvC->y;
	((vmp3DQVector *) &vC)->z = pvC->z;
	((vmp3DQVector *) &vC)->w = pvC->w;

	vD = vec_madd( vA, vB, vC );					// vmaddfp d,a,b,c

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}



//
//	(Quad) Vector Product Sum (Multiplication-Add) - Aligned
//

void vmp_QVecMAddAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA,
                    const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
{	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);
	ASSERT_PTR16(pvC);

	*(vector float *)pvD = vec_madd( (*(vector float *)pvA), (*(vector float *)pvB), (*(vector float *)pvC) );		// vmaddfp d,a,b,c
}



//______________________________________________________
//
//	(Quad) Vector Product Dif (Multiplication-Sub) - Unaligned
//

void vmp_QVecMSubAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA,
                   const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
{	
	vector float vD, vA, vB, vC;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvC);
		
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = pvB->w;
	
	((vmp3DQVector *) &vC)->x = pvC->x;
	((vmp3DQVector *) &vC)->y = pvC->y;
	((vmp3DQVector *) &vC)->z = pvC->z;
	((vmp3DQVector *) &vC)->w = pvC->w;

	vD = vec_sub( vZero, vC );
	vD = vec_madd( vA, vB, vD );					// vnmsubfp d,a,b,c

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}



//
//	(Quad) Vector Product Dif (Multiplication-Sub) - Aligned
//

void vmp_QVecMSubAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA,
                    const vmp3DQVector * const pvB, const vmp3DQVector * const pvC)
{	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);
	ASSERT_PTR16(pvC);

	*(vector float *)pvD = vec_sub( vZero, (*(vector float *)pvC) );
	*(vector float *)pvD = vec_madd( (*(vector float *)pvA), (*(vector float *)pvB), (*(vector float *)pvD) );		// vmaddfp d,a,b,c
}



//______________________________________________________
//
//	Vector Scalar Division - ***FAST*** Unaligned
//
// WARNING: vec_re is 1/4096 accurate!!!

void vmp_QVecDivFastAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar )
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_ZERO(fScalar);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;
	
	((vmp3DQVector *) &vB)->w = ((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = fScalar;
	
	vB = vec_re( vB );				// 1/n
	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}


//	Vector Scalar Division - ***FAST*** Aligned
//
// WARNING: vec_re is 1/4096 accurate!!!

void vmp_QVecDivFastAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar )
{
	vector float vB;

	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_ZERO(fScalar);
	
	((vmp3DQVector *) &vB)->w = ((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = fScalar;
	
	vB = vec_re( vB );				// 1/n
	*(vector float *)pvD = vec_madd( (*(vector float *)pvA), vB, vZero );					// vmaddfp d,a,b,c
}



//______________________________________________________
//
//	(Quad) Scale Vector w/Scalar - Unaligned
//

void vmp_QVecScaleAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar )
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = ((vmp3DQVector *) &vB)->w = fScalar;
	
	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
}



//
//	(Quad) Scale Vector w/Scalar - Aligned
//

void vmp_QVecScaleAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, float fScalar )
{
	vector float vB;

	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = ((vmp3DQVector *) &vB)->w = fScalar;
	
	*(vector float *)pvD = vec_madd( (*(vector float *)pvA), vB, vZero );	// vmaddfp d,a,b,c
}



//______________________________________________________
//
//	Dot Product - Unaligned
//
// f = pvA->x * pvB->x + pvA->y * pvB->y + pvA->z * pvB->z;
//

void vmp_QDotProductAltivec0(float * const pfD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c
	
	*pfD = ((vmp3DQVector *) &vD)->x
	    + ((vmp3DQVector *) &vD)->y
	    + ((vmp3DQVector *) &vD)->z;
}



//
//	Dot Product - Aligned
//

void vmp_QDotProductAltivec(float * const pfD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{
	vector float vD;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);
	
	vD = vec_madd( (*(vector float *)pvA), (*(vector float *)pvB), vZero );					// vmaddfp d,a,b,c
	
	*pfD = ((vmp3DQVector *) &vD)->x
	    + ((vmp3DQVector *) &vD)->y
	    + ((vmp3DQVector *) &vD)->z;
}



//______________________________________________________
//
//	Cross Product - Unaligned
//
//	x = pvA->y * pvB->z - pvA->z * pvB->y;    1 2   2 1
//	y = pvA->z * pvB->x - pvA->x * pvB->z;    2 0   0 2
//	z = pvA->x * pvB->y - pvA->y * pvB->x;    0 1   1 0

void vmp_QCrossProductAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{	
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->z;
	((vmp3DQVector *) &vA)->y = pvA->x;
	((vmp3DQVector *) &vA)->z = pvA->y;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->y;
	((vmp3DQVector *) &vB)->y = pvB->z;
	((vmp3DQVector *) &vB)->z = pvB->x;
	((vmp3DQVector *) &vB)->w = 0.0;

	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c
	vD = vec_sub( vZero, vD );
		
	((vmp3DQVector *) &vA)->x = pvA->y;
	((vmp3DQVector *) &vA)->y = pvA->z;
	((vmp3DQVector *) &vA)->z = pvA->x;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->z;
	((vmp3DQVector *) &vB)->y = pvB->x;
	((vmp3DQVector *) &vB)->z = pvB->y;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	vD = vec_madd( vA, vB, vD );
	
	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = 0.0f;
}



//
//	Cross Product - Aligned
//

void vmp_QCrossProductAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB)
{	
	vector float vD, vA, vB;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->z;
	((vmp3DQVector *) &vA)->y = pvA->x;
	((vmp3DQVector *) &vA)->z = pvA->y;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->y;
	((vmp3DQVector *) &vB)->y = pvB->z;
	((vmp3DQVector *) &vB)->z = pvB->x;
	((vmp3DQVector *) &vB)->w = 0.0;

	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c
	vD = vec_sub( vZero, vD );
		
	((vmp3DQVector *) &vA)->x = pvA->y;
	((vmp3DQVector *) &vA)->y = pvA->z;
	((vmp3DQVector *) &vA)->z = pvA->x;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->z;
	((vmp3DQVector *) &vB)->y = pvB->x;
	((vmp3DQVector *) &vB)->z = pvB->y;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	*(vector float *)pvD = vec_madd( vA, vB, vD );
	pvD->w = 0.0f;
}



//______________________________________________________
//
//	Setup Vec3D module
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void QVmd3DSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);
	
	if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_QVecMul =			vmp_QVecMulAltivec;
		vmp_QVecMul0 =			vmp_QVecMulAltivec0;
		vmp_QVecMAdd =			vmp_QVecMAddAltivec;
		vmp_QVecMAdd0 =			vmp_QVecMAddAltivec0;
		vmp_QVecMSub =			vmp_QVecMSubAltivec;
		vmp_QVecMSub0 =			vmp_QVecMSubAltivec0;
		vmp_QVecScale =			vmp_QVecScaleAltivec;
		vmp_QVecScale0 =		vmp_QVecScaleAltivec0;
		vmp_QVecDiv =			vmp_QVecDivGeneric;
		vmp_QVecDiv0 =			vmp_QVecDivGeneric;
		vmp_QVecDivFast =		vmp_QVecDivFastAltivec;
		vmp_QVecDivFast0 =		vmp_QVecDivFastAltivec0;
		vmp_QDotProduct =		vmp_QDotProductAltivec;
		vmp_QDotProduct0 =		vmp_QDotProductAltivec0;
		vmp_QCrossProduct =		vmp_QCrossProductAltivec;
		vmp_QCrossProduct0 =	vmp_QCrossProductAltivec0;
	  }

	else
	  {
		vmp_QVecMul =			vmp_QVecMulGeneric;
		vmp_QVecMul0 =			vmp_QVecMulGeneric;
		vmp_QVecMAdd =			vmp_QVecMAddGeneric;
		vmp_QVecMAdd0 =			vmp_QVecMAddGeneric;
		vmp_QVecMSub =			vmp_QVecMSubGeneric;
		vmp_QVecMSub0 =			vmp_QVecMSubGeneric;
		vmp_QVecScale =			vmp_QVecScaleGeneric;
		vmp_QVecScale0 =		vmp_QVecScaleGeneric;
		vmp_QVecDiv =			vmp_QVecDivGeneric;
		vmp_QVecDiv0 =			vmp_QVecDivGeneric;
		vmp_QVecDivFast =		vmp_QVecDivGeneric;
		vmp_QVecDivFast0 =		vmp_QVecDivGeneric;
		vmp_QDotProduct =		vmp_QDotProductGeneric;
		vmp_QDotProduct0 =		vmp_QDotProductGeneric;
		vmp_QCrossProduct =		vmp_QCrossProductGeneric;
		vmp_QCrossProduct0 =	vmp_QCrossProductGeneric;
	  }
}



