
//
//	Vmd3DAltivec.cpp - 3D Vector Multiplication & Division
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//	NOTE: Because these a 3 float Vectors the 4th float has to be handled as 
//		a special case thus slowing the code down. This is the same as an AoS 
//		orientation. Instead a SoA should be used in conjunction with the quad
//		set of functions.
//
// *** Also in aligned code could process pvA, pvB & pvC directly but .w is unknown!!! and could
//		be a a non-valid float!
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include	"CpuAsm.h"									// CPU
#include	"Vmp3D.h"
#include	"Vmp3DEx.h"
#include	"Vmd3D.h"									// 3D Math
#include	"Vmd3DEx.h"





//
//	Vector Product (Multiplication) - (Un)Aligned
//

void vmp_VecMulAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{	
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c

	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
}



//______________________________________________________
//
//	Vector Product Sum (Multiplication-Add) - (Un)Aligned
//

void vmp_VecMAddAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA,
				  const vmp3DVector * const pvB, const vmp3DVector * const pvC)
{	
	vector float vD, vA, vB, vC;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	ASSERT_PTR4(pvC);
		
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	((vmp3DQVector *) &vC)->x = pvC->x;
	((vmp3DQVector *) &vC)->y = pvC->y;
	((vmp3DQVector *) &vC)->z = pvC->z;
	((vmp3DQVector *) &vC)->w = 0.0;
	
	vD = vec_madd( vA, vB, vC );					// vmaddfp d,a,b,c

	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
}



//______________________________________________________
//
//	Vector Product Sum (Multiplication-Add) - (Un)Aligned
//

void vmp_VecMSubAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA,
				 const vmp3DVector * const pvB, const vmp3DVector * const pvC)
{	
	vector float vD, vA, vB, vC;
	
	ASSERT_PTR4(pvC);
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
		
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0;

	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	((vmp3DQVector *) &vC)->x = pvC->x;
	((vmp3DQVector *) &vC)->y = pvC->y;
	((vmp3DQVector *) &vC)->z = pvC->z;
	((vmp3DQVector *) &vC)->w = 0.0;

	vD = vec_sub( vZero, vC );
	vD = vec_madd( vA, vB, vD );

	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
}




//______________________________________________________
//
//	Scale Vector w/Scalar - (Un)aligned
//

void vmp_VecScaleAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar )
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = fScalar;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c

	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
}



// WARNING: vec_re is 1/4096 accurate so there is no 'vector' division!!!


//______________________________________________________
//
//	Vector Scalar Division - ***FAST*** (Un)Aligned
//
// WARNING: vec_re is 1/4096 accurate!!!

void vmp_VecDivFastAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA, float fScalar )
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_ZERO(fScalar);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = ((vmp3DQVector *) &vB)->y = ((vmp3DQVector *) &vB)->z = fScalar;
	((vmp3DQVector *) &vB)->w = 1.0;
	
	vB = vec_re( vB );				// 1/n
	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c

	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
}


//______________________________________________________
//
//	Dot Product - (Un)Aligned
//
// f = pvA->x * pvB->x + pvA->y * pvB->y + pvA->z * pvB->z;
//
// NOTE: Not very efficient, should do four dot products in parallel!

void vmp_DotProductAltivec(float * const pfD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c
	
	*pfD = ((vmp3DVector *) &vD)->x
	    + ((vmp3DVector *) &vD)->y
	    + ((vmp3DVector *) &vD)->z;
}



//______________________________________________________
//
//	Cross Product - (Un)Aligned
//
//	x = pvA->y * pvB->z - pvA->z * pvB->y;    1 2   2 1
//	y = pvA->z * pvB->x - pvA->x * pvB->z;    2 0   0 2
//	z = pvA->x * pvB->y - pvA->y * pvB->x;    0 1   1 0

void vmp_CrossProductAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB)
{	
	vector float vD, vA, vB;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);
	
	((vmp3DQVector *) &vA)->x = pvA->z;
	((vmp3DQVector *) &vA)->y = pvA->x;
	((vmp3DQVector *) &vA)->z = pvA->y;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->y;
	((vmp3DQVector *) &vB)->y = pvB->z;
	((vmp3DQVector *) &vB)->z = pvB->x;
	((vmp3DQVector *) &vB)->w = 0.0;

	vD = vec_madd( vA, vB, vZero );					// vmaddfp d,a,b,c
	vD = vec_sub( vZero, vD );
		
	((vmp3DQVector *) &vA)->x = pvA->y;
	((vmp3DQVector *) &vA)->y = pvA->z;
	((vmp3DQVector *) &vA)->z = pvA->x;
	((vmp3DQVector *) &vA)->w = 0.0;
	
	((vmp3DQVector *) &vB)->x = pvB->z;
	((vmp3DQVector *) &vB)->y = pvB->x;
	((vmp3DQVector *) &vB)->z = pvB->y;
	((vmp3DQVector *) &vB)->w = 0.0;
	
	vD = vec_madd( vA, vB, vD );					// vnmsubfp d,a,b,c
	
	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
}



//______________________________________________________
//
//	Setup Vec3D module
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void Vmd3DSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);
	
	if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_VecMul =			vmp_VecMulAltivec;
		vmp_VecMul0 =			vmp_VecMulAltivec;
		vmp_VecMAdd =			vmp_VecMAddAltivec;
		vmp_VecMAdd0 =			vmp_VecMAddAltivec;
		vmp_VecMSub =			vmp_VecMSubAltivec;
		vmp_VecMSub0 =			vmp_VecMSubAltivec;
		vmp_VecDiv =			vmp_VecDivGeneric;
		vmp_VecDiv0 =			vmp_VecDivGeneric;
		vmp_VecDivFast = 		vmp_VecDivFastAltivec;
		vmp_VecDivFast0 =		vmp_VecDivFastAltivec;
		vmp_VecScale =			vmp_VecScaleAltivec;
		vmp_VecScale0 =			vmp_VecScaleAltivec;
		vmp_DotProduct =		vmp_DotProductAltivec;
		vmp_DotProduct0 =		vmp_DotProductAltivec;
		vmp_CrossProduct =		vmp_CrossProductAltivec;
		vmp_CrossProduct0 =		vmp_CrossProductAltivec;
	  }

	else
	  {
		vmp_VecMul =			vmp_VecMulGeneric;
		vmp_VecMul0 =			vmp_VecMulGeneric;
		vmp_VecMAdd =			vmp_VecMAddGeneric;
		vmp_VecMAdd0 =			vmp_VecMAddGeneric;
		vmp_VecMSub =			vmp_VecMSubGeneric;
		vmp_VecMSub0 =			vmp_VecMSubGeneric;
		vmp_VecDiv =			vmp_VecDivGeneric;
		vmp_VecDiv0 =			vmp_VecDivGeneric;
		vmp_VecDivFast =		vmp_VecDivGeneric;
		vmp_VecDivFast0 =		vmp_VecDivGeneric;
		vmp_VecScale =			vmp_VecScaleGeneric;
		vmp_VecScale0 =			vmp_VecScaleGeneric;
		vmp_DotProduct =		vmp_DotProductGeneric;
		vmp_DotProduct0 =		vmp_DotProductGeneric;
		vmp_CrossProduct =		vmp_CrossProductGeneric;
		vmp_CrossProduct0 =		vmp_CrossProductGeneric;
	  }
}



