
//
//  Bench.cpp - (Scalar) Special Function (Float)
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include <iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"
#include "Fsf.h"
#include "FsfEx.h"



typedef enum
{
	iTEST_FLT_SQRT
} iTEST;



//
//	Test 
//

static void testVal( bool bAlign, iTEST iFunc, bool bFast )
{
	vmp3DVector *pvA, *pvB, *pvTmp, *pvTmp2;
	int x, y, z, dif;
	float f, f2, fPrecision, fA, fB;
	float buf[4*4  +3];			// enough space + 12 bytes (3 floats) extra

		// Force proper alignment (Crude just for this demo)

	if (bAlign)
	  {
		pvA    = (vmp3DVector*) ALIGN16((int)(buf));
		pvB    = (vmp3DVector*) ALIGN16((int)(pvA+1));
		pvTmp  = (vmp3DVector*) ALIGN16((int)(pvB+1));
		pvTmp2 = (vmp3DVector*) ALIGN16((int)(pvTmp+1));
	  }
	else
	  {
		pvA    = (vmp3DVector*) buf;
		pvB    = pvA + 1;
		pvTmp  = pvB + 1;
		pvTmp2 = pvTmp + 1;
	  }

	dif = 4;

	fPrecision = (bFast) ? 0.0001f : 0.0000001f;
	
	for ( z = -dif; z <= dif; z++ )
	  {
		pvA->z = 0.123f *(float)z;
		pvB->y = 0.123f *(float)z;

		for ( y = -dif; y <= dif; y++ )
		  {
			pvA->y = 0.123f *(float)y;
			pvB->x = 0.123f *(float)y;

			for ( x = -dif; x <= dif; x++ )
			  {	
				pvA->x = 0.123f *(float)x;
				pvB->z = 0.123f *(float)y;
				fA = pvA->x;
				fB = pvB->z;

				switch(iFunc)
				  {
					case iTEST_FLT_SQRT:
						if ( 0.0f > fA ) fA = -fA;
						if ( 0.0f > fB ) fB = -fB;

						vmp_FSqrtGeneric( &f2, fA );
						if ( bFast )
						  {
							vmp_FSqrtFast( &f, fA );
						  }
						else
						  {
							vmp_FSqrt( &f, fA );
						  }
						break;
				  }

				if (!vmp_IsFEqual(f, f2, fPrecision))
				  {
					cout << " " << f2 << "   vA.x:" << pvA->x << " vA.y:" << pvA->y << " vA.z:" << pvA->z << endl;
					cout << " vB.x:" << pvB->x << " vB.y:" << pvB->y << " vB.z:" << pvB->z << endl; 
					cout << "    " << f << "  *** MISMATCH!!! ***\n" << endl;
				  }
			  }
		  }
	  }
}



int main( int argc, const char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

#pragma unused(argc, argv)

    CpuDetect( &cinfo );								// Detect CPU

	cout << "\nTesting (Scalar) Square Root\n\n";
	cout << CpuInfoStr( szBuf, &cinfo ) << endl;

	FsfSetup( &cinfo );								// Setup (float) stuff

	cout << "Testing Square Root" << endl;
	testVal( false, iTEST_FLT_SQRT, false );

	cout << "Testing Square Root *FAST (Low Precision)*" << endl;
	testVal( false, iTEST_FLT_SQRT, true );

	cout << "Fin!\n\n" << flush;

    return 0;
}

