
//
//  Bench.cpp - 3D (Quad) Vector Special Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include <iostream.h>
#include <stdio.h>

    //  Book Development Library Modules
    
#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"

#include "QVsf3D.h"
#include "QVsf3DEx.h"


#include	"vmp3DEx.h"


typedef enum
{
	iTEST_MAGNITUDE,
	iTEST_MAGNITUDE_FAST,
	iTEST_NORMALIZE,
	iTEST_NORMALIZE_FAST,
	iTEST_VEC_SQRT,
	iTEST_VEC_SQRT_FAST,
	iTEST_MIN,
	iTEST_MAX
} iTEST;



//
//	Test 
//

static void testQVec( bool bAlign, iTEST iFunc )
{
	vmp3DQVector *pvA, *pvB, *pvTmp, *pvTmp2;
	int w, x, y, z, dif;
	float f, f2, fPrecision, fScalar;
	float buf[4 *4  +3];			// enough space + 12 bytes (3 floats) extra
	bool bF;

		// Force proper alignment (Crude just for this demo)

	bF = false;

	if ( bAlign )
	  {
		pvA    = (vmp3DQVector*) ALIGN16((int)(buf));
		pvB    = (vmp3DQVector*) ALIGN16((int)(pvA+1));
		pvTmp  = (vmp3DQVector*) ALIGN16((int)(pvB+1));
		pvTmp2 = (vmp3DQVector*) ALIGN16((int)(pvTmp+1));
	  }
	else
	  {
		pvA    = (vmp3DQVector*) buf;
		pvB    = pvA + 1;
		pvTmp2 = pvB + 1;
		pvTmp  = pvTmp2 + 1;
	  }

	dif = 4;
	fPrecision = SINGLE_PRECISION;

	for ( w = -dif; w <= dif; w++ )
	  {
	  	pvA->w = 0.123f *(float)w;
		pvB->z = 0.123f *(float)w;
		
		for ( z = -dif; z <= dif; z++ )
		  {
			pvA->z = 0.123f *(float)z;
			pvB->y = 0.123f *(float)z;

			for ( y = -dif; y <= dif; y++ )
			  {
				pvA->y = 0.123f *(float)y;
				pvB->x = 0.123f *(float)y;

				for ( x = -dif; x <= dif; x++ )
				  {	
					pvB->w = 0.123f *(float)x;
					pvA->x = 0.123f *(float)x;

					fScalar = pvA->x;

					switch(iFunc)
					  {
						case iTEST_VEC_SQRT:
							fPrecision = SINGLE_PRECISION;
							bF = false;
							if ( pvA->x < 0.0f ) pvA->x = -pvA->x;
							if ( pvA->y < 0.0f ) pvA->y = -pvA->y;
							if ( pvA->z < 0.0f ) pvA->z = -pvA->z;
							if ( pvA->w < 0.0f ) pvA->w = -pvA->w;

							vmp_QVecSqrtGeneric( pvTmp2, pvA );
							if (bAlign)
								vmp_QVecSqrt( pvTmp, pvA );				// Aligned
							else
								vmp_QVecSqrt0( pvTmp, pvA );
							break;

						case iTEST_VEC_SQRT_FAST:
							fPrecision = FAST_PRECISION;
							bF = false;
							if ( pvA->x < 0.0f ) pvA->x = -pvA->x;
							if ( pvA->y < 0.0f ) pvA->y = -pvA->y;
							if ( pvA->z < 0.0f ) pvA->z = -pvA->z;
							if ( pvA->w < 0.0f ) pvA->w = -pvA->w;

							vmp_QVecSqrtGeneric( pvTmp2, pvA );
							if (bAlign)	
								vmp_QVecSqrtFast( pvTmp, pvA );			// Aligned
							else
								vmp_QVecSqrtFast0( pvTmp, pvA );
							break;

						case iTEST_MAGNITUDE:
							fPrecision = SINGLE_PRECISION;
							bF = true;
							vmp_QVecMagnitudeGeneric( &f2, pvA );
							if (bAlign)
								vmp_QVecMagnitude( &f, pvA );			// Aligned
							else
								vmp_QVecMagnitude0( &f, pvA );
							break;

						case iTEST_MAGNITUDE_FAST:
							fPrecision = FAST_PRECISION;
							bF = true;
							vmp_QVecMagnitudeGeneric( &f2, pvA );
							if (bAlign)
								vmp_QVecMagnitudeFast( &f, pvA );		// Aligned
							else
								vmp_QVecMagnitudeFast0( &f, pvA );
							break;

						case iTEST_NORMALIZE:
							fPrecision = SINGLE_PRECISION;
							bF = false;
							vmp_QVecNormalizeGeneric( pvTmp2, pvA );
							if (bAlign)
								vmp_QVecNormalize( pvTmp, pvA );		// Aligned
							else
								vmp_QVecNormalize0( pvTmp, pvA );
							break;

						case iTEST_NORMALIZE_FAST:
							fPrecision = FAST_PRECISION;
							bF = false;
							vmp_QVecNormalizeGeneric( pvTmp2, pvA );
							if (bAlign)
								vmp_QVecNormalizeFast( pvTmp, pvA );	// Aligned
							else
								vmp_QVecNormalizeFast0( pvTmp, pvA );
							break;

						case iTEST_MIN:
							fPrecision = SINGLE_PRECISION;
							bF = false;
							vmp_QVecMinGeneric( pvTmp2, pvA, pvB );
							if (bAlign)
								vmp_QVecMin( pvTmp, pvA, pvB );			// Aligned
							else
								vmp_QVecMin0( pvTmp, pvA, pvB );
							break;

						case iTEST_MAX:
							fPrecision = SINGLE_PRECISION;
							bF = false;
							vmp_QVecMaxGeneric( pvTmp2, pvA, pvB );
							if (bAlign)
								vmp_QVecMax( pvTmp, pvA, pvB );			// Aligned
							else
								vmp_QVecMax0( pvTmp, pvA, pvB );
							break;
					  }

					if (bF)
					  {
						if (!vmp_IsFEqual(f, f2, fPrecision))
						  {
							cout << " " << f2 << "   vA.x:" << pvA->x << " vA.y:" << pvA->y << " vA.z:" << pvA->z << "\n";
							cout << " vB.x:" << pvB->x << " vB.y:" << pvB->y << " vB.z:" << pvB->z << "\n"; 
							cout << "    " << f << "  *** MISMATCH!!! ***\n" << endl;
						  }
					  }
					else
					  {
						if (!vmp_IsQVEqual(pvTmp, pvTmp2, fPrecision)) 
						  {
							cout << "TMP2x:" << pvTmp2->x << " y:" << pvTmp2->y << " z:" << pvTmp2->z << " w:" << pvTmp2->w << "\n";
							cout << " vA.x:" << pvA->x << " vA.y:" << pvA->y << " vA.z:" << pvA->z << " vA.w:" << pvA->w << "\n";
							cout << " vB.x:" << pvB->x << " vB.y:" << pvB->y << " vB.z:" << pvB->z << " vB.w:" << pvB->w << "\n"; 
							cout << "TMP x:" << pvTmp->x << " y:" << pvTmp->y << " z:" << pvTmp->z << " w:" << pvTmp->w << "   *** MISMATCH!!! ***\n" << endl;
						  }
					  }
				  }
			  }
		  }
	  }
}



int main( int argc, const char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

#pragma unused(argc, argv)

    CpuDetect( &cinfo );								// Detect CPU
    
    cout << "\n(Quad) Vector Special Function Test\n\n";
	cout << CpuInfoStr( szBuf, &cinfo ) << endl;

	QVsf3DSetup( &cinfo );								// Setup our 3D Vector stuff

	cout << "Testing (Quad) VecSqrt" << endl;
	testQVec( false, iTEST_VEC_SQRT );

	cout << "Testing (Quad) VecSqrt (Aligned)" << endl;
	testQVec( true,  iTEST_VEC_SQRT );

	cout << "Testing (Quad) VecSqrt (FAST)" << endl;
	testQVec( false, iTEST_VEC_SQRT_FAST );

	cout << "Testing (Quad) VecSqrt (FAST) (Aligned)" << endl;
	testQVec( true,  iTEST_VEC_SQRT_FAST );

	cout << "Testing (Quad) VecMin" << endl;
	testQVec( false, iTEST_MIN );

	cout << "Testing (Quad) VecMin (Aligned)" << endl;
	testQVec( true,  iTEST_MIN );

	cout << "Testing (Quad) VecMax" << endl;
	testQVec( false, iTEST_MAX );

	cout << "Testing (Quad) VecMax (Aligned)" << endl;
	testQVec( true,  iTEST_MAX );

	cout << "Testing (Quad) VecMagnitude" << endl;
	testQVec( false, iTEST_MAGNITUDE );

	cout << "Testing (Quad) VecMagnitude (Aligned)" << endl;
	testQVec( true,  iTEST_MAGNITUDE );

	cout << "Testing (Quad) VecMagnitude (FAST) possible Low Precision" << endl;
	testQVec( false, iTEST_MAGNITUDE_FAST );

	cout << "Testing (Quad) VecMagnitude (Aligned) (FAST) possible Low Precision (Aligned)" << endl;
	testQVec( true,  iTEST_MAGNITUDE_FAST );

	cout << "Testing (Quad) VecNormalize" << endl;
	testQVec( false, iTEST_NORMALIZE );

	cout << "Testing (Quad) VecNormalize (Aligned)" << endl;
	testQVec( true,  iTEST_NORMALIZE );

	cout << "Testing (Quad) VecNormalize (FAST) possible Low Precision" << endl;
	testQVec( false, iTEST_NORMALIZE_FAST );

	cout << "Testing (Quad) VecNormalize (Aligned) (FAST) possible Low Precision (Aligned)" << endl;
	testQVec( true,  iTEST_NORMALIZE_FAST );

	cout << "Fin\n\n" << flush;

    return 0;
}

