
//
//	QVsf3DAltivec.cpp - 3D (Quad) Vector Special Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>
#include    <math.h>

    //  Book Development Library Modules

#include	"CpuAsm.h"									// CPU
#include	"vmp3D.h"
#include	"vmp3DEx.h"

#include	"QVsf3D.h"									// 3D Math
#include	"QVsf3DEx.h"



vector float vHalf = (vector float) (0.5, 0.5, 0.5, 0.5);
vector float vOne = (vector float) (1.0, 1.0, 1.0, 1.0);
vector float vTiny = (vector float) (SINGLE_PRECISION, SINGLE_PRECISION, SINGLE_PRECISION, SINGLE_PRECISION);




//
//	(Quad) Vector Square Root - Unaligned  *** Fast *** (Low 1/4096 Precision)
//
//	pvD->x = sqrtf( pvA->x );
//	pvD->y = sqrtf( pvA->y );
//	pvD->z = sqrtf( pvA->z );
//	pvD->w = sqrtf( pvA->w );
	
void vmp_QVecSqrtFastAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
#if 01						// with Unions
	vec vA;
	vector float vD;
	vector bool int viM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);
	ASSERT_FNEG(pvA->w);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	vA.qv.w = pvA->w;

	viM = vec_cmpge(vA.vf, vTiny);		// Set 1's in position of floats too close to zero
	
		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(vA.vf);				// vrsqrtefp d,a   Elements of zero are now infinite

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vD = vec_and(vD, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);				// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	vA.vf = vec_madd( vA.vf, vD, vZero ); 
	
	pvD->x = vA.qv.x;
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;
	pvD->w = vA.qv.w;
#else						// w/o Unions
	vector float vA, vD;
	vector bool int viM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);
	ASSERT_FNEG(pvA->w);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;

	viM = vec_cmpge(vA, vTiny);			// Set 1's in position of floats too close to zero
	
		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(vA);				// vrsqrtefp d,a   Elements of zero are now infinite

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vD = vec_and(vD, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);				// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	vD = vec_madd( vA, vD, vZero ); 
	
	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
#endif
}



//
//	(Quad) Vector Square Root - Aligned ***FAST*** (Low 1/4096 Precision)
//
	
void vmp_QVecSqrtFastAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
	vector float vD;
	vector bool int viM;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);
	ASSERT_FNEG(pvA->w);

	viM = vec_cmpge(*(vector float *)pvA, vTiny);			// Set 1's in position of floats too close to zero

		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(*(vector float *)pvA);		// vrsqrtefp d,a

		//	Correct for infinity due to 1/0 =  1/sart(0)
		
	vD = vec_and(vD, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);				// Blended #'s and 1.0's
	
		// sqrt(x)= x * (1/sqrt(x))
	*(vector float *)pvD = vec_madd( *(vector float *)pvA, vD, vZero ); 
}


//______________________________________________________
//	(Quad) Vector Square Root - Unaligned  (Normal 24bit Precision)
//
//	pvD->x = sqrtf( pvA->x );
//	pvD->y = sqrtf( pvA->y );
//	pvD->z = sqrtf( pvA->z );
//	pvD->w = sqrtf( pvA->w );

void vmp_QVecSqrtAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
#if 01						// with Unions
	vec vA;
	vector float vD, vT, vU;
	vector bool int viM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);
	ASSERT_FNEG(pvA->w);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	vA.qv.w = pvA->w;

	viM = vec_cmpge(vA.vf, vTiny);		// Set 1's in position of floats too close to zero
	
		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(vA.vf);				// vrsqrtefp d,a   Elements of zero are now infinite

		//	Increase precision using Newton-Raphson refinement
		
	vT = vec_madd(vD,vD,vZero);						// y0*y0 +0
	vU = vec_madd(vD,vHalf,vZero);					// y0*0.5 +0
	vT = vec_nmsub(vA.vf,vT,vOne);					// t0 = B*t0-1)
	vD = vec_madd(vT,vU,vD);						// y1 = t0*t1+y0

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vD = vec_and(vD, viM);							// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);							// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	vA.vf = vec_madd( vA.vf, vD, vZero ); 
	
	pvD->x = vA.qv.x;
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;
	pvD->w = vA.qv.w;
#else						// w/o Unions
	vector float vA, vD, vT, vU;
	vector bool int viM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);
	ASSERT_FNEG(pvA->w);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;

	viM = vec_cmpge(vA, vTiny);			// Set 1's in position of floats too close to zero
	
		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(vA);				// vrsqrtefp d,a   Elements of zero are now infinite

		//	Increase precision using Newton-Raphson refinement
		
	vT = vec_madd(vD,vD,vZero);						// y0*y0 +0
	vU = vec_madd(vD,vHalf,vZero);					// y0*0.5 +0
	vT = vec_nmsub(vA,vT,vOne);						// t0 = B*t0-1)
	vD = vec_madd(vT,vU,vD);						// y1 = t0*t1+y0

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vD = vec_and(vD, viM);							// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);							// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	vD = vec_madd( vA, vD, vZero ); 
	
	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
#endif
}



//
//	(Quad) Vector Square Root - Aligned  (Normal 24bit Precision)
//

void vmp_QVecSqrtAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
	vector float vD, vT, vU;
	vector bool int viM;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);
	ASSERT_FNEG(pvA->w);

	viM = vec_cmpge(*(vector float *)pvA, vTiny);			// Set 1's in position of floats too close to zero

		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(*(vector float *)pvA);			// vrsqrtefp d,a

		//	Increase precision using Newton-Raphson refinement
		
	vT = vec_madd(vD,vD,vZero);						// y0*y0 +0
	vU = vec_madd(vD,vHalf,vZero);					// y0*0.5 +0
	vT = vec_nmsub(*(vector float *)pvA,vT,vOne);	// t0 = B*t0-1)
	vD = vec_madd(vT,vU,vD);						// y1 = t0*t1+y0

		//	Correct for infinity due to 1/0 =  1/sart(0)
		
	vD = vec_and(vD, viM);							// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);							// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	*(vector float *)pvD = vec_madd( *(vector float *)pvA, vD, vZero ); 
}


//______________________________________________________
//
//	(Quad) Vector Minimum - Unaligned
//
//	pvD->x = (pvA->x < pvB->x) ? pvA->x : pvB->x;
//	pvD->y = (pvA->y < pvB->y) ? pvA->y : pvB->y;
//	pvD->z = (pvA->z < pvB->z) ? pvA->z : pvB->z;
//	pvD->w = (pvA->w < pvB->w) ? pvA->w : pvB->w;
	
void vmp_QVecMinAltivec0( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
#if 01						// with Unions
	vec vA, vB;

	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	vA.qv.w = pvA->w;

	vB.qv.x = pvB->x;
	vB.qv.y = pvB->y;
	vB.qv.z = pvB->z;
	vB.qv.w = pvB->w;
	
	vA.vf = vec_min( vA.vf, vB.vf );

	pvD->x = vA.qv.x;
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;
	pvD->w = vA.qv.w;
#else						// w/o Unions
	vector float vD, vA, vB;

	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;

	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = pvB->w;
	
	vD = vec_min( vA, vB );

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
#endif
}


//
//	(Quad) Vector Minimum - Aligned
//
	
void vmp_QVecMinAltivec( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);

	*(vector float *)pvD = vec_min( (*(vector float *)pvA), (*(vector float *)pvB) );
}



//______________________________________________________
//
//	(Quad) Vector Maximum
//
//	pvD->x = (pvA->x > pvB->x) ? pvA->x : pvB->x;
//	pvD->y = (pvA->y > pvB->y) ? pvA->y : pvB->y;
//	pvD->z = (pvA->z > pvB->z) ? pvA->z : pvB->z;
//	pvD->w = (pvA->w > pvB->w) ? pvA->w : pvB->w;
	
void vmp_QVecMaxAltivec0( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
#if 01						// with Unions
	vec vA, vB;

	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	vA.qv.w = pvA->w;

	vB.qv.x = pvB->x;
	vB.qv.y = pvB->y;
	vB.qv.z = pvB->z;
	vB.qv.w = pvB->w;
	
	vA.vf = vec_max( vA.vf, vB.vf );

	pvD->x = vA.qv.x;
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;
	pvD->w = vA.qv.w;
#else						// w/o Unions
	vector float vD, vA, vB;

	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;

	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	((vmp3DQVector *) &vB)->w = pvB->w;
	
	vD = vec_max( vA, vB );

	pvD->x = ((vmp3DQVector *) &vD)->x;
	pvD->y = ((vmp3DQVector *) &vD)->y;
	pvD->z = ((vmp3DQVector *) &vD)->z;
	pvD->w = ((vmp3DQVector *) &vD)->w;
#endif
}



//
//	(Quad) Vector Maximum - Aligned
//
	
void vmp_QVecMaxAltivec( vmp3DQVector * const pvD, const vmp3DQVector * const pvA, const vmp3DQVector * const pvB )
{
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);
	ASSERT_PTR16(pvB);

	*(vector float *)pvD = vec_max( (*(vector float *)pvA), (*(vector float *)pvB) );
}



//______________________________________________________
//
//	(Quad) Vector Magnitude - Unaligned
//
//	*pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );

void vmp_QVecMagnitudeAltivec0(float * const pfD, const vmp3DQVector * const pvA)
{
#if 01						// with Unions
	vec vA;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	vA.qv.w = 1.0;
	
	vA.vf = vec_madd( vA.vf, vA.vf, vZero );
	
	*pfD = sqrt( vA.qv.x + vA.qv.y + vA.qv.z );
#else						// w/o Unions
	vector float vA;
	float f;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 1.0;
	
	vA = vec_madd( vA, vA, vZero );
	
	f  = ((vmp3DVector *) &vA)->x
	   + ((vmp3DVector *) &vA)->y
	   + ((vmp3DVector *) &vA)->z;

	*pfD = sqrt(f);
#endif
}



//
//	(Quad) Vector Magnitude - Aligned
//

void vmp_QVecMagnitudeAltivec(float * const pfD, const vmp3DQVector * const pvA)
{
	vector float vA;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR16(pvA);

	vA = vec_madd( (*(vector float *)pvA), (*(vector float *)pvA), vZero );
	
	*pfD = sqrt(  ((vmp3DVector *) &vA)->x		// sqrt(x*x + y*y + z*z) 
				+ ((vmp3DVector *) &vA)->y
				+ ((vmp3DVector *) &vA)->z );
}



//______________________________________________________
//
//	(Quad) Vector Magnitude - (Fast) Unaligned
//
//	*pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//

void vmp_QVecMagnitudeFastAltivec0(float * const pfD, const vmp3DQVector * const pvA)
{
#if 01						// with Unions
	vec vA, vM;
	vector bool int viM;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;

	vA.vf = vec_madd( vA.vf, vA.vf, vZero );
	
	vA.qv.x += vA.qv.y + vA.qv.z;

	viM = vec_cmpge(vA.vf, vTiny);				// Set 1's in position of floats too close to zero

	vM.vf = vec_rsqrte(vA.vf);					//	 1 / sqrt(a)

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vM.vf = vec_and(vM.vf, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vM.vf = vec_or(vM.vf, viM);					// Blended #'s and 1.0's
	
	vM.vf = vec_madd( vA.vf, vM.vf, vZero );			// sqrt(a) = a * (1 / sqrt(a)) 
	*pfD = vM.qv.x;
#else						// w/o Unions
	vector float vA, vM;
	vector bool int viM;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;

	vA = vec_madd( vA, vA, vZero );
	
	((vmp3DVector *) &vA)->x = ((vmp3DVector *) &vA)->x
							 + ((vmp3DVector *) &vA)->y
							 + ((vmp3DVector *) &vA)->z;

	viM = vec_cmpge(vA, vTiny);			// Set 1's in position of floats too close to zero

	vM = vec_rsqrte(vA);				//	 1 / sqrt(a)

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vM = vec_and(vM, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vM = vec_or(vM, viM);				// Blended #'s and 1.0's

	vM = vec_madd( vA, vM, vZero );		// sqrt(a) = a * (1 / sqrt(a)) 
	*pfD = ((vmp3DVector *)&vM)->x;
#endif
}



//
//	(Quad) Vector Magnitude - (Fast) Aligned
//

void vmp_QVecMagnitudeFastAltivec(float * const pfD, const vmp3DQVector * const pvA)
{
	vector float vA, vM;
	vector bool int viM;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR16(pvA);

	vA = vec_madd( (*(vector float *)pvA), (*(vector float *)pvA), vZero );
	
	((vmp3DVector *) &vA)->x = ((vmp3DVector *) &vA)->x
							 + ((vmp3DVector *) &vA)->y
							 + ((vmp3DVector *) &vA)->z;

	viM = vec_cmpge(vA, vTiny);			// Set 1's in position of floats too close to zero

	vM = vec_rsqrte(vA);				//	 1 / sqrt(a)

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vM = vec_and(vM, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vM = vec_or(vM, viM);				// Blended #'s and 1.0's

	vM = vec_madd( vA, vM, vZero );		// sqrt(a) = a * (1 / sqrt(a)) 
	*pfD = ((vmp3DVector *)&vM)->x;
}



//______________________________________________________
//
//	(Quad) Vector Normalization - Unaligned
//
//	fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//		fMag = 1.0f / fMag;
//		pvD->x = pvA->x * fMag;
//		pvD->y = pvA->y * fMag;
//		pvD->z = pvA->z * fMag;

void vmp_QVecNormalizeAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
#if 01						// with Unions
	vec vA, vM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	vA.qv.x = pvA->x;							// Copy source vA = *pvA
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	vA.qv.w = pvA->w;
	
	vM.vf = vec_madd( vA.vf, vA.vf, vZero );	// M = A.x*A.x + A.y*A.y + A.z*A.z
	
		// No float MSUM function!	
	vM.qv.x = vM.qv.x + vM.qv.y + vM.qv.z;

	if ( vM.qv.x < SINGLE_PRECISION )			// If too close to zero...
	  {
	  }
	else
	  {
	  	vM.qv.x = 1.0f / sqrt(vM.qv.x);
	  	vM.vf = vec_splat(vM.vf, 0);
		
	  	vA.vf = vec_madd( vA.vf, vM.vf, vZero );
	  }
	  
	pvD->x = vA.qv.x;							// save resulting normalized vector
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;		
	pvD->w = 0.0;
#else						// w/o Unions
	vector float vA, vM;
	float fMag;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;
	
	vM = vec_madd( vA, vA, vZero );
	
		// No float MSUM function!	
	fMag = ((vmp3DQVector *) &vM)->x
	     + ((vmp3DQVector *) &vM)->y
	     + ((vmp3DQVector *) &vM)->z;

	if ( fMag < SINGLE_PRECISION )	// If too close to zero...
	  {
	  }
	else
	  {
		((vmp3DQVector *) &vM)->w = ((vmp3DQVector *) &vM)->x = ((vmp3DQVector *) &vM)->y = ((vmp3DQVector *) &vM)->z = 1.0 / sqrt(fMag);
		
	  	vA = vec_madd( vA, vM, vZero );
	  }
	  
	pvD->x = ((vmp3DVector *) &vA)->x;
	pvD->y = ((vmp3DVector *) &vA)->y;
	pvD->z = ((vmp3DVector *) &vA)->z;		
	pvD->w = 0.0;
#endif
}



//
//	(Quad) Vector Normalization - Aligned
//
//	fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//		fMag = 1.0f / fMag;
//		pvD->x = pvA->x * fMag;
//		pvD->y = pvA->y * fMag;
//		pvD->z = pvA->z * fMag;

void vmp_QVecNormalizeAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
	vector float vM;
	float fMag;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	vM = vec_madd( (*(vector float *)pvA), (*(vector float *)pvA), vZero );

		// No float MSUM function!	
	fMag = ((vmp3DVector *) &vM)->x
	     + ((vmp3DVector *) &vM)->y
	     + ((vmp3DVector *) &vM)->z;

	if ( fMag < SINGLE_PRECISION )	// If too close to zero...
	  {
		*pvD = *pvA;
	  }
	else
	  {
		((vmp3DQVector *) &vM)->w = ((vmp3DQVector *) &vM)->x = ((vmp3DQVector *) &vM)->y = ((vmp3DQVector *) &vM)->z = 1.0 / sqrt(fMag);
		
	  	*(vector float *)pvD = vec_madd( (*(vector float *)pvA), vM, vZero );
	  }
	  
	pvD->w = 0.0;
}



//______________________________________________________
//
//	(Quad) Vector Normalization (FAST) - unaligned
//
//	Note: Uses Estimated squareroot for speed
//
//	fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//		fMag = 1.0f / fMag;
//		pvD->x = pvA->x * fMag;
//		pvD->y = pvA->y * fMag;
//		pvD->z = pvA->z * fMag;

void vmp_QVecNormalizeFastAltivec0(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
#if 01					// with Unions
	vec vA, vM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	vA.qv.x = pvA->x;							// Copy source vA = *pvA
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	vA.qv.w = pvA->w;
	
	vM.vf = vec_madd( vA.vf, vA.vf, vZero );   	// [x*x y*y z*z w*w]
	
		// No float MSUM function!	
	vM.qv.x =  vM.qv.x + vM.qv.y + vM.qv.z;		// [# # # Mag]

	if ( vM.qv.x < SINGLE_PRECISION )			// If fMag too close to zero...
	  {											// preset to copy original Vector
	  }
	else
	  {
		vM.vf = vec_splat(vM.vf, 0);			// [fMag fMag fMag fMag]

			//	vD = 1 / sqrt(fMag)
		vM.vf = vec_rsqrte(vM.vf);				// vrsqrtefp d,a

			// sqrt(x)= x / sqrt(x)
		vA.vf = vec_madd( vA.vf, vM.vf, vZero ); 
	  }

	pvD->x = vA.qv.x;							// Save resulting normalized vector
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;		
	pvD->w = 0.0;								// fourth field
#else						// w/o Unions
	vector float vA, vM;
	float fMag;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;			// Copy source vA = *pvA
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = pvA->w;
	
	vM = vec_madd( vA, vA, vZero );   			// [x*x y*y z*z w*w]
	
		// No float MSUM function!	
	fMag =  ((vmp3DVector *) &vM)->x
		  + ((vmp3DVector *) &vM)->y
		  + ((vmp3DVector *) &vM)->z;

	if ( fMag < SINGLE_PRECISION )				// If too close to zero...
	  {											// preset to copy original Vector
	  }
	else
	  {
		((vmp3DQVector *)&vM)->x = fMag;		// [# # # Mag]

		vM = vec_splat(vM, 0);					// [fMag fMag fMag fMag]

			//	vD = 1 / sqrt(fMag)
		vM = vec_rsqrte(vM);					// vrsqrtefp d,a

			// sqrt(x)= x / sqrt(x)
		vA = vec_madd( vA, vM, vZero ); 
	  }

	pvD->x = ((vmp3DVector *) &vA)->x;			// save resulting normalized vector
	pvD->y = ((vmp3DVector *) &vA)->y;
	pvD->z = ((vmp3DVector *) &vA)->z;		
	pvD->w = 0.0;								// fourth field
#endif
}



//
//	(Quad) Vector Normalization - Aligned
//

void vmp_QVecNormalizeFastAltivec(vmp3DQVector * const pvD, const vmp3DQVector * const pvA)
{
	vector float vM;
	float fMag;
	
	ASSERT_PTR16(pvD);
	ASSERT_PTR16(pvA);

	vM = vec_madd( (*(vector float *)pvA), (*(vector float *)pvA), vZero );   // [x*x y*y z*z w*w]
	
		// No float MSUM function!	
	fMag =  ((vmp3DVector *) &vM)->x
		  + ((vmp3DVector *) &vM)->y
		  + ((vmp3DVector *) &vM)->z;

	if ( fMag < SINGLE_PRECISION )				// If too close to zero...
	  {
		*pvD = *pvA;							// Copy original Vector
	  }
	else
	  {
		((vmp3DQVector *)&vM)->x = fMag;		// [# # # Mag]

		vM = vec_splat(vM, 0);					// [fMag fMag fMag fMag]

			//	vD = 1 / sqrt(fMag)
		vM = vec_rsqrte(vM);					// vrsqrtefp d,a

			// sqrt(x)= x / sqrt(x)
		*(vector float *)pvD = vec_madd( *(vector float *)pvA, vM, vZero ); 
	  }

	pvD->w = 0.0;								// fourth field
}




//______________________________________________________
//
//	Setup Vec3D module
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void QVsf3DSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);

    if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_QVecSqrt =				vmp_QVecSqrtAltivec;
		vmp_QVecSqrt0 =				vmp_QVecSqrtAltivec0;
		vmp_QVecSqrtFast =			vmp_QVecSqrtFastAltivec;
		vmp_QVecSqrtFast0 =			vmp_QVecSqrtFastAltivec0;

		vmp_QVecMin =				vmp_QVecMinAltivec;
		vmp_QVecMin0 =				vmp_QVecMinAltivec0;
		vmp_QVecMax =				vmp_QVecMaxAltivec;
		vmp_QVecMax0 =				vmp_QVecMaxAltivec0;

		vmp_QVecMagnitude =			vmp_QVecMagnitudeAltivec;
		vmp_QVecMagnitude0 =		vmp_QVecMagnitudeAltivec0;
		vmp_QVecMagnitudeFast =		vmp_QVecMagnitudeFastAltivec;
		vmp_QVecMagnitudeFast0 =	vmp_QVecMagnitudeFastAltivec0;

		vmp_QVecNormalize =			vmp_QVecNormalizeAltivec;
		vmp_QVecNormalize0 =		vmp_QVecNormalizeAltivec0;
		vmp_QVecNormalizeFast = 	vmp_QVecNormalizeFastAltivec;
		vmp_QVecNormalizeFast0 =	vmp_QVecNormalizeFastAltivec0; 
	  }

	else
	  {
		vmp_QVecSqrt =				vmp_QVecSqrtGeneric;
		vmp_QVecSqrt0 =				vmp_QVecSqrtGeneric;
		vmp_QVecSqrtFast =			vmp_QVecSqrtGeneric;
		vmp_QVecSqrtFast0 =			vmp_QVecSqrtGeneric;

		vmp_QVecMin =				vmp_QVecMinGeneric;
		vmp_QVecMin0 =				vmp_QVecMinGeneric;
		vmp_QVecMax =				vmp_QVecMaxGeneric;
		vmp_QVecMax0 =				vmp_QVecMaxGeneric;

		vmp_QVecMagnitude =			vmp_QVecMagnitudeGeneric;
		vmp_QVecMagnitude0 =		vmp_QVecMagnitudeGeneric;
		vmp_QVecMagnitudeFast =		vmp_QVecMagnitudeGeneric;
		vmp_QVecMagnitudeFast0 =	vmp_QVecMagnitudeGeneric;

		vmp_QVecNormalize =			vmp_QVecNormalizeGeneric;
		vmp_QVecNormalize0 =		vmp_QVecNormalizeGeneric;
		vmp_QVecNormalizeFast = 	vmp_QVecNormalizeGeneric;
		vmp_QVecNormalizeFast0 =	vmp_QVecNormalizeGeneric;
	  }
}





