
//
//	Vsf3DAltivec.cpp - 3D Vector Special Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//	NOTE: Because these a 3 float Vectors the 4th float has to be handled as 
//		a special case thus slowing the code down. This is the same as an AoS 
//		orientation. Instead a SoA should be used in conjunction with the quad
//		set of functions.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>
#include    <math.h>

    //  Book Development Library Modules

#include	"CpuAsm.h"									// CPU
#include	"vmp3D.h"
#include	"vmp3DEx.h"

#include	"Vsf3D.h"									// 3D Math
#include	"Vsf3DEx.h"



vector float vHalf = (vector float) (0.5, 0.5, 0.5, 0.5);
vector float vOne = (vector float) (1.0, 1.0, 1.0, 1.0);
vector float vTiny = (vector float) (SINGLE_PRECISION, SINGLE_PRECISION, SINGLE_PRECISION, SINGLE_PRECISION);



//
//	Vector Square Root
//
//	pvD->x = sqrtf( pvA->x );
//	pvD->y = sqrtf( pvA->y );
//	pvD->z = sqrtf( pvA->z );
	
void vmp_VecSqrtFastAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
#if 01
	vec vA;
	vector float vD;
	vector bool int viM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;

	viM = vec_cmpge(vA.vf, vTiny);		// Set 1's in position of floats too close to zero
	
		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(vA.vf);				// vrsqrtefp d,a   Elements of zero are now infinite

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vD = vec_and(vD, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);				// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	vA.vf = vec_madd( vA.vf, vD, vZero ); 
	
	pvD->x = vA.qv.x;
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;
#else
	vector float vA, vD;
	vector bool int viM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 1.0;

	viM = vec_cmpge(vA, vTiny);			// Set 1's in position of floats too close to zero
	
		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(vA);				// vrsqrtefp d,a   Elements of zero are now infinite

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vD = vec_and(vD, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);				// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	vD = vec_madd( vA, vD, vZero ); 
	
	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
#endif
}



//______________________________________________________
//	Vector Square Root - Unaligned  (Normal 24bit Precision)
//
//	pvD->x = sqrtf( pvA->x );
//	pvD->y = sqrtf( pvA->y );
//	pvD->z = sqrtf( pvA->z );

void vmp_VecSqrtAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
#if 01						// with Unions
	vec vA;
	vector float vD, vT, vU;
	vector bool int viM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;

	viM = vec_cmpge(vA.vf, vTiny);		// Set 1's in position of floats too close to zero
	
		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(vA.vf);				// vrsqrtefp d,a   Elements of zero are now infinite

		//	Increase precision using Newton-Raphson refinement
		
	vT = vec_madd(vD,vD,vZero);						// y0*y0 +0
	vU = vec_madd(vD,vHalf,vZero);					// y0*0.5 +0
	vT = vec_nmsub(vA.vf,vT,vOne);					// t0 = B*t0-1)
	vD = vec_madd(vT,vU,vD);						// y1 = t0*t1+y0

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vD = vec_and(vD, viM);							// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);							// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	vA.vf = vec_madd( vA.vf, vD, vZero ); 
	
	pvD->x = vA.qv.x;
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;
#else						// w/o Unions
	vector float vA, vD, vT, vU;
	vector bool int viM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	ASSERT_FNEG(pvA->x);
	ASSERT_FNEG(pvA->y);
	ASSERT_FNEG(pvA->z);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	((vmp3DQVector *) &vA)->w = 1.0;

	viM = vec_cmpge(vA, vTiny);			// Set 1's in position of floats too close to zero
	
		//	vD = (1 / sqrt(vA))
	vD = vec_rsqrte(vA);				// vrsqrtefp d,a   Elements of zero are now infinite

		//	Increase precision using Newton-Raphson refinement
		
	vT = vec_madd(vD,vD,vZero);						// y0*y0 +0
	vU = vec_madd(vD,vHalf,vZero);					// y0*0.5 +0
	vT = vec_nmsub(vA,vT,vOne);						// t0 = B*t0-1)
	vD = vec_madd(vT,vU,vD);						// y1 = t0*t1+y0

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vD = vec_and(vD, viM);							// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vD = vec_or(vD, viM);							// Blended #'s and 1.0's

		// sqrt(x)= x * (1/sqrt(x))
	vD = vec_madd( vA, vD, vZero ); 
	
	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
#endif
}



//______________________________________________________
//
//	Vector Minimum
//
//	pvD->x = (pvA->x < pvB->x) ? pvA->x : pvB->x;
//	pvD->y = (pvA->y < pvB->y) ? pvA->y : pvB->y;
//	pvD->z = (pvA->z < pvB->z) ? pvA->z : pvB->z;
	
void vmp_VecMinAltivec( vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB )
{
#if 01						// with Unions
	vec vA, vB;

	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;

	vB.qv.x = pvB->x;
	vB.qv.y = pvB->y;
	vB.qv.z = pvB->z;
	
	vA.vf = vec_min( vA.vf, vB.vf );

	pvD->x = vA.qv.x;
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;
#else						// w/o Unions
	vector float vD, vA, vB;

	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;

	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	
	vD = vec_min( vA, vB );

	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
#endif
}



//______________________________________________________
//
//	Vector Maximum
//
//	pvD->x = (pvA->x > pvB->x) ? pvA->x : pvB->x;
//	pvD->y = (pvA->y > pvB->y) ? pvA->y : pvB->y;
//	pvD->z = (pvA->z > pvB->z) ? pvA->z : pvB->z;
	
void vmp_VecMaxAltivec( vmp3DVector * const pvD, const vmp3DVector * const pvA, const vmp3DVector * const pvB )
{
#if 01						// with Unions
	vec vA, vB;

	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;

	vB.qv.x = pvB->x;
	vB.qv.y = pvB->y;
	vB.qv.z = pvB->z;
	
	vA.vf = vec_max( vA.vf, vB.vf );

	pvD->x = vA.qv.x;
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;
#else						// w/o Unions
	vector float vD, vA, vB;

	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);
	ASSERT_PTR4(pvB);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;

	((vmp3DQVector *) &vB)->x = pvB->x;
	((vmp3DQVector *) &vB)->y = pvB->y;
	((vmp3DQVector *) &vB)->z = pvB->z;
	
	vD = vec_max( vA, vB );

	pvD->x = ((vmp3DVector *) &vD)->x;
	pvD->y = ((vmp3DVector *) &vD)->y;
	pvD->z = ((vmp3DVector *) &vD)->z;
#endif
}



//______________________________________________________
//
//	Vector Magnitude
//
//	*pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );

void vmp_VecMagnitudeAltivec(float * const pfD, const vmp3DVector * const pvA)
{
#if 01						// with Unions
	vec vA;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	
	vA.vf = vec_madd( vA.vf, vA.vf, vZero );
	
	*pfD = sqrt( vA.qv.x + vA.qv.y + vA.qv.z );
#else						// w/o Unions
	vector float vA;
	float f;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	
	vA = vec_madd( vA, vA, vZero );
	
	f  = ((vmp3DVector *) &vA)->x
	   + ((vmp3DVector *) &vA)->y
	   + ((vmp3DVector *) &vA)->z;

	*pfD = sqrt(f);
#endif
}



//______________________________________________________
//
//	Vector Magnitude - (Fast) (Un)aligned
//
//	*pfD = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//

void vmp_VecMagnitudeFastAltivec(float * const pfD, const vmp3DVector * const pvA)
{
#if 01						// with Unions
	vec vA, vM;
	vector bool int viM;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);

	vA.qv.x = pvA->x;
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;

	vA.vf = vec_madd( vA.vf, vA.vf, vZero );
	
	vA.qv.x += vA.qv.y + vA.qv.z;

	viM = vec_cmpge(vA.vf, vTiny);				// Set 1's in position of floats too close to zero

	vM.vf = vec_rsqrte(vA.vf);					//	 1 / sqrt(a)

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vM.vf = vec_and(vM.vf, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vM.vf = vec_or(vM.vf, viM);					// Blended #'s and 1.0's
	
	vM.vf = vec_madd( vA.vf, vM.vf, vZero );	// sqrt(a) = a * (1 / sqrt(a)) 
	*pfD = vM.qv.x;
#else						// w/o Unions
	vector float vA, vM;
	vector bool int viM;
	
	ASSERT_PTR4(pfD);
	ASSERT_PTR4(pvA);

	((vmp3DVector *) &vA)->x = pvA->x;
	((vmp3DVector *) &vA)->y = pvA->y;
	((vmp3DVector *) &vA)->z = pvA->z;

	vA = vec_madd( vA, vA, vZero );
	
	((vmp3DVector *) &vA)->x += ((vmp3DVector *) &vA)->y
							 +  ((vmp3DVector *) &vA)->z;

	viM = vec_cmpge(vA, vTiny);			// Set 1's in position of floats too close to zero
	vM = vec_rsqrte(vA);				//	 1 / sqrt(a)

		//	Correct for infinite due to 1/0 = 1/sart(0)
		
	vM = vec_and(vM, viM);				// Preserve #'s that are not too small, else zeros
	viM = vec_nor(viM, *((vector bool int *)&vZero) );			// Flip the bit masks (XOR)
	viM = vec_and(viM, *((vector bool int *)&vOne) );			// Preserve 1's (in positions that were infinite!)
	vM = vec_or(vM, viM);				// Blended #'s and 1.0's

	vA = vec_madd( vA, vM, vZero );		// sqrt(a) = a * (1 / sqrt(a)) 

	*pfD = ((vmp3DVector *)&vA)->x;
#endif
}



//______________________________________________________
//
//	Vector Normalization
//
//	fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//		fMag = 1.0f / fMag;
//		pvD->x = pvA->x * fMag;
//		pvD->y = pvA->y * fMag;
//		pvD->z = pvA->z * fMag;

void vmp_VecNormalizeAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
#if 01						// with Unions
	vec vA, vM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	vA.qv.x = pvA->x;							// Copy source vA = *pvA
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	
	vM.vf = vec_madd( vA.vf, vA.vf, vZero );	// M = A.x*A.x + A.y*A.y + A.z*A.z
	
		// No float MSUM function!	
	vM.qv.x = vM.qv.x + vM.qv.y + vM.qv.z;

	if ( vM.qv.x < SINGLE_PRECISION )			// If too close to zero...
	  {
	  }
	else
	  {
	  	vM.qv.x = 1.0f / sqrt(vM.qv.x);
	  	vM.vf = vec_splat(vM.vf, 0);
		
	  	vA.vf = vec_madd( vA.vf, vM.vf, vZero );
	  }
	  
	pvD->x = vA.qv.x;							// save resulting normalized vector
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;		
#else						// w/o Unions
	vector float vA, vM;
	float fMag;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	((vmp3DQVector *) &vA)->x = pvA->x;
	((vmp3DQVector *) &vA)->y = pvA->y;
	((vmp3DQVector *) &vA)->z = pvA->z;
	
	vM = vec_madd( vA, vA, vZero );
	
	fMag = ((vmp3DQVector *) &vM)->x
	     + ((vmp3DQVector *) &vM)->y
	     + ((vmp3DQVector *) &vM)->z;

	if ( fMag < SINGLE_PRECISION )	// If too close to zero...
	  {
	  }
	else
	  {
		((vmp3DQVector *) &vM)->x = ((vmp3DQVector *) &vM)->y = ((vmp3DQVector *) &vM)->z = 1.0 / sqrt(fMag);
		
	  	vA = vec_madd( vA, vM, vZero );
	  }
	  
	pvD->x = ((vmp3DVector *) &vA)->x;
	pvD->y = ((vmp3DVector *) &vA)->y;
	pvD->z = ((vmp3DVector *) &vA)->z;		
#endif
}



//______________________________________________________
//
//	(Quad) Vector Normalization (FAST) - unaligned
//
//	Note: Uses Estimated squareroot for speed
//
//	fMag = sqrtf( pvA->x * pvA->x + pvA->y * pvA->y + pvA->z * pvA->z );
//		fMag = 1.0f / fMag;
//		pvD->x = pvA->x * fMag;
//		pvD->y = pvA->y * fMag;
//		pvD->z = pvA->z * fMag;

void vmp_VecNormalizeFastAltivec(vmp3DVector * const pvD, const vmp3DVector * const pvA)
{
#if 01					// with Unions
	vec vA, vM;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	vA.qv.x = pvA->x;							// Copy source vA = *pvA
	vA.qv.y = pvA->y;
	vA.qv.z = pvA->z;
	
	vM.vf = vec_madd( vA.vf, vA.vf, vZero );   	// [x*x y*y z*z w*w]
	
		// No float MSUM function!	
	vM.qv.x =  vM.qv.x + vM.qv.y + vM.qv.z;		// [# # # Mag]

	if ( vM.qv.x < SINGLE_PRECISION )			// If fMag too close to zero...
	  {											// preset to copy original Vector
	  }
	else
	  {
		vM.vf = vec_splat(vM.vf, 0);			// [fMag fMag fMag fMag]

			//	vD = 1 / sqrt(fMag)
		vM.vf = vec_rsqrte(vM.vf);				// vrsqrtefp d,a

			// sqrt(x)= x / sqrt(x)
		vA.vf = vec_madd( vA.vf, vM.vf, vZero ); 
	  }

	pvD->x = vA.qv.x;							// Save resulting normalized vector
	pvD->y = vA.qv.y;
	pvD->z = vA.qv.z;		
#else						// w/o Unions
	vector float vA, vM;
	float fMag;
	
	ASSERT_PTR4(pvD);
	ASSERT_PTR4(pvA);

	((vmp3DVector *) &vA)->x = pvA->x;			// Copy source vA = *pvA
	((vmp3DVector *) &vA)->y = pvA->y;
	((vmp3DVector *) &vA)->z = pvA->z;
	
	vM = vec_madd( vA, vA, vZero );   			// [x*x y*y z*z]
	
		// No float MSUM function!	
	fMag =  ((vmp3DVector *) &vM)->x
		  + ((vmp3DVector *) &vM)->y
		  + ((vmp3DVector *) &vM)->z;

	if ( fMag < SINGLE_PRECISION )				// If too close to zero...
	  {											// preset to copy original Vector
	  }
	else
	  {
		((vmp3DQVector *)&vM)->x = fMag;		// [# # # Mag]

		vM = vec_splat(vM, 0);					// [fMag fMag fMag fMag]

			//	vD = 1 / sqrt(fMag)
		vM = vec_rsqrte(vM);					// vrsqrtefp d,a

			// sqrt(x)= x / sqrt(x)
		vA = vec_madd( vA, vM, vZero ); 
	  }

	pvD->x = ((vmp3DVector *) &vA)->x;			// save resulting normalized vector
	pvD->y = ((vmp3DVector *) &vA)->y;
	pvD->z = ((vmp3DVector *) &vA)->z;		
#endif
}



//______________________________________________________
//
//	Setup Vec3D module
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void Vsf3DSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR4(pcinfo);

    if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_VecSqrt =			vmp_VecSqrtAltivec;
		vmp_VecSqrt0 =			vmp_VecSqrtAltivec;
		vmp_VecSqrtFast =		vmp_VecSqrtFastAltivec;
		vmp_VecSqrtFast0 =		vmp_VecSqrtFastAltivec;

		vmp_VecMin =			vmp_VecMinAltivec;
		vmp_VecMin0 =			vmp_VecMinAltivec;
		vmp_VecMax =			vmp_VecMaxAltivec;
		vmp_VecMax0 =			vmp_VecMaxAltivec;

		vmp_VecMagnitude =		vmp_VecMagnitudeAltivec;
		vmp_VecMagnitude0 =		vmp_VecMagnitudeAltivec;
		vmp_VecMagnitudeFast =	vmp_VecMagnitudeFastAltivec;
		vmp_VecMagnitudeFast0 =	vmp_VecMagnitudeFastAltivec;

		vmp_VecNormalize =		vmp_VecNormalizeAltivec;
		vmp_VecNormalize0 =		vmp_VecNormalizeAltivec;
		vmp_VecNormalizeFast = 	vmp_VecNormalizeFastAltivec;
		vmp_VecNormalizeFast0 =	vmp_VecNormalizeFastAltivec;
      }

	else
	  {
		vmp_VecSqrt =			vmp_VecSqrtGeneric;
		vmp_VecSqrt0 =			vmp_VecSqrtGeneric;
		vmp_VecSqrtFast =		vmp_VecSqrtGeneric;
		vmp_VecSqrtFast0 =		vmp_VecSqrtGeneric;

		vmp_VecMin =			vmp_VecMinGeneric;
		vmp_VecMin0 =			vmp_VecMinGeneric;
		vmp_VecMax =			vmp_VecMaxGeneric;
		vmp_VecMax0 =			vmp_VecMaxGeneric;

		vmp_VecMagnitude =		vmp_VecMagnitudeGeneric;
		vmp_VecMagnitude0 =		vmp_VecMagnitudeGeneric;
		vmp_VecMagnitudeFast =	vmp_VecMagnitudeGeneric;
		vmp_VecMagnitudeFast0 =	vmp_VecMagnitudeGeneric;

		vmp_VecNormalize =		vmp_VecNormalizeGeneric;
		vmp_VecNormalize0 =		vmp_VecNormalizeGeneric;
		vmp_VecNormalizeFast = 	vmp_VecNormalizeGeneric;
		vmp_VecNormalizeFast0 =	vmp_VecNormalizeGeneric;
	  }
}





