
//
//  Bench.cpp - Vector 3D Trigonometric Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"

#include "VTrig3D.h"
#include "VTrig3DEx.h"



typedef enum
{
	iTEST_FLT_SINCOS,
	iTEST_FLT_SIN,
	iTEST_FLT_COS,
	iTEST_FLT_TAN
} iTEST;


 
//
//	Test Trignometric Functions
//

static void testTrig( iTEST iFunc )
{
    float fStep, fAngX, fAngX2;
	float f, f2, e, e2;
	float fPrecision;
	bool bDual;

	fPrecision = 0.0000001f;
	fPrecision = 0.000001f;	 //cossin sin cos
	fPrecision = 0.0001f;  //tan
	fStep = 10.0f;

		// Force proper alignment (Crude just for this demo)

	bDual = false;

	for ( fAngX = -360.0f; fAngX <= 360.0f; fAngX += fStep )
	  {
		fAngX2 = DEG2RAD( fAngX ); /// 182.04f;

		switch(iFunc)
		  {
			case iTEST_FLT_SINCOS:
				bDual = true;
				vmp_FSinCosGeneric( &f2, &e2, fAngX );
				vmp_FSinCos( &f, &e, fAngX );
				break;

			case iTEST_FLT_SIN:
				bDual = false;
				vmp_FSinGeneric( &f2, fAngX );
				vmp_FSin( &f, fAngX );
				break;

			case iTEST_FLT_COS:
				bDual = false;
				vmp_FCosGeneric( &f2, fAngX );
				vmp_FCos( &f, fAngX );
				break;

			case iTEST_FLT_TAN:
				bDual = false;
				vmp_FTanGeneric( &f2, fAngX );
				vmp_FTan( &f, fAngX );
				break;
		  }

		if (!vmp_IsFEqual( f2, f, fPrecision ))
		  {
			cout << f2 << " vs " << f << " failure! " << fAngX << endl;
		  }

		if (bDual)
		  {
			if (!vmp_IsFEqual( e2, e, fPrecision ))
			  {
				cout << e2 << " vs " << e << " failure! " << fAngX << endl;
			  }
		  }
	  }
}





int main( int argc, const char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

#pragma unused(argc, argv)

    CpuDetect( &cinfo );								// Detect CPU

    cout << "\nVector Trigonometry Test\n\n";
	cout << CpuInfoStr( szBuf, &cinfo ) << endl;

	VTrig3DSetup( &cinfo );								// Setup our 3D Vector stuff

	cout << "Testing Trig - Sine:Cosine" << endl;
	testTrig( iTEST_FLT_SINCOS );

	cout << "Testing Trig - Sine" << endl;
	testTrig( iTEST_FLT_SIN );

	cout << "Testing Trig - Cosine" << endl;
	testTrig( iTEST_FLT_COS );

	cout << "Testing Trig - Tangent" << endl;
	testTrig( iTEST_FLT_TAN );

	cout << "Fin!\n\n" << flush;

    return 0;
}



