
//
//  Bench.cpp - Vector 3D Matrix Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 5 Oct 2001 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//

#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"										// CPU Module
#include "vmp3D.h"

#include "Matrix3D.h"
#include "Matrix3DEx.h"




//
//	Matrix Multiplication
//
 
void vmp_MatrixMulFPU( vmp3DMatrix dMx, const vmp3DMatrix aMx, const vmp3DMatrix bMx )
{
	vmp3DMatrix mx;
	float *fD;
	const float *pu, *pv;
	uint u, v;

	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);
	ASSERT_PTR(bMx);

	fD = &mx[0][0];
	pu = &aMx[0][0];
	pv = &bMx[0][0];

	u = 4;
	do {
		v = 4;
		do {
			*fD++ =   *(pu+0) * *(pv+0) 
					+ *(pu+1) * *(pv+4)
					+ *(pu+2) * *(pv+8)
					+ *(pu+3) * *(pv+12);
			pv++;
		  } while(--v);

		pu+=4;
		pv-=4;
	  } while (--u);

	vmp_MatrixCopyGeneric(dMx, mx);
}




//
//	Test GenerateModel with Matrix Alignment
//

static void testGenerateModelAlign( void )
{
	float matrixbuf[9  *16 +3];	// enough space + 128bit alignment padding
    float fAngX, fAngY, fAngZ, fAngX2, fAngY2, fAngZ2, fStep;
	vmp3DMatrix *pmRotX, *pmRotY, *pmRotZ, *pmWorld, *pmTrans, *pmTmpX, *pmTmpY, *pmTmpZ, *pmTmpT;
	vmp3DMatrix mtx1, mtx2, world, world2;
	float fPrecision;

	fPrecision = 0.0000001f;
	fStep = 10.0f;

		// Force proper alignment (Crude just for this demo)

	pmRotX = (vmp3DMatrix*) ALIGN16((int)(matrixbuf));   // Matrix Alignment
	pmRotY = pmRotX + 1;
	pmRotZ = pmRotY + 1;
	pmWorld = pmRotZ + 1;
	pmTrans = pmWorld + 1;
	pmTmpX = pmTrans + 1;
	pmTmpY = pmTmpX + 1;
	pmTmpZ = pmTmpY + 1;
	pmTmpT = pmTmpZ + 1;

	for ( fAngX = -360.0f; fAngX <= 360.0f; fAngX += fStep )
	  {
		for ( fAngY = -180.0f; fAngY <= 180.0f; fAngY += fStep )
		  {
			for ( fAngZ = -360.0f; fAngZ <= 360.0f; fAngZ += fStep )
			  {
			    fAngX2 = DEG2RAD( fAngX ); /// 182.04f;
			    fAngY2 = DEG2RAD( fAngY ); /// 182.04f;
			    fAngZ2 = DEG2RAD( fAngZ ); /// 182.04f;

			    vmp_MatrixSetRotateXGeneric( *pmTmpX, fAngX2 );
				vmp_MatrixSetRotateYGeneric( *pmTmpY, fAngY2 );
			    vmp_MatrixSetRotateZGeneric( *pmTmpZ, fAngZ2 );

			    vmp_MatrixSetRotateX0( *pmRotX, fAngX2 );
				vmp_MatrixSetRotateY0( *pmRotY, fAngY2 );
			    vmp_MatrixSetRotateZ0( *pmRotZ, fAngZ2 );

			    vmp_MatrixSetTranslateGeneric( *pmTrans, fAngX, fAngY, fAngZ );
			    vmp_MatrixSetTranslate( *pmTmpT, fAngX, fAngY, fAngZ );


					// Set a World View using Rotation Matrices
				vmp_MatrixMulGeneric( world2, *pmTmpY, *pmTmpX );
				vmp_MatrixMulGeneric( world2, world2, *pmTmpZ );
//				vmp_MatrixMulGeneric( world2, world2, *pmTmpT );

					//	Set a World View using Rotation Angles
				vmp_MatrixSetIdentityGeneric( world );
				vmp_MatrixRotateXGeneric( world, world, fAngX2 );
				vmp_MatrixRotateYGeneric( world, world, fAngY2 );
				vmp_MatrixRotateZGeneric( world, world, fAngZ2 );
//				vmp_MatrixTranslateGeneric( world, world, fAngX, fAngY, fAngZ );



				vmp_MatrixMulGeneric( mtx1, world, world );
				vmp_MatrixMulFPU( mtx2, world, world );

//				vmp_SIMDEntry();

//				vmp_SIMDExit();

				if ( !vmp_IsMatrixEqual( *pmTmpT, *pmTrans, fPrecision ))
				  {
					cout << "Matrix Trans failure!" << endl;
				  }

				if ( !vmp_IsMatrixEqual( *pmTmpX, *pmRotX, fPrecision ))
				  {
					cout << "Matrix X failure!" << endl;
				  }

				if ( !vmp_IsMatrixEqual( *pmTmpY, *pmRotY, fPrecision ))
				  {
					cout << "Matrix Y failure!" << endl;
				  }

				if ( !vmp_IsMatrixEqual( *pmTmpZ, *pmRotZ, fPrecision ))
				  {
					cout << "Matrix Z failure!" << endl;
				  }

				if ( !vmp_IsMatrixEqual( mtx1, mtx2, fPrecision ))
				  {
					cout << "Matrix MUL failure!" << endl;
				  }

				if ( !vmp_IsMatrixEqual( world, world2, fPrecision ))
				  {
					cout << "Matrix World View failure!" << endl;
				  }

				
			  }
		  }
	  }
}



int main( int argc, const char *argv[] )
{
    CpuInfo cinfo;
	char szBuf[ CPU_SZBUF_MAX ];

#pragma unused(argc, argv)

    CpuDetect( &cinfo );								// Detect CPU

    cout << "\nVector Matrix Test\n\n";
	cout << CpuInfoStr( szBuf, &cinfo ) << endl;

	Matrix3DSetup( &cinfo );								// Setup our 3D Vector stuff

	cout << "Testing Translate, {X,Y,Z} Rotations" << endl;
	testGenerateModelAlign();

	cout << "Fin!\n\n" << flush;

    return 0;
}

