
//
//	Matrix3D.cpp - 3D Vector Matrix Function
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>
#include    <math.h>

    //  Book Development Library Modules

#include	"vmp3D.h"
#include	"vmp3DEx.h"
#include	"Matrix3D.h"									// 3D Math
#include	"Matrix3DEx.h"




vmp_MatrixSetIdentityProc	vmp_MatrixSetIdentity;

vmp_MatrixSetTranslateProc vmp_MatrixSetTranslate;

vmp_MatrixSetRotateProc vmp_MatrixSetRotateX;
vmp_MatrixSetRotateProc vmp_MatrixSetRotateX0;
vmp_MatrixSetRotateProc vmp_MatrixSetRotateY;
vmp_MatrixSetRotateProc vmp_MatrixSetRotateY0;
vmp_MatrixSetRotateProc vmp_MatrixSetRotateZ;
vmp_MatrixSetRotateProc vmp_MatrixSetRotateZ0;

vmp_MatrixCopyProc		vmp_MatrixCopy;
vmp_MatrixMulProc		vmp_MatrixMul;
vmp_MatrixTranslateProc	vmp_MatrixTranslate;
vmp_MatrixRotateXProc	vmp_MatrixRotateX;
vmp_MatrixRotateYProc	vmp_MatrixRotateY;
vmp_MatrixRotateZProc	vmp_MatrixRotateZ;





//                            | 1 0 0 0 |
//	Matrix - Set Identity     | 0 1 0 0 |
//                            | 0 0 1 0 |
//                            | 0 0 0 1 |

void vmp_MatrixSetIdentityGeneric( vmp3DMatrix Mx )
{
	ASSERT_PTR(Mx);

	              Mx[0][1]=     Mx[0][2]=     Mx[0][3] = 
	Mx[1][0]=                   Mx[1][2]=     Mx[1][3] = 
	Mx[2][0]=     Mx[2][1]=                   Mx[2][3] = 
	Mx[3][0]=     Mx[3][1]=     Mx[3][2]=                0.0f;

	Mx[0][0]=     Mx[1][1]=     Mx[2][2]=     Mx[3][3]=  1.0f;
}



//                            | 1  0  0  0 |
//	Matrix - Set Translation  | 0  1  0  0 |
//                            | 0  0  1  0 |
//                            | Tx Ty Tz 1 |

void vmp_MatrixSetTranslateGeneric( vmp3DMatrix Mx, float fTx, float fTy, float fTz )
{
	ASSERT_PTR(Mx);

	              Mx[0][1]=     Mx[0][2]=     Mx[0][3] = 
	Mx[1][0]=                   Mx[1][2]=     Mx[1][3] = 
	Mx[2][0]=     Mx[2][1]=                   Mx[2][3] = 0.0f;
	Mx[0][0]=     Mx[1][1]=     Mx[2][2]=     Mx[3][3] = 1.0f;

    Mx[3][0]=fTx; Mx[3][1]=fTy; Mx[3][2]=fTz;
}



//                            | 1   0   0   0 |
//	Matrix - Set X Rotation   | 0  cos sin  0 |
//                            | 0 -sin cos  0 |
//                            | 0   0   0   1 |

void vmp_MatrixSetRotateXGeneric( vmp3DMatrix Mx, float fRads )
{
	ASSERT_PTR(Mx);

	vmp_MatrixSetIdentityGeneric( Mx );

    Mx[1][2] =               sinf( fRads );
    Mx[2][2] =  Mx[1][1] =   cosf( fRads );
    Mx[2][1] = -Mx[1][2]; //-sinf
}



//                            | cos  0 -sin  0 |
//	Matrix - Set Y Rotation   |  0   1   0   0 |
//                            | sin  0  cos  0 |
//                            |  0   0   0   1 |

void vmp_MatrixSetRotateYGeneric( vmp3DMatrix Mx, float fRads )
{
	ASSERT_PTR(Mx);

	vmp_MatrixSetIdentityGeneric( Mx );

    Mx[2][0] =               sinf( fRads );
    Mx[2][2] =  Mx[0][0] =   cosf( fRads );
    Mx[0][2] = -Mx[2][0]; //-sinf
}



//                            |  cos sin  0   0 |
//	Matrix - Set Z Rotation   | -sin cos  0   0 |
//                            |   0   0   1   0 |
//                            |   0   0   0   1 |

void vmp_MatrixSetRotateZGeneric( vmp3DMatrix Mx, float fRads )
{
	ASSERT_PTR(Mx);

	vmp_MatrixSetIdentityGeneric( Mx );

    Mx[0][1] =               sinf( fRads );
    Mx[1][1] =  Mx[0][0] =   cosf( fRads );
    Mx[1][0] = -Mx[0][1]; //-sinf
}



//
//	Matrix - Copy  mD = mA
//

void vmp_MatrixCopyGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx )
{
	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);

	dMx[0][0]=aMx[0][0];  dMx[0][1]=aMx[0][1];  dMx[0][2]=aMx[0][2];  dMx[0][3]=aMx[0][3];
	dMx[1][0]=aMx[1][0];  dMx[1][1]=aMx[1][1];  dMx[1][2]=aMx[1][2];  dMx[1][3]=aMx[1][3];
	dMx[2][0]=aMx[2][0];  dMx[2][1]=aMx[2][1];  dMx[2][2]=aMx[2][2];  dMx[2][3]=aMx[2][3];
	dMx[3][0]=aMx[3][0];  dMx[3][1]=aMx[3][1];  dMx[3][2]=aMx[3][2];  dMx[3][3]=aMx[3][3];
}



//
//	Matrix - Transpose
//

void vmp_MatrixTransposeGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx )
{
	float f[6];

	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);

	dMx[0][0]=aMx[0][0]; dMx[1][1]=aMx[1][1]; dMx[2][2]=aMx[2][2]; dMx[3][3]=aMx[3][3];

	     f[0]=aMx[0][1];      f[1]=aMx[0][2];      f[2]=aMx[0][3];
                              f[3]=aMx[1][2];      f[4]=aMx[1][3];
                                                   f[5]=aMx[2][3];
	dMx[0][1]=aMx[1][0]; dMx[1][0]=f[0];
    dMx[0][2]=aMx[2][0]; dMx[2][0]=f[1];
    dMx[0][3]=aMx[3][0]; dMx[3][0]=f[2];  
	dMx[1][2]=aMx[2][1]; dMx[2][1]=f[3];
	dMx[1][3]=aMx[3][1]; dMx[3][1]=f[4]; 
	dMx[2][3]=aMx[3][2]; dMx[3][2]=f[5];
}



//
//	Matrix - Inverse   mD = - mA
//

void vmp_MatrixInvGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx )
{
	float f[3];

	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);

		//	First Transpose

	f[0] = -(aMx[0][3]*aMx[0][0] + aMx[1][3]*aMx[1][0] + aMx[2][3]*aMx[2][0]);
	f[1] = -(aMx[0][3]*aMx[0][1] + aMx[1][3]*aMx[1][1] + aMx[2][3]*aMx[2][1]);
	f[2] = -(aMx[0][3]*aMx[0][2] + aMx[1][3]*aMx[1][2] + aMx[2][3]*aMx[2][2]);

		//	Now swap cross diagnals

	dMx[0][0]=aMx[0][0];  dMx[1][0]=aMx[0][1];  dMx[2][0]=aMx[0][2];  dMx[3][0]=0.0f;
	dMx[0][1]=aMx[1][0];  dMx[1][1]=aMx[1][1];  dMx[2][1]=aMx[1][2];  dMx[3][1]=0.0f;
	dMx[0][2]=aMx[2][0];  dMx[1][2]=aMx[2][1];  dMx[2][2]=aMx[2][2];  dMx[3][2]=0.0f;
	dMx[0][3]=f[0];       dMx[1][3]=f[1];       dMx[2][3]=f[2];       dMx[3][3]=1.0f;
}



//
//	Matrix Multiplication   mD = mA x mB
//
void vmp_MatrixMulGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx, const vmp3DMatrix bMx )
{
	vmp3DMatrix mx;
	uint u, v;

	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);
	ASSERT_PTR(bMx);

	for ( u=0; u < 4; u++ )
	  {
		for ( v = 0; v < 4; v++ )
		  {
			mx[u][v] = aMx[u][0] * bMx[0][v] 
					 + aMx[u][1] * bMx[1][v]
					 + aMx[u][2] * bMx[2][v]
					 + aMx[u][3] * bMx[3][v];
		  }
	  }

	vmp_MatrixCopyGeneric(dMx, mx);
}




//                       | 00 01 02 03 |   | 1  0  0  0 |
//	Matrix - Translate   | 10 11 12 13 |   | 0  1  0  0 |
//                     = | 20 21 22 23 | * | 0  0  1  0 |
//                       | 30 31 32 33 |   | Tx Ty Tz 1 |
//
//	NOTE: Copy and Translate

void vmp_MatrixTranslateGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx, float fTx, float fTy, float fTz )
{
	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);

	vmp_MatrixCopyGeneric( dMx, aMx );				// copy the matrix

		//	Now set T{x,y,z}

	dMx[3][0] += fTx;
	dMx[3][1] += fTy;
	dMx[3][2] += fTz;
}




//                        | 00 01 02 03 |   | 1   0   0   0 |
//	Matrix - Rotation X   | 10 11 12 13 |   | 0  cos sin  0 |
//                      = | 20 21 22 23 | * | 0 -sin cos  0 |
//                        | 30 31 32 33 |   | 0   0   0   1 |
//
//	NOTE: Multiply two Matrices

void vmp_MatrixRotateXGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx, float fRads )
{
	vmp3DMatrix mx;

	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);

	vmp_MatrixSetIdentityGeneric( mx );

    mx[1][2] =               sinf( fRads );
    mx[2][2] =  mx[1][1] =   cosf( fRads );
    mx[2][1] = -mx[1][2]; //-sinf

	vmp_MatrixMulGeneric( dMx, mx, aMx );
}



//                        | 00 01 02 03 |   | cos  0 -sin  0 |
//	Matrix - Rotation Y   | 10 11 12 13 |   |  0   1   0   0 |
//                      = | 20 21 22 23 | * | sin  0  cos  0 |
//                        | 30 31 32 33 |   |  0   0   0   1 |

void vmp_MatrixRotateYGeneric( vmp3DMatrix dMx, const vmp3DMatrix aMx, float fRads )
{
	vmp3DMatrix mx;

	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);

	vmp_MatrixSetIdentityGeneric( mx );

    mx[2][0] =               sinf( fRads );
    mx[2][2] =  mx[0][0] =   cosf( fRads );
    mx[0][2] = -mx[2][0]; //-sinf

	vmp_MatrixMulGeneric( dMx, mx, aMx );
}



//                        | 00 01 02 03 |   |  cos sin  0   0 |
//	Matrix - Rotation Z   | 10 11 12 13 |   | -sin cos  0   0 |
//                      = | 20 21 22 23 | * |   0   0   1   0 |
//                        | 30 31 32 33 |   |   0   0   0   1 |

void vmp_MatrixRotateZGeneric( vmp3DMatrix dMx, vmp3DMatrix aMx, float fRads )
{
	vmp3DMatrix mx;

	ASSERT_PTR(dMx);
	ASSERT_PTR(aMx);

	vmp_MatrixSetIdentityGeneric( mx );

    mx[0][1] =               sinf( fRads );
    mx[1][1] =  mx[0][0] =   cosf( fRads );
    mx[1][0] = -mx[0][1]; //-sinf

	vmp_MatrixMulGeneric( dMx, mx, aMx );
}





