
//
//  MCompAltivec.cpp - Motion Compensation
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 21 Feb, 2002 - James Leiterman
//
//  Last Modification
//
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>
#include <stdio.h>
    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "MComp.h"
#include "MCompEx.h"



//
//	 8x8 Horizontal Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][X]		(A+B+1)>>1
//	[ ][ ]

void vmp_MCompHz8RAltivec( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
// 	vector unsigned char vR;
	vector unsigned char vS, vR, vD;
	const uint8 *pS, *pR;
	uint dy, lshift, rshift;

//	if (0 != (15 & ((int)pbSrc)))
//	  {
//		vmp_MCompHz8RPPC(pbDst, pbSrc, iStride);
//		return;
//	  }
//
//	ASSERT_PTR16(pbDst);
//	ASSERT_PTR16(pbSrc);
//	ASSERT_ZERO(iStride);

	dy = 8;
	lshift = ((uint) pbSrc)    & 0x0000000F;
	rshift = ((uint)(pbSrc+1)) & 0x0000000F;
	pS = (pbSrc+0) - lshift;
	pR = (pbSrc+1) - rshift;
cout << lshift << " r " << rshift << endl;

	do {	// Vertical Interpolation with rounding (vavgub d,a,b)
		vS = vec_lvsl( lshift, pS );				// 0[123456789ABCDEF 10]
		vR = vec_lvsl( rshift, pR );				// 0[123456789ABCDEF 10]
		vD = vec_avg( vR, vS );						//  [0123456789ABCDEF]

		pS += iStride;
		pR += iStride;
		
			//	Only save first 8 bytes
		*(((uint32 *)pbDst)+0) = *(((uint32 *)&vD)+0);
		*(((uint32 *)pbDst)+1) = *(((uint32 *)&vD)+1);
		pbDst += iStride;
	} while (--dy);
}



//
//	 16x16 Horizontal Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][X]		(A+B+1)>>1
//	[ ][ ]

void vmp_MCompHz16RAltivec( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	vector unsigned char vS;
	uint dy;

	if (0 != (15 & ((int)pbSrc)))
	  {
		vmp_MCompHz16RPPC(pbDst, pbSrc, iStride);
		return;
	  }

	ASSERT_PTR16(pbDst);
	ASSERT_PTR16(pbSrc);
	ASSERT_ZERO(iStride);

	dy = 8;

	do {	// Vertical Interpolation with rounding (vavgub d,a,b)

		vS = vec_lvsl( 1, pbSrc );														// 0[123456789ABCDEF 10]
		*(vector unsigned char *)pbDst = vec_avg( *(vector unsigned char *)pbSrc, vS );	//  [0123456789ABCDEF]

		pbSrc += iStride;
		pbDst += iStride;
	} while (--dy);
}



//_______________________________________________________________________________________



//
//	 8x8 Vertical Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][ ]		(A+B+1)>>1
//	[X][ ]

void vmp_MCompVt8RAltivec( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
	vector unsigned char vD;
	const uint8 *pR;
	uint dy;

	if (0 != (15 & ((int)pbSrc)))
	  {
		vmp_MCompVt8RPPC(pbDst, pbSrc, iStride);
		return;
	  }

	ASSERT_PTR16(pbDst);
	ASSERT_PTR16(pbSrc);
	ASSERT_ZERO(iStride);

	pR = pbSrc + iStride;			

	dy = 16;
	do {	// Vertical Interpolation with rounding (vavgub d,a,b)
		vD = vec_avg( *(vector unsigned char *)pbSrc, *(vector unsigned char *)pR );

		pbSrc = pR;
		pR    += iStride;

			//	Only save first 8 bytes
		*(((uint32 *)pbDst)+0) = *(((uint32 *)&vD)+0);
		*(((uint32 *)pbDst)+1) = *(((uint32 *)&vD)+1);
		pbDst += iStride;
	} while (--dy);
}



//
//	 16x16 Vertical Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][ ]		(A+B+1)>>1
//	[X][ ]

void vmp_MCompVt16RAltivec( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{
#if 01
	vector unsigned char vD, vS, vR, vShift;
	const uint8 *pS;
	uint dy, lshift;

	ASSERT_PTR(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

	vShift = vec_lvsl( 0, pbSrc );

	lshift = ((uint) pbSrc) & 0x0000000F;
	pS = pbSrc - lshift;
//pS =(unsigned char *)&vS;
	
	dy = 16;
	do {	// Vertical Interpolation with rounding (vavgub d,a,b)

printf( "0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
*(pbSrc+0), *(pbSrc+1), *(pbSrc+2), *(pbSrc+3), *(pbSrc+4), *(pbSrc+5), *(pbSrc+6), *(pbSrc+7) );

//		vS = vec_lvsl( lshift, pS );				// 0[123456789ABCDEF 10]
//		vS = vec_lvsl( 0, pbSrc );				// 0[123456789ABCDEF 10]
//vperm

		vS = vec_lvsl( 0, pbSrc );				// 0[123456789ABCDEF 10]
vD = vec_perm( vS, vS, vShift );

pS =(unsigned char *)&vS;
printf( "0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
*(pS+0), *(pS+1), *(pS+2), *(pS+3), *(pS+4), *(pS+5), *(pS+6), *(pS+7) );

pS =(unsigned char *)&vD;
printf( "0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x\n",
*(pS+0), *(pS+1), *(pS+2), *(pS+3), *(pS+4), *(pS+5), *(pS+6), *(pS+7) );


//		vS = vec_ld( , (vector unsigned char *)pS );				// 0[123456789ABCDEF 10]
		pS += iStride;
//		pbSrc += iStride;
//		vR = vec_lvsl( lshift, pS );				// 0[123456789ABCDEF 10]
		vR = vec_lvsl( 0, pbSrc );				// 0[123456789ABCDEF 10]
//		vR = vec_ld( vShift, pS );				// 0[123456789ABCDEF 10]
break;
		*(vector unsigned char *)pbDst = vec_avg( vS, vR );
		pbDst += iStride;
	} while (--dy);
#else
	const uint8 *pR;
	uint dy;

	if (0 != (15 & ((int)pbSrc)))
	  {
		vmp_MCompVt16RPPC(pbDst, pbSrc, iStride);
		return;
	  }

	ASSERT_PTR16(pbDst);
	ASSERT_PTR16(pbSrc);
	ASSERT_ZERO(iStride);

	pR = pbSrc + iStride;			
	
	dy = 16;
	do {	// Vertical Interpolation with rounding (vavgub d,a,b)
		*(vector unsigned char *)pbDst = vec_avg( *(vector unsigned char *)pbSrc, *(vector unsigned char *)pR );

		pbSrc = pR;
		pR    += iStride;
		pbDst += iStride;
	} while (--dy);
#endif
}



//
//	Setup Motion Compensation
//
//	NOTE: This function sets up the function vectors so the best function
//		is attached to the vector for the processor running.
//

void MCompSetup( const CpuInfo * const pcinfo )
{
	ASSERT_PTR(pcinfo);
	    
	if ( CPUBITS_ALTIVEC & pcinfo->nBits )
      {
		vmp_MCompHz8R =			vmp_MCompHz8RAltivec;
		vmp_MCompHz16R = 		vmp_MCompHz16RAltivec;
		vmp_MCompVt8R =			vmp_MCompVt8RAltivec;
		vmp_MCompVt16R = 		vmp_MCompVt16RAltivec;
      }

	else
	  {
		vmp_MCompHz8R =			vmp_MCompHz8RPPC;
		vmp_MCompHz16R = 		vmp_MCompHz16RPPC;
		vmp_MCompVt8R =			vmp_MCompVt8RPPC;
		vmp_MCompVt16R = 		vmp_MCompVt16RPPC;
	  }
}




