
//
//  MCompPPC.cpp - Packed Motion Compensation
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 21 Feb, 2002 - James Leiterman
//
//  Last Modification
//
//	NOTE: 8x8 and 16x16 SOURCE Prediction Blocks are not always aligned!!!
//	If they were, the code would be MUCH faster!
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "MComp.h"
#include "MCompEx.h"



//______________________________________________________
//
//	 8x8 Horizontal Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][X]		(A+B+1)>>1
//	[ ][ ]

void vmp_MCompHz8RPPC( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{//                              r3                   r4          r5
	unsigned int register e01, e7F, dy, c0, c1, d0, d1, t0, t1;
	unsigned int register lshift, rshift, lmask, rmask;

#ifndef USE_ASSERT
#pragma unused(pbDst, pbSrc, iStride)
#endif

	ASSERT_PTR4(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

__asm {	
			//	test and branch if pbSrc not aligned!

		andi.	r0,r4,3			// Branch if NOT four byte aligned?
		cmpwi	r0,0
		bne		badalign

			/********************\
			|* pbSrc IS aligned *|
			\********************/

		lis		e01,257			// =01010000
		lis		e7F,32639		// =7F7F0000

		li		dy,8			// dy=8  # of scanlines (loops)
		
		addi	e01,e01,257		// =01010101  Error Correction 01
		addi	e7F,e7F,32639	// =7F7F7F7F  Error Correction 7F
		
rem0:	lwz		c0,0(r4)		//  [0123] = *(pbSrc+0)
		lwz		c1,4(r4)		//  [4567] = *(pbSrc+4)
		lwz		d1,8(r4)		//  [89AB] = *(pbSrc+8)

		slwi	d0,c0,8			// 0[123_]
		srwi	d1,d1,24		//  [___8]9AB
		rlwimi	d0,c1,8,24,31	//  [1234]
		rlwimi	d1,c1,8,0,23	//  [5678]
		
//		*(pDst+0) = (( c0 >> 1 ) & eMASK7F) + ((d0 >> 1) & eMASK7F) + ((d0 | c0) & eMASK01);
//		*(pDst+1) = (( c1 >> 1 ) & eMASK7F) + ((d1 >> 1) & eMASK7F) + ((d1 | c1) & eMASK01);

		or		t0,c0,d0		// (c0 | d0)  [0123] | [1234]
		or		t1,c1,d1		// (c1 | d1)  [4567] | [5678]
		srwi	c0,c0,1			// (c0 >> 1)
		srwi	d0,d0,1			// (d0 >> 1)
		srwi	c1,c1,1			// (c1 >> 1)
		srwi	d1,d1,1			// (d1 >> 1)
		
		and		t0,t0,e01		// ((c0 | d0) & e01)
		and		t1,t1,e01		// ((c1 | d1) & e01)
		and		c0,c0,e7F		// ((c0 >> 1) & e7F)
		and		d0,d0,e7F		// ((d0 >> 1) & e7F)
		and		c1,c1,e7F		// ((c1 >> 1) & e7F)
		and		d1,d1,e7F		// ((d1 >> 1) & e7F)
		
		add		t0,t0,c0
		add		t1,t1,c1
		add		t0,t0,d0
		add		t1,t1,d1

		add		r4,r4,r5		// pbSrc += iStride 
		subi	dy,dy,1			// dy--
		
		stw		t0,0(r3)		// *(pbDst+0) = 
		stw		t1,4(r3)		// *(pbDst+4) =
		add		r3,r3,r5		// pbDst += iStride
		
		cmpwi	dy,0			// Loop 8 times
		bne		rem0

		b		xit				// Return

			/*********************\
			|* pbSrc NOT aligned *|
			\*********************/

badalign:  	// r0={1...3} offset
		li		dy,-4					// =FFFFFFFC
		slwi	lshift,r0,3				// x8  1=8, 2=16, 3=24 bits to shift
		and		r4,r4,dy				// Set pbSrc to mod 4 = 0

			//	Generate shifting masks
		
		li		rshift,32
		li		rmask,-1				//  =FFFFFFFF
		sub		rshift,rshift,lshift	// 24=32-8  16=32-16  8=32-24   (n=32-n)
		srw		lmask,rmask,rshift		// 8=000000FF 16=0000FFFF 24=00FFFFFF
		xor		rmask,rmask,lmask		// 8=FFFFFF00 16=FFFF0000 24=FF000000

			// Standard preparation (just like mod 4=0) {properly aligned src}

		lis		e7F,32639		// =7F7F0000
		lis		e01,257			// =01010000

		li		dy,8			// dy=8 # of scanlines
				
		addi	e7F,e7F,32639	// =7F7F7F7F  Error Correction 7F
		addi	e01,e01,257	 	// =01010101  Error Correction 01

			//	lshift=(left shift) rshift=(right shift) lmask=(left mask)
			//	NOTE: r4 is either pbSrc -1,-2,-3  actual data!

									//   1=8   2=16   3=24   (offset to left shift to correct!)
remX:	lwz		c0,0(r4)	 		// [X012] [XX01] [XXX0] = *(pbSrc+0) 
		lwz		c1,4(r4)	 		// [X456] [XX45] [XXX4] = *(pbSrc+4)
		lwz		d1,8(r4)	 		// [X89A] [XX89] [XXX8] = *(pbSrc+8)

		slw		c0,c0,lshift	 	// [012_] [01__] [0___]
		rlwnm	c1,c1,lshift,0,31	// [4563] [4523] [4123]
		rlwnm	d1,d1,lshift,0,31	// [89A7] [8967] [8567]

		and		r0,c1,lmask			// [0003] [0023] [0123]
		and		c1,c1,rmask			// [4560] [4500] [4000]
		and		d0,d1,lmask			// [0007] [0067] [0567]

		or		c0,c0,r0			// [0123] [0123] [0123]
		or		c1,c1,d0			// [4567] [4567] [4567]
			//			d1			   [89A7]		
			// *** d1:c1:c0 Source Data now properly aligned! Handle as normal ***
		
		slwi	d0,c0,8			// 0[123_]
		srwi	d1,d1,24		//  [___8]9AB
		rlwimi	d0,c1,8,24,31	//  [1234]
		rlwimi	d1,c1,8,0,23	//  [5678]

//		*(pDst+0) = (( c0 >> 1 ) & eMASK7F) + ((d0 >> 1) & eMASK7F) + ((d0 | c0) & eMASK01);
//		*(pDst+1) = (( c1 >> 1 ) & eMASK7F) + ((d1 >> 1) & eMASK7F) + ((d1 | c1) & eMASK01);
		
		or		t0,c0,d0		// (c0 | d0)
		or		t1,c1,d1		// (c1 | d1)
		srwi	c0,c0,1			// (c0 >> 1)
		srwi	d0,d0,1			// (d0 >> 1)
		srwi	c1,c1,1			// (c1 >> 1)
		srwi	d1,d1,1			// (d1 >> 1)
		
		and		t0,t0,e01		// ((c0 | d0) & e01)
		and		t1,t1,e01		// ((c1 | d1) & e01)
		and		c0,c0,e7F		// ((c0 >> 1) & e7F)
		and		d0,d0,e7F		// ((d0 >> 1) & e7F)
		and		c1,c1,e7F		// ((c1 >> 1) & e7F)
		and		d1,d1,e7F		// ((d1 >> 1) & e7F)
		
		add		t0,t0,c0
		add		t1,t1,c1
		add		t0,t0,d0
		add		t1,t1,d1

		add		r4,r4,r5		// pbSrc += iStride 
		subi	dy,dy,1			// dy--
		
		stw		t0,0(r3)		// *(pbDst+0) = 
		stw		t1,4(r3)		// *(pbDst+4) =
		add		r3,r3,r5		// pbDst += iStride
		
		cmpwi	dy,0			// Loop 8 times
		bne		remX
		
xit:							// Return
	}
}



//______________________________________________________
//
//	 16x16 Horizontal Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][X]		(A+B+1)>>1
//	[ ][ ]

void vmp_MCompHz16RPPC( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{//                                   r3                   r4          r5
	unsigned int register e01, e7F, dy, c0, c1, c2, c3, c4, d0, d1, d2, d3, t0, t1;
	unsigned int register lshift, lmask;
	
#ifndef USE_ASSERT
#pragma unused(pbDst, pbSrc, iStride)
#endif

	ASSERT_PTR4(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

__asm {	
			//	test and branch if pbSrc not aligned!

		andi.	r0,r4,3			// Branch if NOT four byte aligned?
		cmpwi	r0,0
		bne		badalign

			/********************\
			|* pbSrc IS aligned *|
			\********************/

		lis		e01,257			// =01010000
		lis		e7F,32639		// =7F7F0000

		li		dy,16			// dy=16 # of scanlines
		
		addi	e01,e01,257		//  =01010101  Error Correction 01
		addi	e7F,e7F,32639	//  =7F7F7F7F  Error Correction 7F
		
rem0:	lwz		c0,0(r4)		// 0[0123] = *(pbSrc+0)
		lwz		c1,4(r4)		//  [4567] = *(pbSrc+4)
		lwz		c2,8(r4)		//  [89AB] = *(pbSrc+8)
		lwz		c3,12(r4)		//  [CDEF] = *(pbSrc+12)
		lwz		c4,16(r4)		// 1[0123] = *(pbSrc+16)

		slwi	d0,c0,8			// 0[123_]
		slwi	d1,c1,8			// 4[567_]
		rlwimi	d0,c1,8,24,31	//  [1234]
		rlwimi	d1,c2,8,24,31	//  [5678]

		slwi	d2,c2,8			//  [9AB_]
		srwi	c4,c4,24		// 1[___0]
		rlwimi	d2,c3,8,24,31	//  [9ABC]
		rlwimi	c4,c3,8,0,23	//  [DEF 10]
						
//		*pDst =     (( c0 >> 1 ) & eMASK7F) + ((d0 >> 1) & eMASK7F) + ((d0 | c0) & eMASK01);
//		*(pDst+1) = (( c1 >> 1 ) & eMASK7F) + ((d1 >> 1) & eMASK7F) + ((d1 | c1) & eMASK01);

		or		t0,c0,d0		// (c0 | d0)  [0123] | [1234]
		or		t1,c1,d1		// (c1 | d1)  [4567] | [5678]
		srwi	c0,c0,1			// (c0 >> 1)
		srwi	d0,d0,1			// (d0 >> 1)
		srwi	c1,c1,1			// (c1 >> 1)
		srwi	d1,d1,1			// (d1 >> 1)
		
		and		t0,t0,e01		// ((c0 | d0) & e01)
		and		t1,t1,e01		// ((c1 | d1) & e01)
		and		c0,c0,e7F		// ((c0 >> 1) & e7F)
		and		d0,d0,e7F		// ((d0 >> 1) & e7F)
		and		c1,c1,e7F		// ((c1 >> 1) & e7F)
		and		d1,d1,e7F		// ((d1 >> 1) & e7F)
		
		add		c0,c0,t0
		add		c1,c1,t1
		add		c0,c0,d0
		add		c1,c1,d1

//		*(pDst+0) = (( c2 >> 1 ) & eMASK7F) + ((d2 >> 1) & eMASK7F) + ((d2 | c2) & eMASK01);
//		*(pDst+1) = (( c3 >> 1 ) & eMASK7F) + ((d3 >> 1) & eMASK7F) + ((d3 | c3) & eMASK01);

		or		t0,c2,d2		// (c2 | d2)  [89AB] | [9ABC]
		or		t1,c3,c4		// (c3 | d3)  [CDEF] | [DEF 10]
		srwi	c2,c2,1			// (c2 >> 1)
		srwi	d2,d2,1			// (d2 >> 1)
		srwi	c3,c3,1			// (c3 >> 1)
		srwi	c4,c4,1			// (d3 >> 1)
		
		and		t0,t0,e01		// ((c2 | d2) & e01)
		and		t1,t1,e01		// ((c3 | d3) & e01)
		and		c2,c2,e7F		// ((c2 >> 1) & e7F)
		and		d2,d2,e7F		// ((d2 >> 1) & e7F)
		and		c3,c3,e7F		// ((c3 >> 1) & e7F)
		and		c4,c4,e7F		// ((d3 >> 1) & e7F)
		
		add		c2,c2,t0
		add		c3,c3,t1
		add		c2,c2,d2
		add		c3,c3,c4

		add		r4,r4,r5		// pbSrc += iStride 
		subi	dy,dy,1			// dy--
		
		stw		c0,0(r3)		// *(pbDst+0) = 
		stw		c1,4(r3)		// *(pbDst+4) =
		stw		c2,8(r3)		// *(pbDst+8) =
		stw		c3,12(r3)		// *(pbDst+12) =
		add		r3,r3,r5		// pbDst += iStride
		
		cmpwi	dy,0			// Loop 16 times
		bne		rem0

		b		xit				// Done!!!

			/*********************\
			|* pbSrc NOT aligned *|
			\*********************/
badalign:
		li		dy,-4				// =FFFFFFFC
		slwi	lshift,r0,3			// x8  1=8, 2=16, 3=24 bits to shift
		and		r4,r4,dy			// Set src to mod 4 = 0

			//	Generate shifting masks
		
		li		t0,32
		li		c3,-1				//  =FFFFFFFF
		sub		t0,t0,lshift		// 8=24 16=16 n=32-n
		srw		lmask,c3,t0			// 8=000000FF 16=0000FFFF 24=00FFFFFF

		li		dy,16				// dy=16
				
			//	First we need to blend the bits
			//
			//	lshift=(left shift) lmask=(left mask)
			//	e01,e7F,  t0,r0 scratch
remX:
		li		t0,-1				// lmask=FFFFFFFF
		
		lwz		c0,0(r4)			// c0= *(pbSrc+0)
		lwz		c1,4(r4)			// c1= *(pbSrc+4)
		lwz		c2,8(r4)			// c2= *(pbSrc+8)
		lwz		c3,12(r4)			// c3= *(pbSrc+12)
		lwz		d3,16(r4)			// C4= *(pbSrc+16)

		xor		t0,t0,lmask			// (right mask)
		
		slw		c0,c0,lshift		// 8=xxxxxx00 16=xxxx0000 24=xx000000  <<c0
		rlwnm	c1,c1,lshift,0,31	// 8=22110033 16=11003322 24=00332211  <<c1<<
		rlwnm	c2,c2,lshift,0,31	// 8=22110033 16=11003322 24=00332211  <<c2<<
		rlwnm	c3,c3,lshift,0,31	// 8=22110033 16=11003322 24=00332211  <<c3<<		
		rlwnm	d3,d3,lshift,0,31	// 8=22110033 16=11003322 24=00332211  <<c4<<

		and		r0,c1,lmask			// 8=000000FF 16=0000FFFF 24=00FFFFFF
		and		c1,c1,t0			// 8=FFFFFF00 16=FFFF0000 24=FF000000
		or		c0,c0,r0			//  [0123]

		and		r0,c2,lmask
		and		c2,c2,t0
		or		c1,c1,r0			//  [4567]
		
		and		r0,c3,lmask
		and		c3,c3,t0
		or		c2,c2,r0			//  [89AB]

		and		r0,d3,lmask
		or		c3,c3,r0			//  [CDEF]

		slwi	d0,c0,8				//  [123_]
		rlwimi	d0,c1,8,24,31		//  [1234]

		slwi	d1,c1,8				//  [567_]
		rlwimi	d1,c2,8,24,31		//  [5678]

		slwi	d2,c2,8				//  [9AB_]
		rlwimi	d2,c3,8,24,31		//  [9ABC]
		
		srwi	d3,d3,24			//  [___ 10]
		rlwimi	d3,c3,8,0,23		//  [DEF 10]

//		*pDst =     (( c0 >> 1 ) & eMASK7F) + ((d0 >> 1) & eMASK7F) + ((d0 | c0) & eMASK01);
//		*(pDst+1) = (( c1 >> 1 ) & eMASK7F) + ((d1 >> 1) & eMASK7F) + ((d1 | c1) & eMASK01);

		lis		e01,257			// =01010000
		lis		e7F,32639		// =7F7F0000
		
		or		t0,c0,d0		// (c0 | d0)  [0123] | [1234]
		or		r0,c1,d1		// (c1 | d1)  [4567] | [5678]

		addi	e01,e01,257		// =01010101  Error Correction 01
		addi	e7F,e7F,32639	// =7F7F7F7F  Error Correction 7F
		
		srwi	c0,c0,1			// (c0 >> 1)
		srwi	d0,d0,1			// (d0 >> 1)
		srwi	c1,c1,1			// (c1 >> 1)
		srwi	d1,d1,1			// (d1 >> 1)
		
		and		t0,t0,e01		// ((c0 | d0) & e01)
		and		r0,r0,e01		// ((c1 | d1) & e01)
		and		c0,c0,e7F		// ((c0 >> 1) & e7F)
		and		d0,d0,e7F		// ((d0 >> 1) & e7F)
		and		c1,c1,e7F		// ((c1 >> 1) & e7F)
		and		d1,d1,e7F		// ((d1 >> 1) & e7F)
		
		add		c0,c0,t0
		add		c1,c1,r0
		add		c0,c0,d0
		add		c1,c1,d1

//		*(pDst+0) = (( c2 >> 1 ) & eMASK7F) + ((d2 >> 1) & eMASK7F) + ((d2 | c2) & eMASK01);
//		*(pDst+1) = (( c3 >> 1 ) & eMASK7F) + ((d3 >> 1) & eMASK7F) + ((d3 | c3) & eMASK01);
		
		or		t0,c2,d2		// (c2 | d2)  [89AB] | [9ABC]
		or		r0,c3,d3		// (c3 | d3)  [CDEF] | [DEF 10]
		
		srwi	c2,c2,1			// (c2 >> 1)
		srwi	d2,d2,1			// (d2 >> 1)
		srwi	c3,c3,1			// (c3 >> 1)
		srwi	d3,d3,1			// (d3 >> 1)
		
		and		t0,t0,e01		// ((c2 | d2) & e01)
		and		r0,r0,e01		// ((c3 | d3) & e01)
		and		c2,c2,e7F		// ((c2 >> 1) & e7F)
		and		d2,d2,e7F		// ((d2 >> 1) & e7F)
		and		c3,c3,e7F		// ((c3 >> 1) & e7F)
		and		d3,d3,e7F		// ((d3 >> 1) & e7F)
		
		add		c2,c2,t0
		add		c3,c3,r0
		add		c2,c2,d2
		add		c3,c3,d3

		add		r4,r4,r5		// pS += Stride 
		subi	dy,dy,1			// dy--
		
		stw		c0,0(r3)		// *(pDst+0) = 
		stw		c1,4(r3)		// *(pDst+4) =
		stw		c2,8(r3)		// *(pDst+8) =
		stw		c3,12(r3)		// *(pDst+12) =
		add		r3,r3,r5		// pbDst += iStride
		
		cmpwi	dy,0			// Loop 16 times
		bne		remX
		
xit:							// Done!
	}
}




//______________________________________________________
//
//	 8x8 Vertical Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][ ]		(A+B+1)>>1
//	[X][ ]

void vmp_MCompVt8RPPC( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{//                              r3                   r4          r5
	unsigned int register e01, e7F, dy, pR, c0, c1, c2, d0, d1, d2, t0, t1;
	unsigned int register lshift, lmask, rshift, rmask;

#ifndef USE_ASSERT
#pragma unused(pbDst, pbSrc, iStride)
#endif

	ASSERT_PTR4(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

__asm {	
			//	test and branch if pbSrc not aligned!

		andi.	r0,r4,3			// Branch if NOT four byte aligned?
		cmpwi	r0,0
		bne		badalign

			/********************\
			|* pbSrc IS aligned *|
			\********************/

		lis		e01,257			// =01010000
		lis		e7F,32639		// =7F7F0000

		add		pR,r4,r5		// pR = pbSrc + iStride
		li		dy,8			// dy=8
		
		addi	e01,e01,257		// =01010101  Error Correction 01
		addi	e7F,e7F,32639	// =7F7F7F7F  Error Correction 7F
		
rem0:	lwz		c0,0(r4)		//  [0123] = *(pS+0)
		lwz		c1,4(r4)		//  [4567] = *(pS+4)
		lwz		d0,0(pR)		//  [0123] = *(pR+0)
		lwz		d1,4(pR)		//  [4567] = *(pR+4)
		
//		*pDst = (( *pS >> 1 ) & e7F) + ((*pR >> 1) & e7F) + ((*pS | *pR) & e01);
//		*(pDst+1) = (( *(pS+1) >> 1 ) & eMASK7F) + ((*(pR+1) >> 1) & eMASK7F) + ((*(pS+1) | *(pR+1)) & eMASK01);

		or		t0,c0,d0		// (c0 | d0)  [0123] | [1234]
		or		t1,c1,d1		// (c1 | d1)  [4567] | [5678]
		srwi	c0,c0,1			// (c0 >> 1)
		srwi	d0,d0,1			// (d0 >> 1)
		srwi	c1,c1,1			// (c1 >> 1)
		srwi	d1,d1,1			// (d1 >> 1)
		
		and		t0,t0,e01		// ((c0 | d0) & e01)
		and		t1,t1,e01		// ((c1 | d1) & e01)
		and		c0,c0,e7F		// ((c0 >> 1) & e7F)
		and		d0,d0,e7F		// ((d0 >> 1) & e7F)
		and		c1,c1,e7F		// ((c1 >> 1) & e7F)
		and		d1,d1,e7F		// ((d1 >> 1) & e7F)
		
		add		t0,t0,c0
		add		t1,t1,c1
		add		t0,t0,d0
		add		t1,t1,d1

		add		r4,r4,r5		// pbSrc += iStride 
		add		pR,pR,r5		// pR    += iStride
		subi	dy,dy,1			// dy--
		
		stw		t0,0(r3)		// *(pbDst+0) = 
		stw		t1,4(r3)		// *(pbDst+4) =
		add		r3,r3,r5		// pbDst += iStride
		
		cmpwi	dy,0			// Loop 8 times
		bne		rem0

		b		xit				// Done!!!
		
		
			/*********************\
			|* pbSrc NOT aligned *|
			\*********************/

badalign:  	// r0={1...3} offset
		li		pR,-4				// =FFFFFFFC
		slwi	lshift,r0,3			// x8  1=8, 2=16, 3=24 bits to shift
		and		r4,r4,pR			// Set src to mod 4 = 0

			//	Generate shifting masks
		
		li		rshift,32
		li		rmask,-1				//  =FFFFFFFF
		sub		rshift,rshift,lshift	// 8=24 16=16 n=32-n
		srw		lmask,rmask,rshift		// 8=000000FF 16=0000FFFF 24=00FFFFFF
		xor		rmask,rmask,lmask		// 8=FFFFFF00 16=FFFF0000 24=FF000000

			//	lshift=(left shift) rshift=(right shift) lmask=(left mask) rmask=(right mask)
			// Standard preparation (just like mod 4=0) {properly aligned src}

		lis		e01,257				// =01010000
		lis		e7F,32639			// =7F7F0000

		add		pR,r4,r5			// pR = pS + iStride
		li		dy,8				// dy=8
				
		addi	e01,e01,257			// =01010101  Error Correction 01
		addi	e7F,e7F,32639		// =7F7F7F7F  Error Correction 7F

			//	lshift=(left shift) rshift=(right shift) lmask=(left mask)
			//	NOTE: r4 is either pbSrc -1,-2,-3  actual data!

									//   1=8   2=16   3=24   (offset to left shift to correct!)
remX:	lwz		c0,0(r4)	 		// [X012] [XX01] [XXX0] = *(pS+0) 
		lwz		c1,4(r4)	 		// [X456] [XX45] [XXX4] = *(pS+4)
		lwz		c2,8(r4)	 		// [X89A] [XX89] [XXX8] = *(pS+8)
		lwz		d0,0(pR)	 		// [X012] [XX01] [XXX0] = *(pR+0) 
		lwz		d1,4(pR)	 		// [X456] [XX45] [XXX4] = *(pR+4)
		lwz		d2,8(pR)	 		// [X89A] [XX89] [XXX8] = *(pR+8)
		
		slw		c0,c0,lshift		//0[012_] [01__] [0___]
		slw		d0,d0,lshift		//1[012_] [01__] [0___]
		
		rlwnm	c1,c1,lshift,0,31	//0[4563] [4523] [4123]
		rlwnm	d1,d1,lshift,0,31	//1[4563] [4523] [4123]

		and		r0,c1,lmask			//0[___3] [__23] [_123]
		srw		c2,c2,rshift		//0[___7] [__67] [_567]
		and		c1,c1,rmask			//0[456_] [45__] [4___]
		or		c0,c0,r0			//0[0123] [0123] [0123]
		or		c1,c1,c2			//0[4567] [4567] [4567] 

		and		r0,d1,lmask			//1[___3] [__23] [_123]
		srw		d2,d2,rshift		//1[___7] [__67] [_567]
		and		d1,d1,rmask			//1[456_] [45__] [4___]
		or		d0,d0,r0			//1[0123] [0123] [0123]
		or		d1,d1,d2			//1[4567] [4567] [4567]
		
			// *** c1:c0  and  d1:d0 Sourc Data now properly aligned! Handle as normal ***
			
//		*pDst = (( *pS >> 1 ) & e7F) + ((*pR >> 1) & e7F) + ((*pS | *pR) & e01);
//		*(pDst+1) = (( *(pS+1) >> 1 ) & eMASK7F) + ((*(pR+1) >> 1) & eMASK7F) + ((*(pS+1) | *(pR+1)) & eMASK01);

		or		t0,c0,d0		// (*pS | *pR)
		or		t1,c1,d1		// (*(pS+1) | *(pR+1))
		srwi	c0,c0,1			// (*pS >> 1)
		srwi	d0,d0,1			// (*pR >> 1)
		srwi	c1,c1,1			// (*(pS+1) >> 1)
		srwi	d1,d1,1			// (*(pR+1) >> 1)
		
		and		t0,t0,e01		// ((*pS | *pR) & e01)
		and		t1,t1,e01		// ((*(pS+1) | *(pR+1)) & e01)
		and		c0,c0,e7F		// ((*pS >> 1) & e7F)
		and		d0,d0,e7F		// ((*pR >> 1) & e7F)
		and		c1,c1,e7F		// ((*(pS+1) >> 1) & e7F)
		and		d1,d1,e7F		// ((*(pR+1) >> 1) & e7F)
		
		add		t0,t0,c0
		add		t1,t1,c1
		add		t0,t0,d0
		add		t1,t1,d1

		add		r4,r4,r5		// pS += iStride 
		add		pR,pR,r5		// pR += iStride
		subi	dy,dy,1			// dy--
		
		stw		t0,0(r3)		// *(pbDst+0) = 
		stw		t1,4(r3)		// *(pbDst+4) =
		add		r3,r3,r5		// pbDst += iStride

		cmpwi	dy,0			// Loop 8 times
		bne		remX
		
xit:							// Done!
	}
}



//______________________________________________________
//
//	 16x16 Vertical Motion Compensation w/Rounding
//
//	NOTE: Stride (+) is top down and (-) is bottom up orientation
//		Thus Src and Dst may be beginning of (+=)first or (-=)last scanline.
//
//	[X][ ]		(A+B+1)>>1
//	[X][ ]

void vmp_MCompVt16RPPC( uint8 * pbDst, const uint8 * pbSrc, int iStride )
{//                              r3                   r4          r5
	unsigned int register e01, e7F, pR, c0, c1, c2, c3, c4, d0, d1, d2, d3, d4, t0, t1, dy;
	unsigned int register lshift, lmask, iFlip;

#ifndef USE_ASSERT
#pragma unused(pbDst, pbSrc, iStride)
#endif

	ASSERT_PTR4(pbDst);
	ASSERT_PTR(pbSrc);
	ASSERT_ZERO(iStride);

__asm {	
			//	test and branch if pbSrc not aligned!

		andi.	r0,r4,3			// Branch if NOT four byte aligned?
		cmpwi	r0,0
		bne		badalign

			/********************\
			|* pbSrc IS aligned *|
			\********************/

		lis		e01,257			// =01010000
		lis		e7F,32639		// =7F7F0000

		add		pR,r4,r5		// pR = pS + iStride
		li		dy,16			// dy=16
		
		addi	e01,e01,257		// =01010101  Error Correction 01
		addi	e7F,e7F,32639	// =7F7F7F7F  Error Correction 7F

rem0:	lwz		c0,0(r4)		// 0[0123] = *(pS+0)
		lwz		d0,0(pR)		// 1[0123] = *(pR+0)
		lwz		c1,4(r4)		// 0[4567] = *(pS+4)
		lwz		d1,4(pR)		// 1[4567] = *(pR+4)
		
//		*pDst = (( *pS >> 1 ) & e7F) + ((*pR >> 1) & e7F) + ((*pS | *pR) & e01);
//		*(pDst+1) = (( *(pS+1) >> 1 ) & eMASK7F) + ((*(pR+1) >> 1) & eMASK7F) + ((*(pS+1) | *(pR+1)) & eMASK01);

		or		t0,c0,d0		// (*pS | *pR)         [0123] | [0123]
		or		t1,c1,d1		// (*(pS+1) | *(pR+1)) [4567] | [4567]
		srwi	c0,c0,1			// (*pS >> 1)
		srwi	d0,d0,1			// (*pR >> 1)
		srwi	c1,c1,1			// (*(pS+1) >> 1)
		srwi	d1,d1,1			// (*(pR+1) >> 1)
		
		and		t0,t0,e01		// ((*pS | *pR) & e01)
		and		c0,c0,e7F		// ((*pS >> 1) & e7F)
		and		d0,d0,e7F		// ((*pR >> 1) & e7F)

		and		t1,t1,e01		// ((*(pS+1) | *(pR+1)) & e01)
		and		c1,c1,e7F		// ((*(pS+1) >> 1) & e7F)
		and		d1,d1,e7F		// ((*(pR+1) >> 1) & e7F)
		
		add		t0,t0,c0
		add		t1,t1,c1
		add		t0,t0,d0
		add		t1,t1,d1

		stw		t0,0(r3)		// *(pbDst+0) = 
		stw		t1,4(r3)		// *(pbDst+4) =
		
			//	________________________________________
			//	Next 8 bytes

		lwz		c0,8(r4)		// 0[89AB] = *(pS+8)
		lwz		d0,8(pR)		// 1[89AB] = *(pR+8)
		lwz		c1,12(r4)		// 0[CDEF] = *(pS+12)
		lwz		d1,12(pR)		// 1[CDEF] = *(pR+12)
		
//		*(pDst+2) = (( *(pS+8) >> 1 ) & e7F) + ((*(pR+8) >> 1) & e7F) + ((*(pS+8) | *(pR+8)) & e01);

		or		t0,c0,d0		// (*(pS+8) | *(pR+8))
		srwi	c0,c0,1			// (*(pS+8) >> 1)
		srwi	d0,d0,1			// (*(pR+8) >> 1)

//		*(pDst+3) = (( *(pS+12) >> 1 ) & eMASK7F) + ((*(pR+12) >> 1) & eMASK7F) + ((*(pS+12) | *(pR+12)) & eMASK01);

		or		t1,c1,d1		// (*(pS+12) | *(pR+12))
		srwi	c1,c1,1			// (*(pS+12) >> 1)
		srwi	d1,d1,1			// (*(pR+12) >> 1)
		
		and		t0,t0,e01		// ((*(pS+8) | *(pR+8)) & e01)
		and		c0,c0,e7F		// ((*(pS+8) >> 1) & e7F)
		and		d0,d0,e7F		// ((*(pR+8) >> 1) & e7F)

		and		t1,t1,e01		// ((*(pS+12) | *(pR+12)) & e01)
		and		c1,c1,e7F		// ((*(pS+12) >> 1) & e7F)
		and		d1,d1,e7F		// ((*(pR+12) >> 1) & e7F)
		
		add		t0,t0,c0
		add		t1,t1,c1
		add		t0,t0,d0
		add		t1,t1,d1

		add		r4,r4,r5		// pS += iStride 
		add		pR,pR,r5		// pR += iStride

		subi	dy,dy,1			// dy--
		
		stw		t0,8(r3)		// *(pbDst+8) = 
		stw		t1,12(r3)		// *(pbDst+12) =
		add		r3,r3,r5		// pbDst += iStride			

		cmpwi	dy,0			// Loop 16 times
		bne		rem0

		b		xit				// Done!!!
		
			/*********************\
			|* pbSrc NOT aligned *|
			\*********************/

badalign:  			
		li		d4,-4				// =FFFFFFFC
		slwi	lshift,r0,3			// x8  1=8, 2=16, 3=24 bits to shift
		and		r4,r4,d4			// Set src to mod 4 = 0

			//	Generate shifting masks
		
		li		c3,32
		li		d3,-1				//  =FFFFFFFF
		sub		c3,c3,lshift		// 8=24 16=16 n=32-n
		srw		lmask,d3,c3			//  [___X] [__XX] [_XXX] lmask

			//	lshift=(left shift)  lmask=(left mask) 
			// Standard preparation (just like mod 4=0) {properly aligned src}

		lis		e01,257				// =01010000
		lis		e7F,32639			// =7F7F0000

		li		dy,16				// dy=16
		li		iFlip,-1			// FFFFFFFF
		
		addi	e01,e01,257			// =01010101  Error Correction 01
		addi	e7F,e7F,32639		// =7F7F7F7F  Error Correction 7F		
		xor		iFlip,iFlip,lmask	//  [XXX_] [XX__] [X___] rmask

			//	lshift=(left shift) rshift=(right shift) lmask=(left mask)
			//	NOTE: r4 is either pbSrc -1,-2,-3  actual data!

									//    1=8   2=16   3=24   (offset to left shift to correct!)
remX:	lwz		c0,0(r4)			//  [X012] [XX01] [XXX0] = *(pS+0)
		lwz		c1,4(r4)			//  [3456] [2345] [1234] = *(pS+4)
		lwz		c2,8(r4)			//  [789A] [6789] [5678] = *(pS+8)
		lwz		c3,12(r4)			//  [BCDE] [ABCD] [9ABC] = *(pS+12)
		lwz		c4,16(r4)			//  [F012] [EF01] [DEF0] = *(pS+16)
		
		add		r4,r4,r5			// pS = pS + Stride
		
		slw		c0,c0,lshift		//  [012_] [01__] [0___]
		rlwnm	c1,c1,lshift,0,31	//  [4563] [4523] [4123]
		rlwnm	c2,c2,lshift,0,31	//  [89A7] [8967] [8567]
		rlwnm	c3,c3,lshift,0,31	//  [CDEB] [CDAB] [C9AB]
														
		lwz		d0,0(r4)			//  [X012] [XX01] [XXX0] = *(pR+0)
		lwz		d1,4(r4)			//  [3456] [2345] [1234] = *(pR+4)
		lwz		d2,8(r4)			//  [789A] [6789] [5678] = *(pR+8)
		lwz		d3,12(r4)			//  [BCDE] [ABCD] [9ABC] = *(pR+12)
		lwz		d4,16(r4)			//  [F012] [EF01] [DEF0] = *(pR+16)

		slw		d0,d0,lshift		//  [012_] [01__] [0___]
		rlwnm	d1,d1,lshift,0,31	//  [4563] [4523] [4123]
		rlwnm	d2,d2,lshift,0,31	//  [89A7] [8967] [8567]
		rlwnm	d3,d3,lshift,0,31	//  [CDEB] [CDAB] [C9AB]

									//  [___X] [__XX] [_XXX] lmask
		rlwnm	c4,c4,lshift,0,31	//  [012F] [01EF] [0DEF]
		rlwnm	d4,d4,lshift,0,31	//  [012F] [01EF] [0DEF]

		and		c4,c4,lmask			//  [___F] [__EF] [_DEF]
		and		d4,d4,lmask			//  [___F] [__EF] [_DEF]
															
		and		t0,c1,lmask			//  [___3] [__23] [_123]
		and		c1,c1,iFlip			//  [456_] [45__] [4___]
		or		c0,c0,t0			//  [0123] [0123] [0123]
		
		and		t0,c2,lmask			//  [___7] [__67] [_567]
		and		c2,c2,iFlip			//  [89A_] [89__] [8___]
		or		c1,c1,t0			//  [4567] [4567] [4567]

		and		t0,c3,lmask			//  [___B] [__AB] [_9AB]
		and		c3,c3,iFlip			//  [CDE_] [CD__] [C___]
		or		c2,c2,t0			//  [89AB] [89AB] [89AB]
		or		c3,c3,c4			//  [CDEF] [CDEF] [CDEF]
		
		and		t0,d1,lmask			//  [___3] [__23] [_123]
		and		d1,d1,iFlip			//  [456_] [45__] [4___]
		or		d0,d0,t0			//  [0123] [0123] [0123]
		
		and		t0,d2,lmask			//  [___7] [__67] [_567]
		and		d2,d2,iFlip			//  [89A_] [89__] [8___]
		or		d1,d1,t0			//  [4567] [4567] [4567]

		and		t0,d3,lmask			//  [___B] [__AB] [_9AB]
		and		d3,d3,iFlip			//  [CDE_] [CD__] [C___]
		or		d2,d2,t0			//  [89AB] [89AB] [89AB]
		or		d3,d3,d4			//  [CDEF] [CDEF] [CDEF]

			//		c0/d0=[0123] c1/d1=[4567] c2/d2=[89AB] c3/d3=[CDEF]
			// _______________________________________________________
			
//		*pDst = (( *pS >> 1 ) & e7F) + ((*pR >> 1) & e7F) + ((*pS | *pR) & e01);

			//	Okay now we calculate or second error correction vector
		
		or		c4,c0,d0		// (*pS | *pR)
		srwi	c0,c0,1			// (*pS >> 1)
		srwi	d0,d0,1			// (*pR >> 1)

//		*(pDst+1) = (( *(pS+1) >> 1 ) & eMASK7F) + ((*(pR+1) >> 1) & eMASK7F) + ((*(pS+1) | *(pR+1)) & eMASK01);
		
		or		d4,c1,d1		// (*(pS+1) | *(pR+1))
		srwi	c1,c1,1			// (*(pS+1) >> 1)
		srwi	d1,d1,1			// (*(pR+1) >> 1)
		
		and		c4,c4,e01		// ((*pS | *pR) & e01)
		and		c0,c0,e7F		// ((*pS >> 1) & e7F)
		and		d0,d0,e7F		// ((*pR >> 1) & e7F)

		and		d4,d4,e01		// ((*(pS+1) | *(pR+1)) & e01)
		and		c1,c1,e7F		// ((*(pS+1) >> 1) & e7F)
		and		d1,d1,e7F		// ((*(pR+1) >> 1) & e7F)
		
		add		c0,c0,c4
		add		c1,c1,d4
		add		c0,c0,d0
		add		c1,c1,d1

//		*(pDst+1) = (( *(pS+2) >> 1 ) & eMASK7F) + ((*(pR+2) >> 1) & eMASK7F) + ((*(pS+2) | *(pR+2)) & eMASK01);
//		*(pDst+1) = (( *(pS+3) >> 1 ) & eMASK7F) + ((*(pR+3) >> 1) & eMASK7F) + ((*(pS+3) | *(pR+3)) & eMASK01);

		or		c4,c2,d2		// (*(pS+2) | *(pR+2))
		or		d4,c3,d3		// (*(pS+3) | *(pR+3))
		srwi	c2,c2,1			// (*(pS+2) >> 1)
		srwi	d2,d2,1			// (*(pR+2) >> 1)
		srwi	c3,c3,1			// (*(pS+3) >> 1)
		srwi	d3,d3,1			// (*(pR+3) >> 1)

		and		c4,c4,e01		// ((*(pS+2) | *(pR+2)) & e01)
		and		c2,c2,e7F		// ((*(pS+2) >> 1) & e7F)
		and		d2,d2,e7F		// ((*(pR+2) >> 1) & e7F)

		and		d4,d4,e01		// ((*(pS+3) | *(pR+3)) & e01)
		and		c3,c3,e7F		// ((*(pS+3) >> 1) & e7F)
		and		d3,d3,e7F		// ((*(pR+3) >> 1) & e7F)

		add		c2,c2,c4
		add		c3,c3,d4
		add		c2,c2,d2
		add		c3,c3,d3
		
		subi	dy,dy,1			// dy--
		
		stw		c0,0(r3)		// *(pbDst+0) = 
		stw		c1,4(r3)		// *(pbDst+4) =
		stw		c2,8(r3)		// *(pbDst+8) =
		stw		c3,12(r3)		// *(pbDst+12) =
		add		r3,r3,r5		// pbDst += iStride

		cmpwi	dy,0			// Loop 16 times
		bne		remX
		
xit:							// Done!
	}		
}

