
//
//  yuv12PPC.cpp - YUV12 to RGB
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 21 Feb, 2002 - James Leiterman
//
//  Last Modification
//
//	NOTE: 8x8 and 16x16 SOURCE Prediction Blocks are not always aligned!!!
//	If they were, the code would be MUCH faster!
//
//	Borrowed from optimized code I wrote for ProjectMayo Open Source Project.
//	www.projectmayo.com  DivX;-) Open Source Code! yuv2rgb_ppc.c
//	I've modified it a bit to match my use of it in the book.
//
//  stab=4
//



#include "KariType.h"                                   // Master Definitions

	//	3rd Party Libraries
	
#include	<assert.h>
#include	<iostream.h>

    //  Book Development Library Modules

#include "CpuAsm.h"    
#include "vmp3D.h"
#include "vmp3DEx.h"
#include "yuv12.h"
#include "yuv12Ex.h"


#define USE_PPC_MULTBL


#ifdef USE_PPC_MULTBL
	//	Mutiplication Lookup table
static int UTblMul[256];
static int VTblMul[256];
static int YTblMul[256];
#endif


#define VAL_Ru			0x3343		// Ru=13123v
#define VAL_Gv			0x0c92		// Gv= 3218u
#define VAL_RGBy		0x2568		// Y = 9576y
#define VAL_Gu			0x1a1e		// Gu= 6686u
#define VAL_Bv			0x40cf		// Bv=16591v



//______________________________________________________
//	YUV12 to 32bit RGB Conversion
//
//	Borrowed from optimized code I wrote for ProjectMayo
//	Open Source Project.  DivX;-) Open Source Code! yuv2rgb_ppc.c
//	I've modified it a bit to match my use of it in the book.

void vmp_yuv12toRGB32PPC(
	void *pImg,				// r3
	uint nWidth,			// r4
	uint nHeight,			// r5
	int iStride,			// r6
	const uint8 * pBaseY,	// r7
	uint nStrideY, 			// r8
	const uint8 * pBaseU,	// r9
	const uint8 * pBaseV,	// r10
	                        // *too many passed stack args*
	uint nStrideUV )		// r12 ?
{
	unsigned int register iDir, iSAdj, iBit, w, suv;
	unsigned int register y0, u0, w0, rd0, gn0, bl0;
	unsigned int register y1, u1, w1, rd1, gn1, bl1;
		
#ifndef USE_ASSERT
#pragma unused(pBaseY, pBaseU, pBaseV, pImg, nStrideY, nStrideUV, nWidth, nHeight, iStride)
#endif

#ifdef USE_PPC_MULTBL
	int register * pYTbl;
	int register * pUTbl;
	int register * pVTbl;

	pUTbl = UTblMul;
	pVTbl = VTblMul;
	pYTbl = YTblMul;
#endif

	ASSERT_PTR16(pBaseY);
	ASSERT_PTR16(pBaseU);
	ASSERT_PTR16(pBaseV);
	ASSERT_PTR16(pImg);
	ASSERT_ZERO(nStrideY);
	ASSERT_ZERO(nStrideUV);
	ASSERT_ZERO(nWidth);
	ASSERT_ZERO(nHeight);
	ASSERT_ZERO(iStride);
	
		//	Low on registers so generate a stepping (odd/even) key for UV
	suv = -(nWidth >> 1);						// # of UV's needed
	iDir = suv ^ (nStrideUV + suv);			// Stepping Key
	iSAdj = iStride - (nWidth<<2);

	// Compiler not assigning r6 to iStride (too many arguments)!!!
__asm {
		sub		r8,r8,r4		// nStrideY -= nWidth
		li		iBit,0xFFFF		// = 0xFFFFFFFF

high0:	mr		w,r4			// w=nWidth

		// Note: y0 & y1 share  u0 & w0, but we process pixel pairs!
wide0:	lbz		w0,0(r10) 		// v=pBaseV[x>>1]-128
		lbz		u0,0(r9) 		// u=pBaseU[x>>1]-128
		lbz		y0,0(r7) 		// y=pBaseY[x] -16
		lbz		y1,1(r7) 		// y=pBaseY[x+1]-16

	// YUV to RGB Math
	//	R=(9576y + 13123u) / 8192           _R(y,u,v) (0x2568*(y)  	       + 0x3343*(u)) /0x2000
	//	G=(9576y - 3218v - 6686u) / 8192    _G(y,u,v) (0x2568*(y) - 0x0c92*(v) - 0x1a1e*(u)) /0x2000
	//	B=(9576y + 16591v) / 8192           _B(y,u,v) (0x2568*(y) + 0x40cf*(v)) / 0x2000		

#ifdef USE_PPC_MULTBL
			//	Use Multiplication YUV Lookup tables
		slwi	w0,w0,2			// x4 (int table index)
		slwi	u0,u0,2
		slwi	y0,y0,2
		slwi	y1,y1,2

		lwzx	gn0,pVTbl,w0	// lookup int
		lwzx	rd0,pUTbl,u0
		lwzx	y0,pYTbl,y0
		lwzx	y1,pYTbl,y1

		slwi	w0,gn0,2		//x4	12872v=(3218v*4)
		srawi	u0,rd0,1		// 13123u / 2        {2.0}
		srawi	rd1,rd0,7		// 13123u / 128      {1.969237695}

		add		u0,u0,rd1

		add		bl0,w0,gn0		//x5	16090v=(3218v*5)
		sub		gn1,y1,gn0
		sub		gn0,y0,gn0
		
#else
		subi	w0,w0,128
		subi	u0,u0,128
		subi	y0,y0,16
		subi	y1,y1,16

		mulli	gn0,w0,VAL_Gv	// (Gv)   3218v
		mulli	y1,y1,VAL_RGBy
		mulli	y0,y0,VAL_RGBy	// (RGBy) 9576y
		sub		gn1,y1,gn0
		sub		gn0,y0,gn0

		mulli	rd0,u0,VAL_Ru	// (Ru)   13123u
		mulli	bl0,w0,VAL_Bv	// (Bv)   16591v      {5.155686762}
		mulli	u0,u0,VAL_Gu	// (Gu)   6686u       {1.962758002}
#endif

		add		rd1,rd0,y1
		add		rd0,rd0,y0
		sub		gn1,gn1,u0
		sub		gn0,gn0,u0
		add		bl1,bl0,y1
		add		bl0,bl0,y0

		srawi	rd1,rd1,13
		srawi	rd0,rd0,13		// r/=8192
		srawi	gn1,gn1,13
		srawi	gn0,gn0,13		// g/=8192
		srawi	bl1,bl1,13
		srawi	bl0,bl0,13		// b/=8192

			// Pack signed 32bit into unsigned 8bit
			// Saturation MAX( n, 0 )
		srawi	y1,rd1,31
		srawi	y0,rd0,31		// (+)=00000000  (-)=FFFFFFFF  (OR) {Shift sign bit into all bits!}
		srawi	u1,gn1,31
		srawi	u0,gn0,31		// (+)=00000000  (-)=FFFFFFFF  (OR) {Shift sign bit into all bits!}
		srawi	w1,bl1,31
		srawi	w0,bl0,31		// (+)=00000000  (-)=FFFFFFFF  (OR) {Shift sign bit into all bits!}

		xor		y1,y1,iBit
		xor		y0,y0,iBit		// (+)=FFFFFFFF  (-)=00000000 AND MASK   {Neg #'s become 0}
		xor		u1,u1,iBit
		xor		u0,u0,iBit		// (+)=FFFFFFFF  (-)=00000000 AND MASK   {Neg #'s become 0}
		xor		w1,w1,iBit
		xor		w0,w0,iBit		// (+)=FFFFFFFF  (-)=00000000 AND MASK   {Neg #'s become 0}

		and		rd1,rd1,y1
		and		rd0,rd0,y0		// (-)'s become 0, else 1...7FFF
		and		gn1,gn1,u1
		and		gn0,gn0,u0		// (-)'s become 0, else 1...7FFF
		and		bl1,bl1,w1
		and		bl0,bl0,w0		// (-)'s become 0, else 1...7FFF

			// Saturation MIN( n, 255)
		addi	y1,rd1,0xFF00
		addi	y0,rd0,0xFF00	// -=256 is += 0xFFFFFF00		
		addi	u1,gn1,0xFF00
		addi	u0,gn0,0xFF00	// -=256 is += 0xFFFFFF00		
		addi	w1,bl1,0xFF00
		addi	w0,bl0,0xFF00	// -=256 is += 0xFFFFFF00		

		srawi	y1,y1,16
		srawi	y0,y0,16		// (0...FF)=FFFFFFFF else 00000000
		srawi	u1,u1,16
		srawi	u0,u0,16		// (0...FF)=FFFFFFFF else 00000000
		srawi	w1,w1,16
		srawi	w0,w0,16		// (0...FF)=FFFFFFFF else 00000000

		xor		y1,y1,iBit
		xor		y0,y0,iBit		// OR MASK
		xor		u1,u1,iBit
		xor		u0,u0,iBit		// OR MASK
		xor		w1,w1,iBit
		xor		w0,w0,iBit		// OR MASK

		subi	w,w,2			// w--

		or		rd1,rd1,y1
		or		rd0,rd0,y0		// Use OR Mask (>255) mask=FF, otherwise mask=00
		or		gn1,gn1,u1
		or		gn0,gn0,u0		// Use OR Mask (>255) mask=FF, otherwise mask=00
		or		bl1,bl1,w1
		or		bl0,bl0,w0		// Use OR Mask (>255) mask=FF, otherwise mask=00

		// Output RGB  [0]=0 [1]=b [2]=g [3]=r  (Big Endian)

		cmpwi	w,0
	
		rlwimi	rd1,gn1,8,16,23
		rlwimi	rd0,gn0,8,16,23	// = [....ggrr]		

		addi	r7,r7,2   		// Y-tbl++
		addi	r9,r9,1			// pBaseU++

		rlwimi	rd1,bl1,16,8,15
		rlwimi	rd0,bl0,16,8,15	// = [..bbggrr]		

		addi	r10,r10,1		// pBaseV++

		stw		rd0,0(r3)		// Write even 0RGB pixel
		stw		rd1,4(r3)		// Write odd  0RGB pixel
		addi	r3,r3,8

		bne		wide0			// w==0?  *** Loop for all pixels in scanline ***

			// ----------------------
			// We're going to play flip/flop games with the UV stride

		subi	r5,r5,1			// h--
		add		r7,r7,r8		// pBaseY += (delta) nStrideY     Next scanline!

		cmpwi	r5,0

				// Only advance on completion of an (odd) scanline!

		add		r9,r9,suv		// U Table same/next scan pair
		add		r10,r10,suv		// V Table same/next scan pair
		xor		suv,suv,iDir	// Forward/Reverse

		add		r3,r3,iSAdj		// pImg += iStride (adjustment)
		bne		high0			// r5==0?  *** Loop for all scanlines in image ***
	}
}



//
//	Initialize yuv2rgb Function Pointers
//

void vmp_yuv12PPCInit( void )
{
#ifdef USE_PPC_MULTBL
	int i, uv, y;

		//	Build Multiplication Lookup table
	for ( i = 0; i < 256; i++ )
	  {
		uv = i - 128;
		y  = i - 16;

		VTblMul[i] = VAL_Gv   * uv;
		UTblMul[i] = VAL_Ru   * uv;
		YTblMul[i] = VAL_RGBy * y;
	  }
#endif
}



