
//
//  KariType.h - Master Definitions
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//  Created - 2 Jun 98 - James Leiterman
//
//  Last Modification
//
//  stab=4
//



#ifndef __KariType_H
#define __KariType_H

#ifdef __cplusplus
extern "C" {
#endif


#define CC_VMP_MAC										// MAC



#ifndef byte
typedef unsigned char  byte;
#endif

typedef unsigned char  uint8;
typedef unsigned short uint16;
typedef unsigned long  uint32;

#ifndef uint
typedef unsigned int   uint;
#endif

typedef signed char    int8;
typedef signed short   int16;
typedef signed long    int32;


union vmp_qword {
	byte	b[8];
	uint16	h[4];
	uint32	w[2];

	char	c[8];
	int16	s[4];
	int32	i[2];

	float	f[2];
	double	d[1];
};



union vmp_oword {
	byte	b[16];
	uint16	h[8];
	uint32	w[4];

	char	c[16];
	int16	s[8];
	int32	i[4];

	float	f[4];
	double	d[2];
};


#ifndef CC_VMP_MAC
#ifndef bool
typedef enum
{
	vmp_false = 0,
	vmp_true,
} vmp_bool;

	//	Don't use vmp_false, vmp_true... instead use { bool, false, true } !!!
	//	Makes bool true : false compatible with C and C++

#define false   vmp_false
#define true    vmp_true
#define bool    vmp_bool
#endif
#endif



#define ALIGN2( len ) (( len + 1 ) & ~1 )     // round up to 16bits
#define ALIGN4( len ) (( len + 3 ) & ~3 )     // round up to 32bits
#define ALIGN8( len ) (( len + 7 ) & ~7 )     // round up to 64bits
#define ALIGN16( len ) (( len + 15 ) & ~15 )  // round up to 128bits

#define ALIGN2048( len ) (( len + 2047 ) & ~2047 )  // round up to 1 CD Sector


#define USE_ASSERT

#ifdef USE_ASSERT
#define ASSERT( arg1 ) assert( (arg1) )
#define ASSERT_PTR( arg1 ) assert(NULL!=(arg1))
#define ASSERT_PTR4( arg1 ) assert((NULL!=(arg1)) && (0==(((long)(arg1))&3)))
#define ASSERT_PTR16( arg1 ) assert((NULL!=(arg1)) && (0==(((long)(arg1))&15)))
#define ASSERT_FNEG( arg1 ) assert(0.0f<=(arg1))
#define ASSERT_FZERO( arg1 ) assert(0.0f!=(arg1))
#define ASSERT_NEG( arg1 ) assert(0<=(arg1))
#define ASSERT_ZERO( arg1 ) assert(0!=(arg1))

#else
#define ASSERT( arg1 )
#define ASSERT_PTR( arg1 )
#define ASSERT_PTR16( arg1 )
#define ASSERT_FNEG( arg1 )
#define ASSERT_FZERO( arg1 )
#define ASSERT_NEG( arg1 )
#define ASSERT_ZERO( arg1 )
#endif



#if defined CC_VMP_WIN32
#define 					VMP_LITTLE_ENDIAN
#elif defined CC_VMP_XBOX
#define 					VMP_LITTLE_ENDIAN
#elif defined CC_VMP_MAC
#define 					VMP_BIG_ENDIAN
#elif defined CC_VMP_PS2
#define 					VMP_LITTLE_ENDIAN
#else
#define 					VMP_LITTLE_ENDIAN
#endif



#define VMP_ENDIAN32( a ) (((a>>24)&0x000000ff) | ((a>>8)&0x0000ff00) | ((a<<8)&0x00ff0000) | (a<<24))
#define VMP_ENDIAN16( a ) (((a>>8)&0x00ff) | (a<<8))



//	A little Pretzel logic here, but If the 32bit data being read is in Little Endian
//	but is to be read by Big and Little Endian machines, put a VMP_LITTLE_ENDIAN32()
//	around it.  Indicating the data is Little Endian.  The macros below will resolve it.
//
//	If the Data being read is in Big Endian, put a VMP_BIG_ENDIAN32() around it.  It will
//	be resolved as well.  This is by stubbed Macro, or data conversion!

#ifdef VMP_BIG_ENDIAN			// Big Endian Processor

	//	Big Endian Data on Big Endian Processor
#define VMP_BIG_ENDIAN32
#define VMP_BIG_ENDIAN16
	//	Little Endian Data on Big Endian Processor
#define	VMP_LITTLE_ENDIAN32		VMP_ENDIAN32
#define VMP_LITTLE_ENDIAN16		VMP_ENDIAN16

#else							// Little Endian Processor

	//	Big Endian Data on Little Endian Processor
#define	VMP_BIG_ENDIAN32		VMP_ENDIAN32
#define VMP_BIG_ENDIAN16		VMP_ENDIAN16
	//	Little Endian Data on Little Endian Processor
#define VMP_LITTLE_ENDIAN32
#define VMP_LITTLE_ENDIAN16

#endif


#ifdef __cplusplus
}
#endif

#endif //__KariType_H



