  
//
//  vmp3D.h - 3D Math Core Header
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to accomodate the book.
//
//		This module is designed for fast 3D math
//
//  Created - 3 Oct 2001 - James Leiterman
//
//  Last Modification
//
//  stab=4
//


#include    "KariType.h"                                // Kari Games Master Definitions


#ifndef __vmpCore_H
#define __vmpCore_H

#ifdef __cplusplus
extern "C" {
#endif



#define PI          (3.141592f)
#define DEG2RAD(x)  (((x) * PI) / 180.0F)
#define RAD2DEG(x)  (((x) * 180.0f) / PI)


#define SINGLE_PRECISION	0.0000001f
#define FAST_PRECISION		0.001f




//
// 3D Coordinate (Single Precision Floating Point)
//

typedef struct vmp3DVectorType
{
    float x;
    float y;
    float z;
} vmp3DVector;



//
// 3D Coordinate Quad Vector (Single Precision Floating Point)
//

typedef struct vmp3DQVectorType
{
    float x;
    float y;
    float z;
	float w;
} vmp3DQVector;




//
//	Texture Coordinate
//

typedef struct vmp3DVertexType
{
	float x;
	float y;
	float z;

	float nx;											// Normals
	float ny;
	float nz;

	float tu;											// Texture U-coord.
	float tv;											// Texture V-coord.
} vmp3DVertex;



//
//  3D Matrix
//

typedef struct vmp3DMatrix1Type
{
		    float _11, _12, _13, _14;
			float _21, _22, _23, _24;
			float _31, _32, _33, _34;
			float _41, _42, _43, _44; 
} vmp3DMatrix1;			

typedef struct vmp3DMatrix0Type
{
		    float _00, _01, _02, _03;
			float _10, _11, _12, _13;
			float _20, _21, _22, _23;
			float _30, _31, _32, _33; 
} vmp3DMatrix0;
					

#ifndef CC_VMP_PS2
typedef float  vmp3DMatrix[4][4];
#else
//typedef int		IVECTOR[4] __attribute__((aligned (16)));
//typedef float		FVECTOR[4] __attribute__((aligned (16)));
typedef float  vmp3DMatrix[4][4] __attribute__((aligned (16)));
#endif


// I would love to use a union but not very portable!



#ifdef CC_VMP_MAC
//	Code visual cleanup (no casting)
//
//	NOTE: there are two different schools of thought as to casting versus union's, so I've included a copy of both!

union vec 
{
	vmp3DQVector qv;
	vector float vf;
};
#endif



//
//	SIMD Entry Point
//
#ifdef CC_VMP_WIN32
typedef void (*vmp_SIMDEntryProc)(void);				// Function pointer
extern vmp_SIMDEntryProc vmp_SIMDEntry;
#else
#define vmp_SIMDEntry()
#endif



//
//	SIMD Exit Point
//
#ifdef CC_VMP_WIN32
typedef void (*vmp_SIMDExitProc)(void);					// Function pointer
extern vmp_SIMDExitProc vmp_SIMDExit;
#else
#define vmp_SIMDExit()
#endif


bool vmp_IsFEqual(float fA, float fB, float fPrecision);
bool vmp_IsVEqual( const vmp3DVector * const pvA, const vmp3DVector * const pvB, float fPrecision );
bool vmp_IsQVEqual( const vmp3DQVector * const pvA, const vmp3DQVector * const pvB, float fPrecision );

bool vmp_IsMatrixEqual( vmp3DMatrix aMx, vmp3DMatrix bMx, float fPrecision);



#ifdef __cplusplus
}
#endif

#endif	//__vmp3D_H


